/* db_vmint.c - bytecode interpreter for a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "db_vm_v010.h"

/* prototypes for local functions */
static void Abort(char *fmt, ...);

/* Execute - execute the main code */
void Execute(Interpreter *i)
{
    int16_t tmp, tmp2, ind1, ind2;
    
    i->pc = i->code + i->mainCodeOffset;
    i->sp = i->fp = i->stack - 1;
	i->pollCount = BC_POLL_INTERVAL;

	for (;;) {
		if (--i->pollCount <= 0) {
			i->pollCount = BC_POLL_INTERVAL;
			(*i->poll)();
		}
#ifdef DEBUGIOx
		ShowStack(i);
		DecodeInstruction(0, i->code, i->pc);
#endif
		switch (*i->pc++) {
		case OP_HALT:
			return;
		case OP_BRT:
			tmp = *i->pc++;
			tmp |= (int16_t)(*i->pc++ << 8);
			if (Pop(i))
				i->pc += tmp;
			break;
		case OP_BRTSC:
			tmp = *i->pc++;
			tmp |= (int16_t)(*i->pc++ << 8);
			if (*i->sp)
				i->pc += tmp;
			else
				Drop(i, 1);
			break;
		case OP_BRF:
			tmp = *i->pc++;
			tmp |= (int16_t)(*i->pc++ << 8);
			if (!Pop(i))
				i->pc += tmp;
			break;
		case OP_BRFSC:
			tmp = *i->pc++;
			tmp |= (int16_t)(*i->pc++ << 8);
			if (!*i->sp)
				i->pc += tmp;
			else
				Drop(i, 1);
			break;
		case OP_BR:
			tmp = *i->pc++;
			tmp |= (int16_t)(*i->pc++ << 8);
			i->pc += tmp;
			break;
		case OP_NOT:
			*i->sp = (*i->sp ? FALSE : TRUE);
			break;
		case OP_NEG:
			*i->sp = -*i->sp;
			break;
		case OP_ADD:
			tmp = Pop(i);
			*i->sp += tmp;
			break;
		case OP_SUB:
			tmp = Pop(i);
			*i->sp -= tmp;
			break;
		case OP_MUL:
			tmp = Pop(i);
			*i->sp *= tmp;
			break;
		case OP_DIV:
			tmp = Pop(i);
			*i->sp = (tmp == 0 ? 0 : *i->sp / tmp);
			break;
		case OP_REM:
			tmp = Pop(i);
			*i->sp = (tmp == 0 ? 0 : *i->sp % tmp);
			break;
		case OP_BNOT:
			*i->sp = ~*i->sp;
			break;
		case OP_BAND:
			tmp = Pop(i);
			*i->sp &= tmp;
			break;
		case OP_BOR:
			tmp = Pop(i);
			*i->sp |= tmp;
			break;
		case OP_BXOR:
			tmp = Pop(i);
			*i->sp ^= tmp;
			break;
		case OP_SHL:
			tmp = Pop(i);
			*i->sp <<= tmp;
			break;
		case OP_SHR:
			tmp = Pop(i);
			*i->sp >>= tmp;
			break;
		case OP_LT:
			tmp = Pop(i);
			*i->sp = (*i->sp < tmp ? TRUE : FALSE);
			break;
		case OP_LE:
			tmp = Pop(i);
			*i->sp = (*i->sp <= tmp ? TRUE : FALSE);
			break;
		case OP_EQ:
			tmp = Pop(i);
			*i->sp = (*i->sp == tmp ? TRUE : FALSE);
			break;
		case OP_NE:
			tmp = Pop(i);
			*i->sp = (*i->sp != tmp ? TRUE : FALSE);
			break;
		case OP_GE:
			tmp = Pop(i);
			*i->sp = (*i->sp >= tmp ? TRUE : FALSE);
			break;
		case OP_GT:
			tmp = Pop(i);
			*i->sp = (*i->sp > tmp ? TRUE : FALSE);
			break;
		case OP_LIT:
			tmp = *i->pc++;
			CPush(i, (int16_t)(*i->pc++ << 8) | tmp);
			break;
		case OP_GREF:
			tmp = *i->pc++;
			CPush(i, i->variables[(int16_t)(*i->pc++ << 8) | tmp]);
			break;
		case OP_GSET:
			tmp = *i->pc++;
			i->variables[(int16_t)(*i->pc++ << 8) | tmp] = Pop(i);
			break;
		case OP_LREF:
			CPush(i, i->fp[(int)*(int8_t *)(i->pc)++]);
			break;
		case OP_LSET:
			i->fp[(int)*(int8_t *)(i->pc)++] = Pop(i);
			break;
		case OP_LOFF:
			tmp = *i->pc++;
			CPush(i, (int16_t)(i->fp - i->variables) + tmp);
			break;
		case OP_AREF1:
			ind1 = Pop(i);
			tmp = i->sp[0];
			if (ind1 < 1 || ind1 > i->variables[tmp])
				Abort("array subscript out of bounds: %d", tmp);
			i->sp[0] = i->variables[tmp + ind1];
			break;
		case OP_ASET1:
			tmp2 = Pop(i);
			ind1 = Pop(i);
			tmp = Pop(i);
			if (ind1 < 1 || ind1 > i->variables[tmp])
				Abort("array subscript out of bounds: %d", tmp);
			i->variables[tmp + ind1] = tmp2;
			break;
		case OP_AREF2:
			ind2 = Pop(i);
			ind1 = Pop(i);
			tmp = i->sp[0];
			if (ind1 < 1 || ind1 > i->variables[tmp])
				Abort("array subscript out of bounds: %d", ind1);
			else if (ind2 < 1 || ind2 > i->variables[tmp - 1])
				Abort("array subscript out of bounds: %d", ind2);
			i->sp[0] = i->variables[tmp + ((ind2 - 1) * i->variables[tmp]) + ind1];
			break;
		case OP_ASET2:
			tmp2 = Pop(i);
			ind2 = Pop(i);
			ind1 = Pop(i);
			tmp = Pop(i);
			if (ind1 < 1 || ind1 > i->variables[tmp])
				Abort("array subscript out of bounds: %d", ind1);
			else if (ind2 < 1 || ind2 > i->variables[tmp - 1])
				Abort("array subscript out of bounds: %d", ind2);
			i->variables[tmp + ((ind2 - 1) * i->variables[tmp]) + ind1] = tmp2;
			break;
		case OP_RESERVE:
			tmp = *i->pc++;
			Reserve(i, tmp);
			break;
		case OP_CALL:
			tmp = i->sp[-(int16_t)*i->pc++];
			if (tmp & 1)
				(*i->intrinsics[tmp >> 1])(i);
			else {
				tmp2 = i->fp - i->stack;
				i->fp = i->sp;
				Reserve(i, F_SIZE);
				i->fp[F_FP] = tmp2;
				i->fp[F_PC] = i->pc - i->code;
				i->pc = i->code + tmp;
			}
			break;
		case OP_FRET:
			i->pc = i->code + i->fp[F_PC];
			i->sp = i->fp;
			Drop(i, i->pc[-1]);
			i->fp = i->stack + i->fp[F_FP];
			break;
		case OP_SRET:
			i->pc = i->code + i->fp[F_PC];
			i->sp = i->fp;
			Drop(i, i->pc[-1] + 1);
			i->fp = i->stack + i->fp[F_FP];
			break;
		default:
			Abort("undefined opcode 0x%02x", i->pc[-1]);
			break;
		}
	}
}

void StackOverflow(void)
{
    Abort("stack overflow");
}

static void Abort(char *fmt, ...)
{
#ifdef DEBUGIO
    va_list ap;
    va_start(ap, fmt);
    VM_puts("error: ");
    VM_vprintf(fmt, ap);
    va_end(ap);
    VM_putchar('\n');
#endif
    exit(1);
}
