#include "XGS_PIC_SYSTEM_V010.h"            
#include "XGS_PIC_UART_DRV_V010.h"
#include "FSIO.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "db_vm_v010.h"
#include "db_xmrcv_v010.h"

typedef enum {
    RCV_WAIT_SOH,
    RCV_BLK1,
    RCV_BLK2,
    RCV_DATA,
    RCV_CRC1,
    RCV_CRC2,
    RCV_PURGE
} RcvState;

#define RETRY_INFINITE	(-1)

typedef struct {
    RcvState state;
    int nak;
    int retry;
    int cancel;
    int timeout;
    uint8_t packet[SBLKLEN];
    uint8_t nextBlk;
    uint8_t blk1;
    uint8_t blk2;
    uint16_t chksum;
    int dataCount;
    FSFILE *fp;
    long size;
    uint16_t timerCount;
    uint16_t lastTime;
} XmState;

static XmState xm;

#define EnableTimerInt()    (IEC0bits.T3IE = 1)
#define DisableTimerInt()   (IEC0bits.T3IE = 0)

static void Init(void);
static int HandlePacket(void);
static int ProcessOpen(void);
static int ProcessData(void);
static void ProcessClose(void);

void XmInit(void)
{
    /* start timer 3 running at 10Hz */
    TMR3 = 0;               // clear the current timer value
    PR3 = 19531;            // 50mhz / 256 / 19531 = ~.1 seconds
    T3CONbits.TCKPS = 3;    // 1:256 prescale
    T3CONbits.TCS = 0;      // use internal clock source
    IPC2bits.T3IP = 4;      // set interrupt priority to 4 (0->7)
    IFS0bits.T3IF = 0;      // clear interrupt flag in status register
    T3CONbits.TON = 1;      // turn the timer on

    /* initialize the xmodem receive state */
    Init();
    xm.fp = NULL;
    xm.timerCount = 0;

    /* notify the sender that we're ready to receive */
    UART_putchar(xm.nak);

    EnableTimerInt();
}

void XmPoll(void)
{
    uint8_t ch;

    /* handle received characters */
    if (UART_getchar(&ch)) {
        do {
            xm.lastTime = xm.timerCount;
            switch (xm.state) {
            case RCV_WAIT_SOH:
                switch (ch) {
                case SOH:
                    xm.timeout = 10;
                    xm.state = RCV_BLK1;
                    break;
                case EOT:
                    ProcessClose();
                    UART_putchar(ACK);
                    Init();
                    break;
                case CAN:
                    if (++xm.cancel >= 2) {
                        ProcessClose();
                        Init();
                    }
                    break;
                }
                break;
            case RCV_BLK1:
                xm.blk1 = ch;
                xm.state = RCV_BLK2;
                break;
            case RCV_BLK2:
                xm.blk2 = ch;
                xm.dataCount = 0;
                xm.chksum = 0;
                xm.state = RCV_DATA;
                break;
            case RCV_DATA:
                xm.packet[xm.dataCount] = ch;
                xm.chksum = updcrc(xm.chksum, ch);
                if (++xm.dataCount >= SBLKLEN)
                    xm.state = RCV_CRC1;
                break;
            case RCV_CRC1:
                xm.chksum = updcrc(xm.chksum, ch);
                xm.state = RCV_CRC2;
                break;
            case RCV_CRC2:
                xm.chksum = updcrc(xm.chksum, ch);
                if (HandlePacket())
                    UART_putchar(ACK);
                else
                    UART_putchar(NAK);
                xm.state = RCV_WAIT_SOH;
                break;
            case RCV_PURGE:
                break;
            default:
                xm.retry = 0;
                xm.state = RCV_PURGE;
                break;
            }
        } while (UART_getchar(&ch));
    }

    /* check for timeout */
    else {
        int thisTime = xm.timerCount;
        if (thisTime - xm.lastTime >= xm.timeout) {
            if (xm.retry == RETRY_INFINITE || --xm.retry >= 0) {
                xm.lastTime = thisTime;
                UART_putchar(xm.nak);
            }
            else
                Init();
        }
    }
}

static void Init(void)
{
    xm.state = RCV_WAIT_SOH;
    xm.nextBlk = 1;
    xm.nak = CRC;
    xm.retry = RETRY_INFINITE;
    xm.cancel = 0;
    xm.timeout = 10;
}

static int HandlePacket(void)
{
    /* check the block numbers */
    if (xm.blk1 != (~xm.blk2 & 0xff))
        return FALSE;
    
    /* check the checksum */
    else if (xm.chksum != 0x0000)
        return FALSE;

    /* check for a filename block */
    if (xm.blk1 == 0) {
        ProcessClose();
        return ProcessOpen();
    }
    
    /* check for a duplicate block */
    else if (xm.blk1 == ((xm.nextBlk - 1) & 0xff))
        return TRUE;

    /* check for a block sequence error */
    else if (xm.blk1 != xm.nextBlk)
        return FALSE;

    /* update the next block number */
    ++xm.nextBlk;

    /* handle a data block */
    return ProcessData();
}

static int ProcessOpen(void)
{
    char *name;
	
	/* close any currently open file */
	ProcessClose();
	
	/* get the filename */
	name = (char *)xm.packet;

	/* check for the start of a new transfer */
    if (name[0]) {
        if (!(xm.fp = FSfopen(name, "wb")))
            return FALSE;
        xm.size = atol((char *)&xm.packet[strlen(name) + 1]);
    }
	
	/* file was successfully open or end of batch transfer was handled */
    return TRUE;
}

static int ProcessData(void)
{
    if (!xm.fp) {
        xm.fp = FSfopen("xmodem.dat", "wb");
        xm.size = 0;
    }
    if (xm.fp) {
        int count = (!xm.size || xm.size > SBLKLEN ? SBLKLEN : (int)xm.size);
        if (FSfwrite(xm.packet, 1, count, xm.fp) != (size_t)count)
            return FALSE;
        if (xm.size)
            xm.size -= count;
    }
    return TRUE;
}

static void ProcessClose(void)
{
    if (xm.fp) {
        FSfclose(xm.fp);
        xm.fp = NULL;
    }
}

/* timer usage:

    1:   video
    2:   sound
    3:   xmodem
    4-9: unused

*/

void __attribute__((interrupt, no_auto_psv)) _T3Interrupt(void)
{
    ++xm.timerCount;
   IFS0bits.T3IF = 0;       // Clear interrupt flag so it can interrupt again
} 
