/* db_expr.c - expression parser
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <string.h>
#include "db_bascom_v010.h"

/* prototypes */
static void ParseExpr2(ParseContext *c, PVAL *pv);
static void ParseExpr3(ParseContext *c, PVAL *pv);
static void ParseExpr4(ParseContext *c, PVAL *pv);
static void ParseExpr5(ParseContext *c, PVAL *pv);
static void ParseExpr6(ParseContext *c, PVAL *pv);
static void ParseExpr7(ParseContext *c, PVAL *pv);
static void ParseExpr8(ParseContext *c, PVAL *pv);
static void ParseExpr9(ParseContext *c, PVAL *pv);
static void ParseExpr10(ParseContext *c, PVAL *pv);
static void ParseExpr11(ParseContext *c, PVAL *pv);
static void ParseSimplePrimary(ParseContext *c, PVAL *pv);
static void ParseFunctionCall(ParseContext *c, PVAL *pv);
static void ParseCall(ParseContext *c, PVAL *pv);
static void ParseArrayReference(ParseContext *c, PVAL *pv);
static void code_literal(ParseContext *c, int fcn, PVAL *pv);
static void code_global(ParseContext *c, int fcn, PVAL *pv);
static int GREF_inst(Type *type);
static int GSET_inst(Type *type);
static void code_argument(ParseContext *c, int fcn, PVAL *pv);
static int AREF_inst(Type *type);
static int ASET_inst(Type *type);
static void code_local(ParseContext *c, int fcn, PVAL *pv);
static int LREF_inst(Type *type);
static int LSET_inst(Type *type);
static void code_index(ParseContext *c, int fcn, PVAL *pv);
static int AREF1_inst(Type *type);
static int AREF2_inst(Type *type);
static int ASET1_inst(Type *type);
static int ASET2_inst(Type *type);

/* ParseExpr - handle the OR operator */
void ParseExpr(ParseContext *c, PVAL *pv)
{
    Token tkn;
    ParseExpr2(c, pv);
    if ((tkn = GetToken(c)) == T_OR) {
        int end = 0;
        rvalue(c, pv);
        do {
            putcbyte(c, OP_BRTSC);
            end = putcword(c, end);
            ParseExpr2(c, pv);
            rvalue(c, pv);
        } while ((tkn = GetToken(c)) == T_OR);
        fixupbranch(c, end, codeaddr(c));
    }
    SaveToken(c, tkn);
}

/* ParseExpr2 - handle the AND operator */
static void ParseExpr2(ParseContext *c, PVAL *pv)
{
    Token tkn;
    ParseExpr3(c, pv);
    if ((tkn = GetToken(c)) == T_AND) {
        int end = 0;
        rvalue(c, pv);
        do {
            putcbyte(c, OP_BRFSC);
            end = putcword(c, end);
            ParseExpr3(c, pv);
            rvalue(c, pv);
        } while ((tkn = GetToken(c)) == T_AND);
        fixupbranch(c, end, codeaddr(c));
    }
    SaveToken(c, tkn);
}

/* ParseExpr3 - handle the BXOR operator */
static void ParseExpr3(ParseContext *c, PVAL *pv)
{
    Token tkn;
    ParseExpr4(c, pv);
    while ((tkn = GetToken(c)) == '^') {
        PVAL pv2;
        rvalue(c, pv);
        ParseExpr4(c, &pv2);
        rvalue(c, &pv2);
        putcbyte(c, OP_BXOR);
    }
    SaveToken(c,tkn);
}

/* ParseExpr4 - handle the BOR operator */
static void ParseExpr4(ParseContext *c, PVAL *pv)
{
    Token tkn;
    ParseExpr5(c, pv);
    while ((tkn = GetToken(c)) == '|') {
        PVAL pv2;
        rvalue(c, pv);
        ParseExpr5(c, &pv2);
        rvalue(c, &pv2);
        putcbyte(c, OP_BOR);
    }
    SaveToken(c,tkn);
}

/* ParseExpr5 - handle the BAND operator */
static void ParseExpr5(ParseContext *c, PVAL *pv)
{
    Token tkn;
    ParseExpr6(c, pv);
    while ((tkn = GetToken(c)) == '&') {
        PVAL pv2;
        rvalue(c, pv);
        ParseExpr6(c, &pv2);
        rvalue(c, &pv2);
        putcbyte(c, OP_BAND);
    }
    SaveToken(c,tkn);
}

/* ParseExpr6 - handle the '=' and '<>' operators */
static void ParseExpr6(ParseContext *c, PVAL *pv)
{
    Token tkn;
    ParseExpr7(c, pv);
    while ((tkn = GetToken(c)) == '=' || tkn == T_NE) {
        PVAL pv2;
        int op;
        switch (tkn) {
        case '=':
            op = OP_EQ;
            break;
        case T_NE:
            op = OP_NE;
            break;
        }
        rvalue(c, pv);
        ParseExpr7(c, &pv2);
        rvalue(c, &pv2);
        putcbyte(c, op);
    }
    SaveToken(c,tkn);
}

/* ParseExpr7 - handle the '<', '<=', '>=' and '>' operators */
static void ParseExpr7(ParseContext *c, PVAL *pv)
{
    Token tkn;
    ParseExpr8(c, pv);
    while ((tkn = GetToken(c)) == '<' || tkn == T_LE || tkn == T_GE || tkn == '>') {
        PVAL pv2;
        int op;
        switch (tkn) {
        case '<':
            op = OP_LT;
            break;
        case T_LE:
            op = OP_LE;
            break;
        case T_GE:
            op = OP_GE;
            break;
        case '>':
            op = OP_GT;
            break;
        }
        rvalue(c, pv);
        ParseExpr8(c, &pv2);
        rvalue(c, &pv2);
        putcbyte(c, op);
    }
    SaveToken(c,tkn);
}

/* ParseExpr8 - handle the '<<' and '>>' operators */
static void ParseExpr8(ParseContext *c, PVAL *pv)
{
    Token tkn;
    ParseExpr9(c, pv);
    while ((tkn = GetToken(c)) == T_SHL || tkn == T_SHR) {
        PVAL pv2;
        int op;
        switch (tkn) {
        case T_SHL:
            op = OP_SHL;
            break;
        case T_SHR:
            op = OP_SHR;
            break;
        }
        rvalue(c, pv);
        ParseExpr9(c, &pv2);
        rvalue(c, &pv2);
        putcbyte(c, op);
    }
    SaveToken(c,tkn);
}

/* ParseExpr9 - handle the '+' and '-' operators */
static void ParseExpr9(ParseContext *c, PVAL *pv)
{
    Token tkn;
    ParseExpr10(c, pv);
    while ((tkn = GetToken(c)) == '+' || tkn == '-') {
        PVAL pv2;
        ParseExpr10(c, &pv2);
        if (IsLiteral(pv) && IsLiteral(&pv2)) {
            switch (tkn) {
            case '+':
                pv->val = pv->val + pv2.val;
                break;
            case '-':
                pv->val = pv->val - pv2.val;
                break;
            }
        }
        else {
            rvalue(c, pv);
            rvalue(c, &pv2);
            switch (tkn) {
            case '+':
                putcbyte(c, OP_ADD);
                break;
            case '-':
                putcbyte(c, OP_SUB);
                break;
            }
        }
    }
    SaveToken(c, tkn);
}

/* ParseExpr10 - handle the '*', '/' and MOD operators */
static void ParseExpr10(ParseContext *c, PVAL *pv)
{
    Token tkn;
    ParseExpr11(c, pv);
    while ((tkn = GetToken(c)) == '*' || tkn == '/' || tkn == T_MOD) {
        PVAL pv2;
        ParseExpr11(c, &pv2);
        if (IsLiteral(pv) && IsLiteral(&pv2)) {
            switch (tkn) {
            case '*':
                pv->val = pv->val * pv2.val;
                break;
            case '/':
                if (pv2.val == 0)
                    ParseError(c, "division by zero in constant expression");
                pv->val = pv->val / pv2.val;
                break;
            case T_MOD:
                if (pv2.val == 0)
                    ParseError(c, "division by zero in constant expression");
                pv->val = pv->val % pv2.val;
                break;
            }
        }
        else {
            rvalue(c, pv);
            rvalue(c, &pv2);
            switch (tkn) {
            case '*':
                putcbyte(c, OP_MUL);
                break;
            case '/':
                putcbyte(c, OP_DIV);
                break;
            case T_MOD:
                putcbyte(c, OP_REM);
                break;
            }
        }
    }
    SaveToken(c, tkn);
}

/* ParseExpr11 - handle unary operators */
static void ParseExpr11(ParseContext *c, PVAL *pv)
{
    Token tkn;
    switch (tkn = GetToken(c)) {
    case '-':
        ParsePrimary(c, pv);
        if (IsLiteral(pv))
            pv->val = -pv->val;
        else {
            rvalue(c, pv);
            putcbyte(c, OP_NEG);
        }
        break;
    case T_NOT:
        ParsePrimary(c, pv);
        if (IsLiteral(pv))
            pv->val = !pv->val;
        else {
            rvalue(c, pv);
            putcbyte(c, OP_NOT);
        }
        break;
    case '~':
        ParsePrimary(c, pv);
        if (IsLiteral(pv))
            pv->val = ~pv->val;
        else {
            rvalue(c, pv);
            putcbyte(c, OP_BNOT);
        }
        break;
    default:
        SaveToken(c,tkn);
        ParsePrimary(c, pv);
        break;
    }
}

/* ParsePrimary - parse function calls and array references */
void ParsePrimary(ParseContext *c, PVAL *pv)
{
    Token tkn;
    ParseSimplePrimary(c, pv);
    if ((tkn = GetToken(c)) == '(') {
        switch (pv->type->id) {
        case TYPE_FUNCTION:
            ParseFunctionCall(c, pv);
            break;
        case TYPE_ARRAY:
            ParseArrayReference(c, pv);
            break;
        default:
            ParseError(c, "expecting a function or an array");
        }
    }
    else
        SaveToken(c, tkn);
}

/* ParseSubroutineCall - parse a call to a subroutine */
void ParseSubroutineCall(ParseContext *c)
{
    PVAL pv;

    /* parse the function name */
    ParseSimplePrimary(c, &pv);

    /* make sure it's a function we're calling */
    if (pv.type->id != TYPE_FUNCTION)
        ParseError(c, "Attempt to call a non-function");
    else if (pv.type->FunctionInfo.returnType)
        ParseError(c, "attempt to call a function as a subroutine");

    /* parse the call */
    FRequire(c, '(');
    ParseCall(c, &pv);
}

/* ParseFunctionCall - parse a function or subroutine call */
static void ParseFunctionCall(ParseContext *c, PVAL *pv)
{
    /* make sure it's a function we're calling */
    if (!pv->type->FunctionInfo.returnType)
        ParseError(c, "attempt to call a subroutine as a function");

    /* parse the call */
    ParseCall(c, pv);
}

/* ParseCall - parse a function or subroutine call */
static void ParseCall(ParseContext *c, PVAL *pv)
{
    Symbol *arg;
    int argc = 0;
    Token tkn;
    PVAL pv2;
    
    /* get the value of the function */
    rvalue(c, pv);

    /* get the list of argument types */
    arg = pv->type->FunctionInfo.arguments.head;
    if (pv->type->FunctionInfo.returnType)
        arg = arg->next;

    /* compile each argument expression */
    if ((tkn = GetToken(c)) != ')') {
        SaveToken(c, tkn);
        do {
            ParseExpr(c, &pv2);
            rvalue(c, &pv2);
            if (!arg)
                ParseError(c, "too many arguments");
            else if (!CompareTypes(pv2.type, arg->type))
                ParseError(c, "wrong argument type");
            arg = arg->next;
            ++argc;
        } while ((tkn = GetToken(c)) == ',');
        Require(c, tkn, ')');
    }

    /* make sure there werent' too many arguments specified */
    if (arg)
        ParseError(c, "too few arguments");

    /* call the function */
    putcbyte(c, OP_CALL);
    putcbyte(c, argc);

    /* we've got an rvalue now */
    pv->type = pv->type->FunctionInfo.returnType;
    pv->fcn = NULL;
}

/* ParseArrayReference - parse an array reference */
static void ParseArrayReference(ParseContext *c, PVAL *pv)
{
    Token tkn;
    PVAL pv2;

    /* get the array offset */
    rvalue(c, pv);
    
    /* setup the element type */
    pv->type = pv->type->ArrayInfo.elementType;
    pv->fcn = code_index;
    
    /* get the first index */
    ParseExpr(c, &pv2);
    rvalue(c, &pv2);
    
    /* check for a second index */
    if ((tkn = GetToken(c)) == ',') {
        ParseExpr(c, &pv2);
        rvalue(c, &pv2);
        tkn = GetToken(c);
        pv->val = 2;
    }
    
    /* only one index */
    else
        pv->val = 1;
    
    /* check for the close paren */
    Require(c, tkn, ')');
}

/* ParseSimplePrimary - parse a primary expression */
static void ParseSimplePrimary(ParseContext *c, PVAL *pv)
{
    switch (GetToken(c)) {
    case '(':
        ParseExpr(c, pv);
        FRequire(c,')');
        break;
    case T_NUMBER:
        pv->type = &c->integerType;
        pv->fcn = code_literal;
        pv->val = c->value;
        break;
    case T_STRING:
        putcbyte(c, OP_LIT);
        putcword(c, AddString(c, c->token, codeaddr(c)));
        pv->type = &c->stringType;
        pv->fcn = NULL;
        break;
    case T_IDENTIFIER:
        GetLValue(c, c->token, pv);
        break;
    default:
        ParseError(c, "Expecting a primary expression");
        break;
    }
}

/* GetLValue - get an l-value */
void GetLValue(ParseContext *c, char *name, PVAL *pv)
{
    Symbol *sym;

    /* handle local variables within a function or subroutine */
    if (c->function && (sym = FindSymbol(&c->locals, name)) != NULL) {
        if (sym->storageClass == SC_CONSTANT) {
            pv->type = sym->type;
            pv->fcn = code_literal;
            pv->val = sym->value;
        }
        else {
            pv->type = sym->type;
            pv->fcn = code_local;
            pv->val = sym->offset;
        }
    }

    /* handle function or subroutine arguments or the return value symbol */
    else if (c->function && (sym = FindSymbol(&c->functionType->FunctionInfo.arguments, name)) != NULL) {
        pv->type = sym->type;
        pv->fcn = code_argument;
        pv->val = sym->offset - c->functionType->FunctionInfo.argumentCount;
    }

    /* handle global symbols */
    else if ((sym = FindSymbol(&c->globals, c->token)) != NULL) {
        if (sym->storageClass == SC_CONSTANT) {
            pv->type = sym->type;
            pv->fcn = code_literal;
            pv->val = sym->value;
        }
        else {
            pv->type = sym->type;
            pv->fcn = code_global;
            pv->val = sym->offset;
        }
    }

    /* handle undefined symbols */
    else {

        /* if inside a function or subroutine definition, add the symbol as a local */
        if (c->function) {
            if (!(sym = AddSymbol(&c->locals, name, SC_VARIABLE, DefaultType(c, name), c->localCount++ + F_SIZE + 1, 0)))
                Fatal("insufficient memory");
            pv->type = sym->type;
            pv->fcn = code_local;
            pv->val = sym->offset;
        }

        /* otherwise, add it as a global */
        else {
            if (!(sym = AddSymbol(&c->globals, name, SC_VARIABLE, DefaultType(c, name), c->globalCount++, 0)))
                Fatal("insufficient memory");
            pv->type = sym->type;
            pv->fcn = code_global;
            pv->val = sym->offset;
        }
    }
}

/* DefaultType - get the default type based on the variable name */
Type *DefaultType(ParseContext *c, char *name)
{
    Type *type;
    switch (name[strlen(name) - 1]) {
    case '$':
        type = &c->stringType;
        break;
    case '%':
        type = &c->integerType;
        break;
    default:
        type = &c->integerType; // should be &c->floatType
        break;
    }
    return type;
}

/* IsLiteral - check to see if a partial value is a literal (number) */
int IsLiteral(PVAL *pv)
{
    return pv->fcn == code_literal;
}

/* code_literal - compile a literal reference */
static void code_literal(ParseContext *c, int fcn, PVAL *pv)
{
    switch (fcn) {
    case PV_LOAD:
        putcbyte(c, OP_LIT);
        putcword(c, pv->val);
        break;
    case PV_STORE:
        ParseError(c, "can't store into a literal");
        break;
    }
}

/* code_global - compile a global variable reference */
static void code_global(ParseContext *c, int fcn, PVAL *pv)
{
    switch (fcn) {
    case PV_LOAD:
        putcbyte(c, GREF_inst(pv->type));
        putcword(c, pv->val);
        break;
    case PV_STORE:
        putcbyte(c, GSET_inst(pv->type));
        putcword(c, pv->val);
        break;
    }
}

/* GREF_inst - select a global reference instruction */
static int GREF_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_FLOAT:
        inst = OP_GREFF;
        break;
    case TYPE_INTEGER:
        inst = OP_GREF;
        break;
    case TYPE_BYTE:
        inst = OP_GREFB;
        break;
    case TYPE_STRING:
    case TYPE_ARRAY:
        inst = OP_LIT;
        break;
    default:
        Fatal("internal error: GREF_inst");
        break;
    }
    return inst;
}

/* GSET_inst - select a global set instruction */
static int GSET_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_FLOAT:
        inst = OP_GSETF;
        break;
    case TYPE_INTEGER:
    case TYPE_ARRAY:
    case TYPE_STRING:
        inst = OP_GSET;
        break;
    case TYPE_BYTE:
        inst = OP_GSETB;
        break;
    default:
        Fatal("internal error: GSET_inst");
        break;
    }
    return inst;
}

/* code_argument - compile an argument reference */
static void code_argument(ParseContext *c, int fcn, PVAL *pv)
{
    switch (fcn) {
    case PV_LOAD:
        putcbyte(c, AREF_inst(pv->type));
        putcbyte(c, pv->val);
        break;
    case PV_STORE:
        putcbyte(c, ASET_inst(pv->type));
        putcbyte(c, pv->val);
        break;
    }
}

/* AREF_inst - select an argument reference instruction */
static int AREF_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_FLOAT:
        inst = OP_LREFF;
        break;
    case TYPE_INTEGER:
        inst = OP_LREF;
        break;
    case TYPE_BYTE:
        inst = OP_LREFB;
        break;
    case TYPE_ARRAY:
    case TYPE_STRING:
        inst = OP_LREF;
        break;
    default:
        Fatal("internal error: AREF_inst");
        break;
    }
    return inst;
}

/* ASET_inst - select an argument set instruction */
static int ASET_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_FLOAT:
        inst = OP_LSETF;
        break;
    case TYPE_INTEGER:
    case TYPE_ARRAY:
    case TYPE_STRING:
        inst = OP_LSET;
        break;
    case TYPE_BYTE:
        inst = OP_LSETB;
        break;
    default:
        Fatal("internal error: ASET_inst");
        break;
    }
    return inst;
}

/* code_local - compile an local reference */
static void code_local(ParseContext *c, int fcn, PVAL *pv)
{
    switch (fcn) {
    case PV_LOAD:
        putcbyte(c, LREF_inst(pv->type));
        putcbyte(c, pv->val);
        break;
    case PV_STORE:
        putcbyte(c, LSET_inst(pv->type));
        putcbyte(c, pv->val);
        break;
    }
}

/* LREF_inst - select a local reference instruction */
static int LREF_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_FLOAT:
        inst = OP_LREFF;
        break;
    case TYPE_INTEGER:
        inst = OP_LREF;
        break;
    case TYPE_BYTE:
        inst = OP_LREFB;
        break;
    case TYPE_ARRAY:
    case TYPE_STRING:
        inst = OP_LOFF;
        break;
    default:
        Fatal("internal error: LREF_inst");
        break;
    }
    return inst;
}

/* LSET_inst - select a local set instruction */
static int LSET_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_FLOAT:
        inst = OP_LSETF;
        break;
    case TYPE_INTEGER:
    case TYPE_ARRAY:
    case TYPE_STRING:
        inst = OP_LSET;
        break;
    case TYPE_BYTE:
        inst = OP_LSETB;
        break;
    default:
        Fatal("internal error: LSET_inst");
        break;
    }
    return inst;
}

/* code_index - compile an array reference */
static void code_index(ParseContext *c, int fcn, PVAL *pv)
{
    switch (fcn) {
    case PV_LOAD:
        if (pv->val == 1)
            putcbyte(c, AREF1_inst(pv->type));
        else
            putcbyte(c, AREF2_inst(pv->type));
        break;
    case PV_STORE:
        if (pv->val == 1)
            putcbyte(c, ASET1_inst(pv->type));
        else
            putcbyte(c, ASET2_inst(pv->type));
        break;
    }
}

/* AREF1_inst - select a 1 index reference instruction */
static int AREF1_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_FLOAT:
        break;
    case TYPE_INTEGER:
    case TYPE_ARRAY:
    case TYPE_STRING:
        inst = OP_AREF1;
        break;
    case TYPE_BYTE:
        inst = OP_AREF1B;
        break;
    default:
        Fatal("internal error: AREF1_inst");
        break;
    }
    return inst;
}

/* AREF2_inst - select a 2 index reference instruction */
static int AREF2_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_FLOAT:
        break;
    case TYPE_INTEGER:
    case TYPE_ARRAY:
    case TYPE_STRING:
        inst = OP_AREF2;
        break;
    case TYPE_BYTE:
        inst = OP_AREF2B;
        break;
    default:
        Fatal("internal error: AREF2_inst");
        break;
    }
    return inst;
}

/* ASET1_inst - select a 1 index set instruction */
static int ASET1_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_FLOAT:
        break;
    case TYPE_INTEGER:
    case TYPE_ARRAY:
    case TYPE_STRING:
        inst = OP_ASET1;
        break;
    case TYPE_BYTE:
        inst = OP_ASET1B;
        break;
    default:
        Fatal("internal error: ASET1_inst");
        break;
    }
    return inst;
}

/* ASET2_inst - select a 2 index set instruction */
static int ASET2_inst(Type *type)
{
    int inst;
    switch (type->id) {
    case TYPE_FLOAT:
        break;
    case TYPE_INTEGER:
    case TYPE_ARRAY:
    case TYPE_STRING:
        inst = OP_ASET2;
        break;
    case TYPE_BYTE:
        inst = OP_ASET2B;
        break;
    default:
        Fatal("internal error: ASET2_inst");
        break;
    }
    return inst;
}

/* rvalue - get the rvalue of a partial expression */
void rvalue(ParseContext *c, PVAL *pv)
{
    if (pv->fcn) {
        (*pv->fcn)(c, PV_LOAD, pv);
        pv->fcn = NULL;
    }
}

/* chklvalue - make sure we've got an lvalue */
void chklvalue(ParseContext *c, PVAL *pv)
{
    if (pv->fcn == NULL)
        ParseError(c,"Expecting an lvalue");
}
