/* db_stat.c - statement parser
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdlib.h>
#include <string.h>
#include "db_bascom_v010.h"

/* statement handler prototypes */
static void ParseDef(ParseContext *c);
static void ParseSub(ParseContext *c);
static void ParseType(ParseContext *c);
static void ParseDim(ParseContext *c);
static Type *ParseVariableDecl(ParseContext *c, char *name);
static void ParseLet(ParseContext *c);
static void ParseIf(ParseContext *c);
static void ParseElse(ParseContext *c);
static void ParseElseIf(ParseContext *c);
static void ParseEnd(ParseContext *c);
static void ParseEndDef(ParseContext *c);
static void ParseEndSub(ParseContext *c);
static void ParseConstantDef(ParseContext *c, char *name);
static void ParseDefOrSub(ParseContext *c, char *name, int isFunction);
static void ParseEndDefOrSub(ParseContext *c);
static void ParseEndIf(ParseContext *c);
static void ParseFor(ParseContext *c);
static void ParseNext(ParseContext *c);
static void ParseDo(ParseContext *c);
static void ParseDoWhile(ParseContext *c);
static void ParseDoUntil(ParseContext *c);
static void ParseLoop(ParseContext *c);
static void ParseLoopWhile(ParseContext *c);
static void ParseLoopUntil(ParseContext *c);
static void ParseCall(ParseContext *c);
static void ParseStop(ParseContext *c);
static void ParsePrint(ParseContext *c);
static void ParseInput(ParseContext *c);

/* prototypes */
static BlockType CheckBlock(ParseContext *c);
static void PushBlock(ParseContext *c);
static void PopBlock(ParseContext *c);

/* ParseStatement - parse a statement */
void ParseStatement(ParseContext *c)
{
    Token tkn;
    
    /* dispatch on the statement keyword */
    switch (tkn = GetToken(c)) {
    case T_REM:
        /* just a comment so ignore the rest of the line */
        break;
    case T_DEF:
        ParseDef(c);
        break;
    case T_SUB:
        ParseSub(c);
        break;
    case T_TYPE:
        ParseType(c);
        break;
    case T_DIM:
        ParseDim(c);
        break;
    case T_LET:
        ParseLet(c);
        break;
    case T_IF:
        ParseIf(c);
        break;
    case T_ELSE:
        ParseElse(c);
        break;
    case T_ELSE_IF:
        ParseElseIf(c);
        break;
    case T_END:
        ParseEnd(c);
        break;
    case T_END_DEF:
        ParseEndDef(c);
        break;
    case T_END_SUB:
        ParseEndSub(c);
        break;
    case T_END_IF:
        ParseEndIf(c);
        break;
    case T_FOR:
        ParseFor(c);
        break;
    case T_NEXT:
        ParseNext(c);
        break;
    case T_DO:
        ParseDo(c);
        break;
    case T_DO_WHILE:
        ParseDoWhile(c);
        break;
    case T_DO_UNTIL:
        ParseDoUntil(c);
        break;
    case T_LOOP:
        ParseLoop(c);
        break;
    case T_LOOP_WHILE:
        ParseLoopWhile(c);
        break;
    case T_LOOP_UNTIL:
        ParseLoopUntil(c);
        break;
    case T_CALL:
        ParseCall(c);
        break;
    case T_STOP:
        ParseStop(c);
        break;
    case T_PRINT:
        ParsePrint(c);
        break;
    case T_INPUT:
        ParseInput(c);
        break;
    default:    /* default to 'LET' */
        SaveToken(c, tkn);
        ParseLet(c);
        break;
    }
}

/* ParseDef - parse the 'DEF' statement */
static void ParseDef(ParseContext *c)
{
    char name[MAXTOKEN];
    Token tkn;

    /* get the name being defined */
    FRequire(c, T_IDENTIFIER);
    strcpy(name, c->token);

    /* check for a constant definition */
    if ((tkn = GetToken(c)) == '=')
        ParseConstantDef(c, name);
    
    /* otherwise, assume a function definition */
    else {
        SaveToken(c, tkn);
        ParseDefOrSub(c, name, TRUE);
    }
}

/* ParseEndDef - parse the 'END DEF' statement */
static void ParseEndDef(ParseContext *c)
{
    ParseEndDefOrSub(c);
}

/* ParseSub - parse the 'SUB' statement */
static void ParseSub(ParseContext *c)
{
    char name[MAXTOKEN];
    FRequire(c, T_IDENTIFIER);
    strcpy(name, c->token);
    ParseDefOrSub(c, name, FALSE);
}

/* ParseEndSub - parse the 'END SUB' statement */
static void ParseEndSub(ParseContext *c)
{
    ParseEndDefOrSub(c);
}

/* ParseConstantDef - parse a 'DEF <name> =' statement */
static void ParseConstantDef(ParseContext *c, char *name)
{
    Symbol *sym;
    PVAL pv;

    /* get the constant value */
    ParseExpr(c, &pv);

    /* make sure it's a constant */
    if (!IsLiteral(&pv))
        ParseError(c, "expecting a constant expression");

    /* if inside a function or subroutine definition, add the symbol as a local */
    if (c->function) {
        if (!(sym = AddSymbol(&c->locals, name, SC_CONSTANT, &c->integerType, 0, pv.val)))
            Fatal("insufficient memory");
    }

    /* otherwise, add it as a global */
    else {
        if (!(sym = AddSymbol(&c->globals, name, SC_CONSTANT, &c->integerType, 0, pv.val)))
            Fatal("insufficient memory");
    }
    FRequire(c, T_EOL);
}

/* ParseDefOrSub - parse a 'DEF' or 'SUB' statement */
static void ParseDefOrSub(ParseContext *c, char *name, int isFunction)
{
    Token tkn;

    /* don't allow nested functions (for now anyway) */
    if (c->function)
        Fatal("nested functions not supported");

    /* create a type for the function */
    if (!(c->functionType = (Type *)malloc(sizeof(Type))))
        Fatal("insufficient memory");

    /* initialize the type */
    c->functionType->id = TYPE_FUNCTION;
    InitSymbolTable(&c->functionType->FunctionInfo.arguments);
    c->functionType->FunctionInfo.argumentCount = 0;

    /* initialize the local symbol table */
    InitSymbolTable(&c->locals);
    c->localCount = 0;

    /* make sure the function starts at an even address */
    if (c->codeOff & 1) {
        putc('\0', c->ofp);
        ++c->codeOff;
    }
    
    /* set the return type for functions */
    if (isFunction) {
        if (!AddSymbol(&c->functionType->FunctionInfo.arguments, name, SC_VARIABLE, DefaultType(c, name), 0, 0))
            Fatal("insufficient memory");
        c->functionType->FunctionInfo.returnType = &c->integerType;
    }

    /* no return type for subroutines */
    else
        c->functionType->FunctionInfo.returnType = NULL;

    /* get the argument list */
    if ((tkn = GetToken(c)) == '(') {
        if ((tkn = GetToken(c)) != ')') {
            int offset = 1;
            SaveToken(c, tkn);
            do {
                char name[MAXTOKEN];
                Type *type;
                type = ParseVariableDecl(c, name);
                if (!AddSymbol(&c->functionType->FunctionInfo.arguments, name, SC_VARIABLE, type, offset, 0))
                    Fatal("insufficient memory");
                ++c->functionType->FunctionInfo.argumentCount;
                ++offset;
            } while ((tkn = GetToken(c)) == ',');
        }
        Require(c, tkn, ')');
    }
    else
        SaveToken(c, tkn);

    /* enter the function name in the global symbol table */
    if (!(c->function = AddSymbol(&c->globals, name, SC_CONSTANT, c->functionType, 0, c->codeOff)))
        Fatal("insufficient memory");

    /* create the function prolog */
    c->cbase = c->cptr;
    putcbyte(c, OP_RESERVE);
    putcbyte(c, 0);
    
    FRequire(c, T_EOL);
}

/* ParseEndDefOrSub - parse an 'END DEF' or 'END SUB' statement */
static void ParseEndDefOrSub(ParseContext *c)
{
    if (!c->functionType)
        ParseError(c, "not in a function definition");
    else {
        if (c->functionType->FunctionInfo.returnType)
            putcbyte(c, OP_FRET);
        else
            putcbyte(c, OP_SRET);
        c->cbase[1] = c->localCount;
        WriteFunction(c, c->function->name);
#ifdef DEBUGIO
        if (c->functionType->FunctionInfo.arguments.head) {
            printf("arguments:\n");
            DumpSymbols(&c->functionType->FunctionInfo.arguments, FALSE);
        }
        if (c->locals.head) {
            printf("locals:\n");
            DumpSymbols(&c->locals, FALSE);
        }
        printf("\n");
#endif
        c->function = NULL;
        c->functionType = NULL;
        FreeSymbols(&c->locals);
    }
}

/* ParseType - parse the 'TYPE' statement */
static void ParseType(ParseContext *c)
{
    /*
        type <name> as <type>
        type <name> as <type> ( [ <size> ] [ , [ <size> ] ] )
        type <name> as def [ (<args>) ]
        type <name> as sub [ (<args>) ]
        type <name>
          <field-name> [ as <type> ]
          ...
        end type

        <args>:
            <arg>
            <args> , <arg>

        <arg>:
            <name> [ as <type> ]
    */
}

/* ParseDim - parse the 'DIM' statement */
static void ParseDim(ParseContext *c)
{
    int valuePrefix, valueSize;
    char name[MAXTOKEN];
    Symbol *sym;
    Type *type;
    Token tkn;

    /* parse variable declarations */
    do {

        /* parse a variable declaration */
        type = ParseVariableDecl(c, name);

        /* get the size of the value */
        valueSize = ValueSize(type, &valuePrefix);

        /* add to the local symbol table if inside a function definition */
        if (c->function) {
            int offset = c->localCount + F_SIZE + 1 + valuePrefix;
            if (type->id == TYPE_ARRAY) {
                putcbyte(c, OP_LIT);
                putcword(c, type->ArrayInfo.dim1);
                putcbyte(c, OP_LSET);
                putcbyte(c, offset - 1);
                putcbyte(c, OP_LIT);
                putcword(c, type->ArrayInfo.dim1 * type->ArrayInfo.dim2);
                putcbyte(c, OP_LSET);
                putcbyte(c, offset - 2);
            }
            if (!(sym = AddSymbol(&c->locals, name, SC_VARIABLE, type, offset, 0)))
                Fatal("insufficient memory");
            c->localCount += valueSize;
        }
        
        /* otherwise, add to the global symbol table */
        else {
            if (!(sym = AddSymbol(&c->globals, name, SC_VARIABLE, type, c->globalCount + valuePrefix, 0)))
                Fatal("insufficient memory");
            c->globalCount += valueSize;
        }

    } while ((tkn = GetToken(c)) == ',');

    Require(c, tkn, T_EOL);
}

/* ParseVariableDecl - parse a variable declaration */
static Type *ParseVariableDecl(ParseContext *c, char *name)
{
    Type *type, **pBaseType;
    Token tkn;

    /* parse the variable name */
    FRequire(c, T_IDENTIFIER);
    strcpy(name, c->token);
    
    /* determine the default type based on the variable name */
    type = DefaultType(c, name);
    pBaseType = &type;

    /* handle arrays */
    if ((tkn = GetToken(c)) == '(') {
        Type *arrayType;
        int dim1, dim2;

        /* check for an array with unspecified dimensions */
        if ((tkn = GetToken(c)) == ')')
            dim1 = dim2 = 0;

        /* otherwise, parse the dimensions */
        else {
            PVAL pv;

            /* put back the token */
            SaveToken(c, tkn);

            /* get the array size */
            ParseExpr(c, &pv);

            /* make sure it's a constant */
            if (!IsLiteral(&pv) || pv.val <= 0)
                ParseError(c, "expecting a positive constant expression");
            dim1 = pv.val;

            /* check for a second dimension */
            if ((tkn = GetToken(c)) == ',') {
                ParseExpr(c, &pv);
                if (!IsLiteral(&pv) || pv.val <= 0)
                    ParseError(c, "expecting a positive constant expression");
                dim2 = pv.val;
                tkn = GetToken(c);
            }

            /* no second dimension */
            else
                dim2 = 1;

            /* only two dimensions allowed for now */
            Require(c, tkn, ')');
        }

        /* create a type for the function */
        if (!(arrayType = (Type *)malloc(sizeof(Type))))
            Fatal("insufficient memory");

        /* initialize the type */
        arrayType->id = TYPE_ARRAY;
        arrayType->ArrayInfo.elementType = type;
        pBaseType = &arrayType->ArrayInfo.elementType;
        arrayType->ArrayInfo.dim1 = dim1;
        arrayType->ArrayInfo.dim2 = dim2;
        type = arrayType;

        tkn = GetToken(c);
    }

    /* check for a type specification */
    if (tkn == T_AS) {
        Type *declaredType;
        FRequire(c, T_IDENTIFIER);
        if (!(declaredType = FindType(&c->types, c->token)))
            ParseError(c, "unknown type: %s", c->token);
        *pBaseType = declaredType;
    }

    /* just use the default type */
    else
        SaveToken(c, tkn);

    /* return the type */
    return type;
}

/* ParseLet - parse the 'LET' statement */
static void ParseLet(ParseContext *c)
{
    PVAL pv, pv2;
    ParsePrimary(c, &pv);
    chklvalue(c, &pv);
    FRequire(c, '=');
    ParseExpr(c, &pv2);
    rvalue(c, &pv2);
    (*pv.fcn)(c, PV_STORE, &pv);
    FRequire(c, T_EOL);
}

/* ParseIf - parse the 'IF' statement */
static void ParseIf(ParseContext *c)
{
    PVAL pv;
    ParseExpr(c, &pv);
    rvalue(c, &pv);
    FRequire(c, T_THEN);
    PushBlock(c);
    c->bptr->type = BLOCK_IF;
    putcbyte(c, OP_BRF);
    c->bptr->IfBlock.nxt = putcword(c, 0);
    c->bptr->IfBlock.end = 0;
    FRequire(c, T_EOL);
}

/* ParseElseIf - parse the 'ELSE IF' statement */
static void ParseElseIf(ParseContext *c)
{
    PVAL pv;
    switch (CheckBlock(c)) {
    case BLOCK_IF:
        putcbyte(c, OP_BR);
        c->bptr->IfBlock.end = putcword(c, c->bptr->IfBlock.end);
        fixupbranch(c, c->bptr->IfBlock.nxt, codeaddr(c));
        ParseExpr(c, &pv);
        rvalue(c, &pv);
        FRequire(c, T_THEN);
        putcbyte(c, OP_BRF);
        c->bptr->IfBlock.nxt = putcword(c, 0);
        FRequire(c, T_EOL);
        break;
    default:
        ParseError(c, "ELSE IF without a matching IF");
        break;
    }
}

/* ParseElse - parse the 'ELSE' statement */
static void ParseElse(ParseContext *c)
{
    int end;
    switch (CheckBlock(c)) {
    case BLOCK_IF:
        putcbyte(c, OP_BR);
        end = putcword(c, c->bptr->IfBlock.end);
        fixupbranch(c, c->bptr->IfBlock.nxt, codeaddr(c));
        c->bptr->type = BLOCK_ELSE;
        c->bptr->ElseBlock.end = end;
        break;
    default:
        ParseError(c, "ELSE without a matching IF");
        break;
    }
    FRequire(c, T_EOL);
}

/* ParseEndIf - parse the 'END IF' statement */
static void ParseEndIf(ParseContext *c)
{
    switch (CheckBlock(c)) {
    case BLOCK_IF:
        fixupbranch(c, c->bptr->IfBlock.nxt, codeaddr(c));
        PopBlock(c);
        break;
    case BLOCK_ELSE:
        fixupbranch(c, c->bptr->ElseBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    default:
        ParseError(c, "ELSE without a matching IF");
        break;
    }
    FRequire(c, T_EOL);
}

/* ParseEnd - parse the 'END' statement */
static void ParseEnd(ParseContext *c)
{
    putcbyte(c, OP_HALT);
    FRequire(c, T_EOL);
}

/* ParseFor - parse the 'FOR' statement */
static void ParseFor(ParseContext *c)
{
    int test, body, inst;
    PVAL pv, pv2;
    Token tkn;

    PushBlock(c);
    c->bptr->type = BLOCK_FOR;

    /* initialize the for variable */
    FRequire(c, T_IDENTIFIER);
    GetLValue(c, c->token, &pv);
    FRequire(c, '=');
    ParseExpr(c, &pv2);
    rvalue(c, &pv2);

    /* check to see if we've reached the end */
    test = codeaddr(c);
    (*pv.fcn)(c, PV_STORE, &pv);
    (*pv.fcn)(c, PV_LOAD, &pv);
    FRequire(c, T_TO);
    ParseExpr(c, &pv2);
    rvalue(c, &pv2);
    putcbyte(c, OP_LE);
    putcbyte(c, OP_BRT);
    body = putcword(c, 0);
    putcbyte(c, OP_BR);
    c->bptr->ForBlock.end = putcword(c, 0);
    
    /* update the for variable after an iteration of the loop */
    c->bptr->ForBlock.nxt = codeaddr(c);
    (*pv.fcn)(c, PV_LOAD, &pv);
    if ((tkn = GetToken(c)) == T_STEP) {
        ParseExpr(c, &pv2);
        rvalue(c, &pv2);
        tkn = GetToken(c);
    }
    else {
        putcbyte(c, OP_LIT);
        putcword(c, 1);
    }
    putcbyte(c, OP_ADD);
    inst = putcbyte(c, OP_BR);
    putcword(c, test - inst - 3);

    fixupbranch(c, body, codeaddr(c));
    Require(c, tkn, T_EOL);
}

/* ParseNext - parse the 'NEXT' statement */
static void ParseNext(ParseContext *c)
{
    int inst;
    switch (CheckBlock(c)) {
    case BLOCK_FOR:
        FRequire(c, T_IDENTIFIER);
        inst = putcbyte(c, OP_BR);
        putcword(c, c->bptr->ForBlock.nxt - inst - 3);
        fixupbranch(c, c->bptr->ForBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    default:
        ParseError(c, "NEXT without a matching FOR");
        break;
    }
    FRequire(c, T_EOL);
}

/* ParseDo - parse the 'DO' statement */
static void ParseDo(ParseContext *c)
{
    PushBlock(c);
    c->bptr->type = BLOCK_DO;
    c->bptr->DoBlock.nxt = codeaddr(c);
    c->bptr->DoBlock.end = 0;
    FRequire(c, T_EOL);
}

/* ParseDoWhile - parse the 'DO WHILE' statement */
static void ParseDoWhile(ParseContext *c)
{
    PVAL pv;
    PushBlock(c);
    c->bptr->type = BLOCK_DO;
    c->bptr->DoBlock.nxt = codeaddr(c);
    ParseExpr(c, &pv);
    rvalue(c, &pv);
    putcbyte(c, OP_BRF);
    c->bptr->DoBlock.end = putcword(c, 0);
    FRequire(c, T_EOL);
}

/* ParseDoUntil - parse the 'DO UNTIL' statement */
static void ParseDoUntil(ParseContext *c)
{
    PVAL pv;
    PushBlock(c);
    c->bptr->type = BLOCK_DO;
    c->bptr->DoBlock.nxt = codeaddr(c);
    ParseExpr(c, &pv);
    rvalue(c, &pv);
    putcbyte(c, OP_BRT);
    c->bptr->DoBlock.end = putcword(c, 0);
    FRequire(c, T_EOL);
}

/* ParseLoop - parse the 'LOOP' statement */
static void ParseLoop(ParseContext *c)
{
    int inst;
    switch (CheckBlock(c)) {
    case BLOCK_DO:
        inst = putcbyte(c, OP_BR);
        putcword(c, c->bptr->DoBlock.nxt - inst - 3);
        fixupbranch(c, c->bptr->DoBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    default:
        ParseError(c, "LOOP without a matching DO");
        break;
    }
    FRequire(c, T_EOL);
}

/* ParseLoopWhile - parse the 'LOOP WHILE' statement */
static void ParseLoopWhile(ParseContext *c)
{
    int inst;
    PVAL pv;
    switch (CheckBlock(c)) {
    case BLOCK_DO:
        ParseExpr(c, &pv);
        rvalue(c, &pv);
        inst = putcbyte(c, OP_BRT);
        putcword(c, c->bptr->DoBlock.nxt - inst - 3);
        fixupbranch(c, c->bptr->DoBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    default:
        ParseError(c, "LOOP without a matching DO");
        break;
    }
    FRequire(c, T_EOL);
}

/* ParseLoopUntil - parse the 'LOOP UNTIL' statement */
static void ParseLoopUntil(ParseContext *c)
{
    int inst;
    PVAL pv;
    switch (CheckBlock(c)) {
    case BLOCK_DO:
        ParseExpr(c, &pv);
        rvalue(c, &pv);
        inst = putcbyte(c, OP_BRF);
        putcword(c, c->bptr->DoBlock.nxt - inst - 3);
        fixupbranch(c, c->bptr->DoBlock.end, codeaddr(c));
        PopBlock(c);
        break;
    default:
        ParseError(c, "LOOP without a matching DO");
        break;
    }
    FRequire(c, T_EOL);
}

/* ParseCall - parse the 'CALL' statement */
static void ParseCall(ParseContext *c)
{
    ParseSubroutineCall(c);
    FRequire(c, T_EOL);
}

/* ParseStop - parse the 'STOP' statement */
static void ParseStop(ParseContext *c)
{
    putcbyte(c, OP_HALT);
    FRequire(c, T_EOL);
}

/* ParsePrint - parse the 'PRINT' statement */
static void ParsePrint(ParseContext *c)
{
    int needNewline = TRUE;
    Token tkn;
    PVAL pv;

    while ((tkn = GetToken(c)) != T_EOL) {
        switch (tkn) {
        case ',':
            needNewline = FALSE;
            putcbyte(c, OP_PRINTT);
            break;
        case ';':
            needNewline = FALSE;
            break;
        default:
            needNewline = TRUE;
            SaveToken(c, tkn);
            ParseExpr(c, &pv);
            rvalue(c, &pv);
            switch (pv.type->id) {
            case TYPE_STRING:
                putcbyte(c, OP_PRINTS);
                break;
            case TYPE_FLOAT:
                putcbyte(c, OP_PRINTF);
                break;
            case TYPE_INTEGER:
            case TYPE_BYTE:
                putcbyte(c, OP_PRINTI);
                break;
            default:
                ParseError(c, "invalid argument to print");
                break;
            }
            break;
        }
    }

    if (needNewline)
        putcbyte(c, OP_PRINTN);
}

/* ParseInput - parse the 'INPUT' statement */
static void ParseInput(ParseContext *c)
{
}

/* CheckBlock - make sure there is a block on the stack */
static BlockType CheckBlock(ParseContext *c)
{
    return c->bptr < c->blockBuf ? BLOCK_NONE : c->bptr->type;
}

/* PushBlock - push a block on the block stack */
static void PushBlock(ParseContext *c)
{
    if (++c->bptr >= c->btop)
        Fatal("statements too deeply nested");
}

/* PopBlock - pop a block off the block stack */
static void PopBlock(ParseContext *c)
{
    --c->bptr;
}
