/* db_life.c - the game of life
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

/* needs to be built with:

XGS_PIC_GFX_DRV_V010.c
XGS_PIC_NTSC_160_192_2_V010.s
XGS_PIC_GAMEPAD_DRV_V010.c
XGS_PIC_SYSTEM_V010.c

*/

#include "XGS_PIC_SYSTEM_V010.h"			
#include "XGS_PIC_NTSC_160_192_2_V010.h"
#include "XGS_PIC_GFX_DRV_V010.h"
#include "XGS_PIC_GAMEPAD_DRV_V010.h"
#include "db_text_v010.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#define WORLDW		16
#define WORLDH		16

#define CELLW		4
#define CELLH		8

#define TOPMARGIN	20
#define LEFTMARGIN	40

#define ARRAYW		(1 + WORLDW + 1 + WORLDW + 1)
#define ARRAYH		(1 + WORLDH + 1)

#define WORLD1		1
#define WORLD2		(1 + WORLDW + 1)

static short world[ARRAYW][ARRAYH];

static int thisx, nextx;

// Not defined but actually located in the gcc library!
int vsprintf( char *, const char *, va_list );
int vsnprintf(char *, size_t, const char *, va_list);

unsigned char	g_VRAMBuffer[VRAM_BUF_SIZE] __attribute__((far));	
unsigned short 	g_PaletteMap[PALETTE_MAP_SIZE];
unsigned char	g_Palettes[MIN_PALETTE_SIZE];

static int visit(int x, int y);
static void clearall(void);
static void fill(void);
static void nextgen(void);
static void showgen(void);
static void clear(int color);
static void fillrect(int x, int y, int w, int h, int color);
static void waitforvsync(void);
static int gamepad(int which);
static void text(int x, int y, char *str);

int main(void)
{
	SYS_ConfigureClock(FCY_NTSC);

	Gamepad_Init();

	GFX_InitBitmap(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP, g_VRAMBuffer, g_Palettes, g_PaletteMap);
	GFX_SetClipRegion(0, SCREEN_HEIGHT - 1, 0, SCREEN_WIDTH - 1);
	
	memset(g_PaletteMap, 0, sizeof(g_PaletteMap));
	
	g_Palettes[0] = NTSC_BLACK;
	g_Palettes[1] = NTSC_RED;
	g_Palettes[2] = NTSC_GREEN;
	g_Palettes[3] = NTSC_BLUE;

	GFX_StartDrawing(SCREEN_TYPE_NTSC);

	clear(3);
	text(10, 5, "The Game of Life");
	fillrect(LEFTMARGIN - 1, TOPMARGIN - 1, WORLDW * (CELLW + 1) + 1, WORLDH * (CELLH + 1) + 1, 0);

	clearall();

	for (;;) {
		int generations = 30;
		int generation = 0;
		int done;

		thisx = WORLD1;
		nextx = WORLD2;
		fill();
		
		do {
			int tmp, buttons;

		    // count the generation
		    ++generation;

		    // wait for vsync to display the current generation
		    waitforvsync();
		    showgen();
		    
		    // compute the next generation
		    nextgen();
		    
		    // swap worlds
		    tmp = thisx;
		    thisx = nextx;
		    nextx = tmp;
		
		    // determine if it's time to stop
		    buttons = gamepad(0);
		    done = !(buttons & GAMEPAD_YELLOW) && (generation >= generations);
	
		} while (!done);
	}

	return 0;
}

static int visit(int x, int y)
{
    int neighbors = 0;
    int x1, y1;
    for (y1 = y-1; y1 <= y+1; ++y1)
		for (x1 = x-1; x1 <= x+1; ++x1)
			if (world[x1][y1] > 0)
				++neighbors;
	return neighbors;
}

static void clearall(void)
{
	int x1, y1;
	for (y1 = 0; y1 < ARRAYH; ++y1)
		for (x1 = 0; x1 < ARRAYW; ++x1)
			world[x1][y1] = 0;
}

static void fill(void)
{
	int x1, y1, cell;
	for (y1 = 1; y1 <= WORLDH; ++y1)
		for (x1 = 0; x1 < WORLDW; ++x1) {
			if ((rand() % 4) == 3)
				cell = 1;
			else
				cell = 0;
			world[thisx+x1][y1] = cell;
		}
}

static void nextgen(void)
{
	int x1, y1, neighbors, new;
	for (y1 = 1; y1 <= WORLDH; ++y1)
		for (x1 = 0; x1 < WORLDW; ++x1) {
			neighbors = visit(thisx+x1, y1);
			if (world[thisx+x1][y1] == 0) {
				if (neighbors == 3)
					new = 1;
				else
					new = 0;
			}
			else {
				if (neighbors == 3 || neighbors == 4)
					new = 1;
				else
					new = 0;
			}
			world[nextx+x1][y1] = new;
		}
}

static void showgen(void)
{
	int x1, y1, xoff, yoff, color;
	yoff = TOPMARGIN;
	for (y1 = 1; y1 <= WORLDH; ++y1) {
		xoff = LEFTMARGIN;
		for (x1 = 0; x1 < WORLDW; ++x1) {
			if (world[thisx+x1][y1] == 0)
				color = 0;
			else
				color = 2;
			fillrect(xoff, yoff, CELLW, CELLH, color);
			xoff += CELLW + 1;
		}
		yoff += CELLH + 1;
	}
}

static void clear(int color)
{
    GFX_FillScreen_2BPP(color, g_VRAMBuffer);
}

static void fillrect(int x, int y, int w, int h, int color)
{
	GFX_FillRegion_2BPP(x, y, w, h, color, g_VRAMBuffer);
}

static void waitforvsync(void)
{
	WAIT_FOR_VSYNC_END();	
	WAIT_FOR_VSYNC_START();	
}

static int gamepad(int which)
{
	return Gamepad_Read(which == 0 ? GAMEPAD_0 : GAMEPAD_1);
}

static void text(int x, int y, char *str)
{
	DrawBitmapFontString(str, x, y, g_VRAMBuffer);
}
