/* db_packet_v010.c - packet transfer routines
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "db_packet_v010.h"
#include "../runtime/db_vm_v010.h"

typedef enum {
    RCV_WAIT_SOH,
    RCV_PKTN,
    RCV_TYPE,
    RCV_LEN,
    RCV_CHK,
    RCV_DATA,
    RCV_CRC1,
    RCV_CRC2,
    RCV_PURGE
} RcvState;

#define RETRY_INFINITE	(-1)

typedef struct {
    RcvState state;
    int retry;
    int cancel;
    int timeout;
    uint8_t rcv_packet[PKTMAXLEN];
    uint8_t rcv_nextPktn;
    uint8_t rcv_pktn;
    uint8_t rcv_type;
    uint8_t rcv_len;
    uint8_t rcv_chk;
    uint16_t rcv_crc;
    int rcv_dataCount;
    uint8_t snd_frame[FRAMELEN];
    uint8_t snd_pktn;
    uint8_t snd_len;
    VMFILE *fp;
} PktState;

static PktState pkt;

static void Init(void);
static int CheckHeader(void);
static int HandlePacket(int *pType);
static void ProcessDebug(void);
static void ProcessOpen(void);
static void ProcessData(void);
static void ProcessClose(void);
static void ProcessRun(void);

int PktInit(char *port)
{
    if (!md_init(port))
        return VMFALSE;
    Init();
    pkt.fp = NULL;
    return VMTRUE;
}

void PktSendFrame(int type, uint8_t *buf, int len)
{
    PktBuildFrame(type, buf, len);
    PktSendCurrentFrame();
}

void PktBuildFrame(int type, uint8_t *buf, int len)
{
    uint16_t crc = 0;
    uint8_t *p;

    /* setup the frame header */
    p = pkt.snd_frame;
    p[0] = SOH;					/* SOH */
    p[1] = pkt.snd_pktn;		/* packet# */
    p[2] = type;				/* type type */
    p[3] = len;					/* data length */
    p[4] = p[1] + p[2] + p[3];	/* header checksum */
    p += 5;

    /* compute the crc */
    while (--len >= 0)
        crc = updcrc(crc, (*p++ = *buf++));
    crc = updcrc(crc, '\0');
    crc = updcrc(crc, '\0');

    /* add the crc to the frame */
    *p++ = (uint8_t)(crc >> 8);
    *p++ = (uint8_t)crc;

    /* save the frame length */
    pkt.snd_len = (uint8_t)(p - pkt.snd_frame);

    /* increment the packet number */
    if (type >= PKT_BASE)
        ++pkt.snd_pktn;
}

void PktSendCurrentFrame(void)
{
    md_write(pkt.snd_frame, pkt.snd_len);
}

int PktPoll(int timeout)
{
    int ch, type;

    /* handle received characters */
    if ((ch = md_get(timeout)) != MD_TIMEOUT) {
        do {
            switch (pkt.state) {
            case RCV_WAIT_SOH:
                switch (ch) {
                case SOH:
                    pkt.timeout = 10;
                    pkt.state = RCV_PKTN;
                    break;
                case CAN:
                    if (++pkt.cancel >= 2) {
                        ProcessClose();
                        Init();
                    }
                    break;
                }
                break;
            case RCV_PKTN:
                pkt.rcv_pktn = ch;
                pkt.state = RCV_TYPE;
                break;
            case RCV_TYPE:
                pkt.rcv_type = ch;
                pkt.state = RCV_LEN;
                break;
            case RCV_LEN:
                pkt.rcv_len = ch;
                pkt.state = RCV_CHK;
                break;
             case RCV_CHK:
                pkt.rcv_chk = ch;
                if (CheckHeader()) {
                    if (pkt.rcv_len == 0)
                        pkt.state = RCV_CRC1;
                    else {
                        pkt.rcv_dataCount = 0;
                        pkt.rcv_crc = 0;
                        pkt.state = RCV_DATA;
                    }
                }
                else {
                    /* handle bad header */
                }
                break;
           case RCV_DATA:
                pkt.rcv_packet[pkt.rcv_dataCount] = ch;
                pkt.rcv_crc = updcrc(pkt.rcv_crc, ch);
                if (++pkt.rcv_dataCount >= pkt.rcv_len)
                    pkt.state = RCV_CRC1;
                break;
            case RCV_CRC1:
                pkt.rcv_crc = updcrc(pkt.rcv_crc, ch);
                pkt.state = RCV_CRC2;
                break;
            case RCV_CRC2:
                pkt.rcv_crc = updcrc(pkt.rcv_crc, ch);
                pkt.state = RCV_WAIT_SOH;
                if (HandlePacket(&type)) {
                    if (type >= PKT_BASE)
                        PktSendFrame(PKT_ACK, NULL, 0);
                }
                else
                    PktSendFrame(PKT_NAK, NULL, 0);
                if (type == PKT_ACK || type == PKT_NAK)
                    return type;
                break;
            case RCV_PURGE:
                break;
            default:
                pkt.retry = 0;
                pkt.state = RCV_PURGE;
                break;
            }
        } while ((ch = md_get(timeout)) != MD_TIMEOUT);
    }

    return PKT_TIMEOUT;
}

static void Init(void)
{
    pkt.state = RCV_WAIT_SOH;
    pkt.snd_pktn = 1;
    pkt.rcv_nextPktn = 1;
    pkt.retry = RETRY_INFINITE;
    pkt.cancel = 0;
    pkt.timeout = 10;
}

static int CheckHeader(void)
{
    /* check the header checksum */
    if (pkt.rcv_pktn + pkt.rcv_type + pkt.rcv_len != pkt.rcv_chk)
        return VMFALSE;

    /* header is okay */
    return VMTRUE;
}

static int HandlePacket(int *pType)
{
    /* check for a duplicate block */
    if (pkt.rcv_pktn == ((pkt.rcv_nextPktn - 1) & 0xff))
        return VMTRUE;

    /* check for a block sequence error */
    else if (pkt.rcv_pktn != pkt.rcv_nextPktn)
        return VMFALSE;

    /* check the packet length */
    else if (pkt.rcv_len > PKTMAXLEN)
        return VMFALSE;

    /* update the next packet number */
    if (pkt.rcv_type >= PKT_BASE)
        ++pkt.rcv_nextPktn;

    /* handle the packet */
    switch (pkt.rcv_type) {
    case PKT_DEBUG:
        ProcessDebug();
        break;
    case PKT_ACK:
        break;
    case PKT_NAK:
        break;
    case PKT_OPEN:
        ProcessOpen();
        break;
    case PKT_DATA:
        ProcessData();
        break;
    case PKT_CLOSE:
        ProcessClose();
        break;
    case PKT_RUN:
        ProcessRun();
        break;
    default:
         break;
    }

    /* packet was processed */
    *pType = pkt.rcv_type;
    return VMTRUE;
}

static void ProcessDebug(void)
{
#ifdef WIN32
    char *str;
    
    /* get the debug string */
    str = (char *)pkt.rcv_packet;

    /* display it */
    fwrite(pkt.rcv_packet, pkt.rcv_len, 1, stderr);
#endif
}

static void ProcessOpen(void)
{
    char *name;
    
    /* close any currently open file */
    ProcessClose();
    
    /* get the filename */
    name = (char *)pkt.rcv_packet;

    /* check for the start of a new transfer */
    pkt.fp = VM_fopen(name, "wb");
}

static void ProcessData(void)
{
    if (pkt.fp)
        VM_fwrite(pkt.rcv_packet, 1, pkt.rcv_len, pkt.fp);
}

static void ProcessClose(void)
{
    if (pkt.fp) {
        VM_fclose(pkt.fp);
        pkt.fp = NULL;
    }
}

static void ProcessRun(void)
{
#ifndef WIN32
    char *name;
    
    /* get the filename */
    name = (char *)pkt.rcv_packet;

    /* load a new program */
    LoadProgram(name);
#endif
}
