/* db_vmmain.c - a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

/* needs to be built with:

db_crctab_v010.c
db_packet_v010.c
db_picuart_v010.c
db_vmdebug_v010.c
db_vmfcn_v010.c
db_vmimage_v010.c
db_vmint_v010.c
db_vmmain_v010.c
FSIO.c
SD-SPI.c
XGS_PIC_GAMEPAD_DRV_V010.c
XGS_PIC_GFX_DRV_V010.c
XGS_PIC_SOUND_DRV_V010.c
XGS_PIC_SYSTEM_V010.c
XGS_PIC_UART_DRV_V010.c

For the bitmap target define the symbol BITMAP_GRAPHICS and include:

db_text_v010.c
XGS_PIC_NTSC_160_192_2_V010.s

For the tile target define the symbol TILE_GRAPHICS and include:

XGS_PIC_NTSC_TILE2_V010.s

In order to make serial downloads reliable, it is necessary to change the following definition in XGS_PIC_UART_DRV_V010.h:

  #define UART1_RX_BUFFER_SIZE	135

The receive buffer needs to be large enough to hold a 5+128+2=135 byte packet.

In MPLAB under Project -> Build Options -> Project
and then under the "MPLAB LINK30" tab, set the heap size to 512. 

*/

#include "XGS_PIC_SYSTEM_V010.h"			
#include "XGS_PIC_GFX_DRV_V010.h"
#include "XGS_PIC_UART_DRV_V010.h"
#include "XGS_PIC_GAMEPAD_DRV_V010.h"
#include "XGS_PIC_SOUND_DRV_V010.h"

#ifdef BITMAP_GRAPHICS
#include "XGS_PIC_NTSC_160_192_2_V010.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "db_vm_v010.h"
#include "db_text_v010.h"
#include "db_packet_v010.h"

#define VMIMAGE_SIZE    2000

static uint8_t vmimage[VMIMAGE_SIZE] __attribute__((far));	

// Not defined but actually located in the gcc library!
int vsprintf( char *, const char *, va_list );
int vsnprintf(char *, size_t, const char *, va_list);

#ifdef BITMAP_GRAPHICS
unsigned char	g_VRAMBuffer[VRAM_BUF_SIZE] __attribute__((far));	
unsigned short 	g_PaletteMap[PALETTE_MAP_SIZE];
unsigned char	g_Palettes[MIN_PALETTE_SIZE];
#define TARGET_TYPE TARGET_BITMAP
#endif

#ifdef TILE_GRAPHICS
#define TARGET_TYPE TARGET_TILE
#endif

static VMFILE *FileOpen(char *name, char *mode);
static void FileClose(VMFILE *fp);
static size_t FileRead(void *buf, size_t count, VMFILE *fp);
static size_t FileWrite(void *buf, size_t count, VMFILE *fp);
static int FileSeek(VMFILE *fp, long offset, int whence);
static long FileTell(VMFILE *fp);
static void Poll(void);

int main(void)
{
    int imageLoaded;
    Interpreter i;

    SYS_ConfigureClock(FCY_NTSC);

    UART_Init(9600);
    UART_printf("XGS Basic Runtime 1.0\r\n");

    PktInit(NULL);

    Gamepad_Init();
    SND_Init();
    FSInit();

#ifdef BITMAP_GRAPHICS
    GFX_InitBitmap(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP, g_VRAMBuffer, g_Palettes, g_PaletteMap);
    GFX_SetClipRegion(0, SCREEN_HEIGHT - 1, 0, SCREEN_WIDTH - 1);
    
    memset(g_PaletteMap, 0, sizeof(g_PaletteMap));
    
    g_Palettes[0] = NTSC_BLACK;
    g_Palettes[1] = NTSC_RED;
    g_Palettes[2] = NTSC_GREEN;
    g_Palettes[3] = NTSC_BLUE;

    GFX_StartDrawing(SCREEN_TYPE_NTSC);
#endif

    i.fileOpen = FileOpen;
    i.fileClose = FileClose;
    i.fileRead = FileRead;
    i.fileWrite = FileWrite;
    i.fileSeek = FileSeek;
    i.fileTell = FileTell;
    i.poll = Poll;

    InitFcn(&i);

    InitInterpreter(&i, TARGET_TYPE);

    if (!(imageLoaded = LoadImage(&i, "autorun.bai"))) {
#ifdef BITMAP_GRAPHICS
        DrawBitmapFontString("XGS Basic Runtime 1.0", 0, 20, g_VRAMBuffer);
        DrawBitmapFontString("load failed: autorun.bai", 0, 40, g_VRAMBuffer);
#else
        VM_printf("load failed: autorun.bai\n");
#endif
    }   

    for (;;) {
        if (imageLoaded)
            Execute(&i);
        else
            PktPoll(0);
    }

    return 0;
}
/* Poll - check for incoming packets */
static void Poll(void)
{
    PktPoll(0);
}

/* AllocateImage - allocate space for an image */
int AllocateImage(Interpreter *i, int codeSize, int dataSize)
{
    /* make sure there is enough memory */
    if (codeSize + dataSize > VMIMAGE_SIZE)
        return VMFALSE;

    /* setup the code and data space (code last in case it is an odd number of bytes) */
    i->variables = (int16_t *)vmimage;
    i->code = vmimage + dataSize;

    /* return successfully */
    return VMTRUE;
}

/* FreeImage - free an image */
void FreeImage(Interpreter *i)
{
}

void VM_printf(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_vprintf(fmt, ap);
    va_end(ap);
}

void VM_vprintf(char *fmt, va_list ap)
{
    char buf[128];
    vsnprintf(buf, sizeof(buf), fmt, ap);
    VM_puts(buf);
}

void VM_puts(char *str)
{
    while (*str)
        VM_putchar(*str++);
    VM_flush();
}

static char debugBuf[16];
static int debugLen = 0;

static void DebugBufferCh(int ch)
{
    if (debugLen < sizeof(debugBuf)) {
        debugBuf[debugLen++] = ch;
        if (debugLen == sizeof(debugBuf))
            VM_flush();
    }
}

void VM_putchar(int ch)
{
    if (ch == '\n')
        DebugBufferCh('\r');
    DebugBufferCh(ch);
    if (ch == '\n')
        VM_flush();
}

void VM_flush(void)
{
    if (debugLen > 0) {
        PktSendFrame(PKT_DEBUG, (uint8_t *)debugBuf, debugLen);
        debugLen = 0;
    }
}

static VMFILE *FileOpen(char *name, char *mode)
{
    return (VMFILE *)FSfopen(name, mode);
}

static void FileClose(VMFILE *fp)
{
    FSfclose((FSFILE *)fp);
}

static size_t FileRead(void *buf, size_t count, VMFILE *fp)
{
    return FSfread(buf, 1, count, (FSFILE *)fp);
}

static size_t FileWrite(void *buf, size_t count, VMFILE *fp)
{
    return FSfwrite(buf, 1, count, (FSFILE *)fp);
}

static int FileSeek(VMFILE *fp, long offset, int whence)
{
    return FSfseek((FSFILE *)fp, offset, whence);
}

static long FileTell(VMFILE *fp)
{
    return FSftell((FSFILE *)fp);
}
