/* db_bascom.c - a simple basic compiler
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "db_bascom.h"
#include "../common/db_packet.h"

/* local function prototypes */
static Usage(void);
static void DownloadImage(char *imagename, char *dstname, char *port);
static void AddIntrinsic(ParseContext *c, char *name, char *argtypes, char *retType, int index);
static void PlaceStrings(ParseContext *c, int offset);
static void WriteStrings(ParseContext *c);
static void CreateOutputFile(ParseContext *c, char *name);
static void WriteOutputFile(ParseContext *c);

/* main - the main routine */
int main(int argc, char *argv[])
{
    char *infile, *outfile, *dstname, *port;
    int compileOnly, debugMode, i;
    ParseContext c;

    /* initialize */
    infile = NULL;
    outfile = NULL;
	dstname = "autorun.bai";
    port = "COM4";
    compileOnly = VMFALSE;
    debugMode = VMFALSE;

    /* get the arguments */
    for(i = 1; i < argc; ++i) {

        /* handle switches */
        if(argv[i][0] == '-') {
            switch(argv[i][1]) {
            case 'c':
                compileOnly = VMTRUE;
                break;
            case 'd':
                debugMode = VMTRUE;
                break;
            case 'o':
                if(argv[i][2])
                    outfile = &argv[i][2];
                else if(++i < argc)
                    outfile = argv[i];
                else
                    Usage();
                break;
            case 'p':
                if(argv[i][2])
                    port = &argv[i][2];
                else if(++i < argc)
                    port = argv[i];
                else
                    Usage();
                break;
            case 'r':
                dstname = NULL;
                break;
            default:
                Usage();
            }
        }

        /* handle the input filename */
        else {
            if (infile)
                Usage();
            infile = argv[i];
        }
    }

    /* make sure there was at least an input file specified */
    if (!infile)
        Usage();

    /* default the output file name */
    if (!outfile) {
        static char buf[MAXLINE];
        char *end = strrchr(infile, '.');
        if (end && !strchr(end, '/') && !strchr(end, '\\')) {
            strncpy(buf, infile, end - infile);
            buf[end - infile] = '\0';
        }
        else
            strcpy(buf, infile);
        strcat(buf, ".bai");
        outfile = buf;
    }

    /* initialize the parse context */
    InitParseContext(&c);

    /* create the output file */
    CreateOutputFile(&c, outfile);

    /* initialize the global symbol table */
    InitSymbolTable(&c.globals);
    c.globalCount = 0;

    /* target is unknown */
    c.target = TARGET_UNKNOWN;

    /* remember whether we're in debug mode */
    c.debugMode = debugMode;

    /* initialize the type table */
    InitTypeTable(&c.types);
    AddType(&c.types, "FLOAT",      &c.floatType);
    AddType(&c.types, "INTEGER",    &c.integerType);
    AddType(&c.types, "BYTE",       &c.byteType);
    AddType(&c.types, "STRING",     &c.stringType);
    
    /* add the intrinsic functions */
    AddIntrinsic(&c, "RND",             "i",        "i",    0);
    AddIntrinsic(&c, "WAITFORVSYNC",    "",         NULL,   1);
    AddIntrinsic(&c, "GAMEPAD",         "i",        "i",    2);
    AddIntrinsic(&c, "SOUND",           "i",        NULL,   3);
    
    /* open the input file */
    if (!PushFile(&c, infile))
        Fatal("opening '%s'", infile);

    /* get the next line */
    while (GetLine(&c)) {
        Token tkn;
        if ((tkn = GetToken(&c)) != T_EOL) {
            SaveToken(&c, tkn);
            ParseStatement(&c);
        }
    }

    /* write the output file */
    WriteOutputFile(&c);

    /* close the parse context */
    CloseParseContext(&c);

    /* make sure a target was defined */
    if (c.target == TARGET_UNKNOWN)
        Fatal("target unknown, please add an OPTION target = [ tile | bitmap ] to your source file");

    /* download the file if necessary */
    if (!compileOnly)
        DownloadImage(outfile, dstname, port);

    /* exit successfully */
    return 0;
}

/* Usage - display a usage message and exit */
static Usage(void)
{
    fprintf(stderr, "\
usage: xgsbasic [ -o <output-file> ] [ -p <port> ] [ -c ] [ -r ] <input-file>\n");
    exit(1);
}

/* DownloadImage - download a compiled image */
static void DownloadImage(char *imagename, char *dstname, char *port)
{
    /* open the communications line */
    if (!PktOpen(port))
        Fatal("can't open serial port: %s", port);

    /* initialize the link */
    if (!PktInit())
        Fatal("can't initialize link");

    if (!SendFile(imagename, dstname))
        Fatal("download failed: %s", imagename);

    if (dstname && !RunFile("autorun.bai"))
        Fatal("error starting program");
    
	for (;;)
        PktPoll(0);

    /* restore the communications line */
    md_done();
}

/* SelectRuntime - select the target runtime */
void SelectRuntime(ParseContext *c, char *target)
{
    if (strcasecmp(target, "bitmap") == 0) {
        AddIntrinsic(c, "PLOT",            "iii",      NULL,   4);
        AddIntrinsic(c, "LINE",            "iiiii",    NULL,   5);
        AddIntrinsic(c, "HLINE",           "iiii",     NULL,   6);
        AddIntrinsic(c, "VLINE",           "iiii",     NULL,   7);
        AddIntrinsic(c, "CLEAR",           "i",        NULL,   8);
        AddIntrinsic(c, "FILLRECT",        "iiiii",    NULL,   9);
        AddIntrinsic(c, "CIRCLE",          "iiii",     NULL,   10);
        AddIntrinsic(c, "TEXT",            "iis",      NULL,   11);
        c->target = TARGET_BITMAP;
    }
    else if (strcasecmp(target, "tile") == 0) {
        AddIntrinsic(c, "SETTILEMAP",      "B",        NULL,   4);
        AddIntrinsic(c, "SETTILES",        "B",        NULL,   5);
        AddIntrinsic(c, "SETCOLORTABLE",   "B",        NULL,   6);
        AddIntrinsic(c, "SETSPRITE",       "iiii",     NULL,   7);
        c->target = TARGET_TILE;
    }
    else
        ParseError(c, "unknown runtime: %s", target);
}

/* AddString - add a string to the string table */
int AddString(ParseContext *c, char *value, int offset)
{
    String *str;
    
    /* check to see if the string is already in the table */
    for (str = c->strings; str != NULL; str = str->next)
        if (strcmp(value, str->value) == 0) {
            int link;
            if (!(link = str->fixups))
                return str->offset;
            else {
                str->fixups = offset;
                return link;
            }
        }

    /* allocate the string structure */
    if (!(str = (String *)malloc(sizeof(String) + strlen(value))))
        Fatal("insufficient memory");

    /* initialize the string */
    memset(str, 0, sizeof(String));
    strcpy(str->value, value);
    str->fixups = offset;
    str->next = c->strings;
    c->strings = str;

    /* return zero to terminate the fixup list */
    return 0;
}

/* PlaceStrings - place any string literals defined in the current function */
static void PlaceStrings(ParseContext *c, int offset)
{
    String *str;
    for (str = c->strings; str != NULL; str = str->next) {
        if (str->fixups) {
            int size = (int)strlen(str->value) + 1;
            str->offset = offset;
            fixup(c, str->fixups, str->offset);
            offset += size;
        }
    }
}

/* WriteStrings - write any string literals defined in the current function */
static void WriteStrings(ParseContext *c)
{
    String *str;
    for (str = c->strings; str != NULL; str = str->next) {
        if (str->fixups) {
            int size = (int)strlen(str->value) + 1;
            //printf("string: %04x '%s'\n", str->offset, str->value);
            fwrite(str->value, size, 1, c->ofp);
            c->codeOff += size;
            str->fixups = 0;
        }
    }
}

/* AddIntrinsic - add an intrinsic function to the global symbol table */
static void AddIntrinsic(ParseContext *c, char *name, char *argTypes, char *retType, int index)
{
    Type *type, *argType;

    /* create a type for the function */
    if (!(type = (Type *)malloc(sizeof(Type))))
        Fatal("insufficient memory");

    /* initialize the type */
    type->id = TYPE_FUNCTION;
    InitSymbolTable(&type->FunctionInfo.arguments);

    /* set the return type */
    if (retType) {
        switch (*retType) {
        case 'f':
            type->FunctionInfo.returnType = &c->floatType;
            break;
        case 'i':
            type->FunctionInfo.returnType = &c->integerType;
            break;
        case 'b':
            type->FunctionInfo.returnType = &c->byteType;
            break;
        case 's':
            type->FunctionInfo.returnType = &c->stringType;
            break;
        default:
            Fatal("unknown return type: %s", retType);
            break;
        }
    }
    else
        type->FunctionInfo.returnType = NULL;


    /* add a dummy symbol for the return value */
    if (retType) {
        if (!AddSymbol(&type->FunctionInfo.arguments, name, SC_VARIABLE, NULL, 0, 0))
            Fatal("insufficient memory");
    }

    /* add the argument types */
    if (argTypes) {
        while (*argTypes) {
            switch (*argTypes++) {
            case 'f':
                argType = &c->floatType;
                break;
            case 'F':
                argType = &c->floatArrayType;
                break;
            case 'i':
                argType = &c->integerType;
                break;
            case 'I':
                argType = &c->integerArrayType;
                break;
            case 'b':
                argType = &c->byteType;
                break;
            case 'B':
                argType = &c->byteArrayType;
                break;
            case 's':
                argType = &c->stringType;
                break;
            default:
                Fatal("unknown argument type: %c", argTypes[-1]);
                break;
            }
            if (!AddSymbol(&type->FunctionInfo.arguments, "", SC_VARIABLE, argType, 0, 0))
                Fatal("insufficient memory");
        }
    }

    /* add a global symbol for the intrinsic function */
    if (!AddSymbol(&c->globals, name, SC_CONSTANT, type, 0, (index << 1) | 1))
        Fatal("insufficient memory");
}

/* CreateOutputFile - create the output file */
static void CreateOutputFile(ParseContext *c, char *name)
{
    /* create the output file */
    if (!(c->ofp = fopen(name, "wb")))
        Fatal("creating output file: %s", name);

    /* skip past where the header will be */
    fseek(c->ofp, sizeof(FileHdr), SEEK_SET);
}

/* WriteFunction - write byte for a function to the output file */
void WriteFunction(ParseContext *c, char *name)
{
    size_t codeSize = c->cptr - c->codeBuf;

    /* place string literals defined in this function */
    PlaceStrings(c, (int)c->codeOff + codeSize);
    
    if (c->debugMode) {
        printf("%s:\n", name);
        DecodeFunction(c->codeOff, c->codeBuf, (int)codeSize);
    }

    /* write the procedure to the output file */
    if (fwrite(c->codeBuf, 1, codeSize, c->ofp) != codeSize)
        Fatal("error writing function to output file");

    /* update the offset into the code area */
    c->codeOff += codeSize;

    /* write the string literals defined in this function */
    WriteStrings(c);
    
    /* restore the pointers to the main code */
    c->cptr = c->codeBuf;
}

/* WriteOutputFile - write the output file */
static void WriteOutputFile(ParseContext *c)
{
    uint16_t mainCodeOffset;
    FileHdr hdr;
    Symbol *sym;

    /* end the main code with a halt */
    putcbyte(c, OP_HALT);
    
    /* write the main code */
    mainCodeOffset = c->codeOff;
    WriteFunction(c, "main");

    /* setup the file header structure */
    strcpy(hdr.tag, FILE_TAG);
    hdr.version = FILE_VERSION;
    hdr.target = c->target;
    hdr.codeOffset = sizeof(FileHdr);
    hdr.codeSize = c->codeOff;
    hdr.mainCodeOffset = mainCodeOffset;
    hdr.initOffset = hdr.codeOffset + ((hdr.codeSize + 1) & ~1);
    hdr.initSize = 0;
    hdr.variableCount = c->globalCount;
    hdr.stackSize = 100;

    /* write the initializers */
    fseek(c->ofp, hdr.initOffset, SEEK_SET);
    for (sym = c->globals.head; sym != NULL; sym = sym->next)
        if (sym->storageClass == SC_VARIABLE) {
            if (sym->type->id == TYPE_ARRAY) {
                int arraySize = sym->type->ArrayInfo.dim1 * sym->type->ArrayInfo.dim2;
                uint16_t offset = sym->offset - 2;
                uint16_t count = 2;
                fwrite(&offset, sizeof(uint16_t), 1, c->ofp);
                fwrite(&count, sizeof(uint16_t), 1, c->ofp);
                fwrite(&sym->type->ArrayInfo.dim1, sizeof(int16_t), 1, c->ofp);
                fwrite(&arraySize, sizeof(int16_t), 1, c->ofp);
                hdr.initSize += (2 + count) * sizeof(int16_t);
            }
        }

    /* write the output file header */
    fseek(c->ofp, 0, SEEK_SET);
    if (fwrite(&hdr, 1, sizeof(FileHdr), c->ofp) != sizeof(FileHdr))
        Fatal("error writing output file header");

    /* close the output file */
    fclose(c->ofp);

    if (c->debugMode)
        DumpSymbols(&c->globals, VMTRUE);

    printf("\n");
    printf("code size:  %d\n", hdr.codeSize);
    printf("data size:  %d\n", hdr.variableCount  * sizeof(int16_t));
    printf("stack size: %d\n", hdr.stackSize * sizeof(int16_t));
    printf("total sram: %d\n", hdr.codeSize + (hdr.variableCount + hdr.stackSize) * sizeof(int16_t));
}

void VM_printf(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end(ap);
}

void VM_vprintf(char *fmt, va_list ap)
{
	vprintf(fmt, ap);
}

void VM_putchar(int ch)
{
	putchar(ch);
}

VMFILE *VM_fopen(const char *name, const char *mode)
{
    return (VMFILE *)fopen(name, mode);
}

int VM_fclose(VMFILE *file)
{
    return fclose((FILE *)file);
}

size_t VM_fwrite(void *buf, size_t count, VMFILE *file)
{
    return fwrite(buf, 1, count, (FILE *)file);
}

/* Alloc - allocate memory and report a fatal error on failure */
void *Alloc(size_t size)
{
    void *p;
    if (!(p = malloc(size)))
        Fatal("insufficient memory");
    return p;
}

/* Fatal - report a fatal error and exit */
void Fatal(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    fputs("error: ", stderr);
    vfprintf(stderr, fmt, ap);
    putc('\n', stderr);
    va_end(ap);
    exit(1);
}
