/* db_bascom.c - a simple basic compiler
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "db_bascom_v010.h"
#include "db_xmodem_v010.h"

/* local function prototypes */
static Usage(void);
static void AddIntrinsic(ParseContext *c, char *name, char *argtypes, char *retType, int index);
static void PlaceStrings(ParseContext *c, int offset);
static void WriteStrings(ParseContext *c);
static void CreateOutputFile(ParseContext *c, char *name);
static void WriteOutputFile(ParseContext *c);

/* main - the main routine */
int main(int argc, char *argv[])
{
    char *infile, *outfile, *port, *xgsfile;
    ParseContext c;
    int i;

    /* initialize */
    infile = NULL;
    outfile = NULL;
    port = "COM4";
    xgsfile = NULL;
    
    /* get the arguments */
    for(i = 1; i < argc; ++i) {

        /* handle switches */
        if(argv[i][0] == '-') {
            switch(argv[i][1]) {
            case 'd':
                if(argv[i][2])
                    xgsfile = &argv[i][2];
                else if(++i < argc)
                    xgsfile = argv[i];
                else
                    Usage();
                break;
            case 'o':
                if(argv[i][2])
                    outfile = &argv[i][2];
                else if(++i < argc)
                    outfile = argv[i];
                else
                    Usage();
                break;
            case 'p':
                if(argv[i][2])
                    port = &argv[i][2];
                else if(++i < argc)
                    port = argv[i];
                else
                    Usage();
                break;
            default:
                Usage();
            }
        }

        /* handle the input filename */
        else {
            if (infile)
                Usage();
            infile = argv[i];
        }
    }

    /* make sure there was at least an input file specified */
    if (!infile)
        Usage();

    /* default the output file name */
    if (!outfile) {
        static char buf[MAXLINE];
        char *end = strrchr(infile, '.');
        if (end && !strchr(end, '/') && !strchr(end, '\\')) {
            strncpy(buf, infile, end - infile);
            buf[end - infile] = '\0';
        }
        else
            strcpy(buf, infile);
        strcat(buf, ".bai");
        outfile = buf;
    }

    /* initialize the parse context */
    InitParseContext(&c);

    /* create the output file */
    CreateOutputFile(&c, outfile);

    /* initialize the global symbol table */
    InitSymbolTable(&c.globals);
    c.globalCount = 0;

    /* initialize the type table */
    InitTypeTable(&c.types);
    AddType(&c.types, "FLOAT",      &c.floatType);
    AddType(&c.types, "INTEGER",    &c.integerType);
    AddType(&c.types, "BYTE",       &c.byteType);
    AddType(&c.types, "STRING",     &c.stringType);
    
    /* add the intrinsic functions */
    AddIntrinsic(&c, "RND",             "i",        "i",    0);
    AddIntrinsic(&c, "PLOT",            "iii",      NULL,   1);
    AddIntrinsic(&c, "LINE",            "iiiii",    NULL,   2);
    AddIntrinsic(&c, "HLINE",           "iiii",     NULL,   3);
    AddIntrinsic(&c, "VLINE",           "iiii",     NULL,   4);
    AddIntrinsic(&c, "CLEAR",           "i",        NULL,   5);
    AddIntrinsic(&c, "FILLRECT",        "iiiii",    NULL,   6);
    AddIntrinsic(&c, "CIRCLE",          "iiii",     NULL,   7);
    AddIntrinsic(&c, "WAITFORVSYNC",    "",         NULL,   8);
    AddIntrinsic(&c, "GAMEPAD",         "i",        "i",    9);
    AddIntrinsic(&c, "SOUND",           "i",        NULL,   10);
    AddIntrinsic(&c, "TEXT",			"iis",		NULL,   11);
    AddIntrinsic(&c, "DEBUGOUT",        "s",        NULL,   12);
    
    /* open the input file */
    if (!PushFile(&c, infile))
        Fatal("opening '%s'", infile);

    /* get the next line */
    while (GetLine(&c)) {
        Token tkn;
        if ((tkn = GetToken(&c)) != T_EOL) {
            SaveToken(&c, tkn);
            ParseStatement(&c);
        }
    }

    /* write the output file */
    WriteOutputFile(&c);

    /* close the parse context */
    CloseParseContext(&c);

    /* download the file to the xgs */
    if (xgsfile) {
#ifdef WIN32
		char fullPort[20];
		sprintf(fullPort, "\\\\.\\%s", port);
		port = fullPort;
#endif

		/* initialize the communications line */
		if (!md_init(port))
			Fatal("can't open serial port: %s", port);

        if (!SendFileToXGS(port, outfile, xgsfile))
			Fatal("download failed: %s", xgsfile);

		/* restore the communications line */
		md_done();
	}

    /* exit successfully */
    return 0;
}

/* Usage - display a usage message and exit */
static Usage(void)
{
    fprintf(stderr, "\
usage: bascom [ -o <output-file> ] [ -p <port> ] [ -d <download-file> ] <input-file>\n");
    exit(1);
}

/* AddString - add a string to the string table */
int AddString(ParseContext *c, char *value, int offset)
{
    String *str;
    
    /* check to see if the string is already in the table */
    for (str = c->strings; str != NULL; str = str->next)
        if (strcmp(value, str->value) == 0) {
            int link;
            if (!(link = str->fixups))
                return str->offset;
            else {
                str->fixups = offset;
                return link;
            }
        }

    /* allocate the string structure */
    if (!(str = malloc(sizeof(String) + strlen(value))))
        Fatal("insufficient memory");

    /* initialize the string */
    memset(str, 0, sizeof(String));
    strcpy(str->value, value);
    str->fixups = offset;
    str->next = c->strings;
    c->strings = str;

    /* return zero to terminate the fixup list */
    return 0;
}

/* PlaceStrings - place any string literals defined in the current function */
static void PlaceStrings(ParseContext *c, int offset)
{
    String *str;
    for (str = c->strings; str != NULL; str = str->next) {
        if (str->fixups) {
            size_t size = strlen(str->value) + 1;
            str->offset = offset;
            fixup(c, str->fixups, str->offset);
            offset += size;
        }
    }
}

/* WriteStrings - write any string literals defined in the current function */
static void WriteStrings(ParseContext *c)
{
    String *str;
    for (str = c->strings; str != NULL; str = str->next) {
        if (str->fixups) {
            size_t size = strlen(str->value) + 1;
            //printf("string: %04x '%s'\n", str->offset, str->value);
            fwrite(str->value, size, 1, c->ofp);
            c->codeOff += size;
            str->fixups = 0;
        }
    }
}

/* AddIntrinsic - add an intrinsic function to the global symbol table */
static void AddIntrinsic(ParseContext *c, char *name, char *argTypes, char *retType, int index)
{
    Type *type, *argType;

    /* create a type for the function */
    if (!(type = (Type *)malloc(sizeof(Type))))
        Fatal("insufficient memory");

    /* initialize the type */
    type->id = TYPE_FUNCTION;
    InitSymbolTable(&type->FunctionInfo.arguments);

    /* set the return type */
    if (retType) {
        switch (*retType) {
        case 'f':
            type->FunctionInfo.returnType = &c->floatType;
            break;
        case 'i':
            type->FunctionInfo.returnType = &c->integerType;
            break;
        case 'b':
            type->FunctionInfo.returnType = &c->byteType;
            break;
        case 's':
            type->FunctionInfo.returnType = &c->stringType;
            break;
        default:
            Fatal("unknown return type: %s", retType);
            break;
        }
    }
    else
        type->FunctionInfo.returnType = NULL;


    /* add a dummy symbol for the return value */
    if (retType) {
        if (!AddSymbol(&type->FunctionInfo.arguments, name, SC_VARIABLE, NULL, 0, 0))
            Fatal("insufficient memory");
    }

    /* add the argument types */
    if (argTypes) {
        while (*argTypes) {
            switch (*argTypes++) {
            case 'f':
                argType = &c->floatType;
                break;
            case 'i':
                argType = &c->integerType;
                break;
            case 'b':
                argType = &c->byteType;
                break;
            case 's':
                argType = &c->stringType;
                break;
            default:
                Fatal("unknown argument type: %c", argTypes[-1]);
                break;
            }
            if (!AddSymbol(&type->FunctionInfo.arguments, "", SC_VARIABLE, argType, 0, 0))
                Fatal("insufficient memory");
        }
    }

    /* add a global symbol for the intrinsic function */
    if (!AddSymbol(&c->globals, name, SC_CONSTANT, type, 0, (index << 1) | 1))
        Fatal("insufficient memory");
}

/* CreateOutputFile - create the output file */
static void CreateOutputFile(ParseContext *c, char *name)
{
    /* create the output file */
    if (!(c->ofp = fopen(name, "wb")))
        Fatal("creating output file: %s", name);

    /* skip past where the header will be */
    fseek(c->ofp, sizeof(FileHdr), SEEK_SET);
}

/* WriteFunction - write byte for a function to the output file */
void WriteFunction(ParseContext *c, char *name)
{
    size_t codeSize = c->cptr - c->cbase;

    /* place string literals defined in this function */
    PlaceStrings(c, c->codeOff + codeSize);
    
#ifdef DEBUGIO
    printf("%s:\n", name);
    DecodeFunction(c->codeOff, c->cbase, (int)codeSize);
#endif

    /* write the procedure to the output file */
    if (fwrite(c->cbase, 1, codeSize, c->ofp) != codeSize)
        Fatal("error writing function to output file");

    /* update the offset into the code area */
    c->codeOff += codeSize;

    /* write the string literals defined in this function */
    WriteStrings(c);
    
    /* restore the pointers to the main code */
    c->cptr = c->cbase;
    c->cbase = c->codeBuf;
}

/* WriteOutputFile - write the output file */
static void WriteOutputFile(ParseContext *c)
{
    uint16_t mainCodeOffset;
    FileHdr hdr;
    Symbol *sym;

    /* end the main code with a halt */
    putcbyte(c, OP_HALT);
    
    /* write the main code */
    mainCodeOffset = c->codeOff;
    WriteFunction(c, "main");

    /* setup the file header structure */
    strcpy(hdr.tag, FileTag);
    hdr.version = FileVersion;
    hdr.codeOffset = sizeof(FileHdr);
    hdr.codeSize = c->codeOff;
    hdr.mainCodeOffset = mainCodeOffset;
    hdr.variableOffset = hdr.codeOffset + (hdr.codeSize + 1) & ~1;
    hdr.variableCount = c->globalCount;
    hdr.stackSize = 100;

    /* write the variables */
    fseek(c->ofp, hdr.variableOffset, SEEK_SET);
    for (sym = c->globals.head; sym != NULL; sym = sym->next)
        if (sym->storageClass == SC_VARIABLE) {
            if (sym->type->id == TYPE_ARRAY) {
                int arraySize = sym->type->ArrayInfo.dim1 * sym->type->ArrayInfo.dim2;
                int arrayByteSize = ValueByteSize(sym->type->ArrayInfo.elementType) * arraySize + 1;
                fwrite(&sym->type->ArrayInfo.dim1, sizeof(int16_t), 1, c->ofp);
                fwrite(&arraySize, sizeof(int16_t), 1, c->ofp);
                while ((arrayByteSize -= 2) >= 0)
                    fwrite(&sym->value, sizeof(int16_t), 1, c->ofp);
            }
            else
                fwrite(&sym->value, sizeof(int16_t), 1, c->ofp);
        }

    /* write the output file header */
    fseek(c->ofp, 0, SEEK_SET);
    if (fwrite(&hdr, 1, sizeof(FileHdr), c->ofp) != sizeof(FileHdr))
        Fatal("error writing output file header");

    /* close the output file */
    fclose(c->ofp);

#ifdef DEBUGIO
    DumpSymbols(&c->globals, TRUE);
#endif

    printf("\n");
    printf("code size:  %d\n", hdr.codeSize);
    printf("data size:  %d\n", hdr.variableCount  * sizeof(int16_t));
    printf("stack size: %d\n", hdr.stackSize * sizeof(int16_t));
    printf("total sram: %d\n", (hdr.variableCount + hdr.stackSize) * sizeof(int16_t));
}

/* Alloc - allocate memory and report a fatal error on failure */
void *Alloc(size_t size)
{
    void *p;
    if (!(p = malloc(size)))
        Fatal("insufficient memory");
    return p;
}

/* Fatal - report a fatal error and exit */
void Fatal(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    fputs("error: ", stderr);
    vfprintf(stderr, fmt, ap);
    putc('\n', stderr);
    va_end(ap);
    exit(1);
}
