/* db_bascom.h - definitions for a simple basic compiler
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#ifndef __DB_BASCOM_H__
#define __DB_BASCOM_H__

#include "db_vm_v010.h"

/* program limits */
#define MAXLINE     128
#define MAXTOKEN    32

/* forward type declarations */
typedef struct SymbolTable SymbolTable;
typedef struct Symbol Symbol;
typedef struct TypeTable TypeTable;
typedef struct NamedType NamedType;
typedef struct Type Type;

/* lexical tokens */
typedef enum {
    T_NONE,
    T_REM = 0x100,  /* keywords start here */
    T_DEF,
    T_SUB,
    T_TYPE,
    T_DIM,
    T_AS,
    T_LET,
    T_IF,
    T_THEN,
    T_ELSE,
    T_END,
    T_FOR,
    T_TO,
    T_STEP,
    T_NEXT,
    T_DO,
    T_WHILE,
    T_UNTIL,
    T_LOOP,
    T_GOTO,
    T_CALL,
	T_MOD,
    T_AND,
    T_OR,
	T_NOT,
    T_STOP,
    T_PRINT,
    T_INPUT,
    T_ELSE_IF,  /* compound keywords */
    T_END_DEF,
    T_END_SUB,
    T_END_IF,
    T_DO_WHILE,
    T_DO_UNTIL,
    T_LOOP_WHILE,
    T_LOOP_UNTIL,
    T_LE,       /* non-keyword tokens */
    T_NE,
    T_GE,
	T_SHL,
	T_SHR,
    T_IDENTIFIER,
    T_NUMBER,
    T_STRING,
    T_EOL,
    T_EOF
} Token;

typedef enum {
    BLOCK_NONE,
    BLOCK_IF,
    BLOCK_ELSE,
    BLOCK_FOR,
    BLOCK_DO
} BlockType;

typedef struct Block Block;
struct Block {
    BlockType type;
    union {
        struct {
            int nxt;
            int end;
        } IfBlock;
        struct {
            int end;
        } ElseBlock;
        struct {
            int nxt;
            int end;
        } ForBlock;
        struct {
            int nxt;
            int end;
        } DoBlock;
    };
};

typedef struct String String;
struct String {
    String *next;
    int offset;
    int fixups;
    char value[1];
};

/* storage class ids */
typedef enum {
    SC_CONSTANT,
    SC_VARIABLE
} StorageClass;

/* type ids */
typedef enum {
    TYPE_UNKNOWN,
    TYPE_FLOAT,
    TYPE_INTEGER,
    TYPE_BYTE,
    TYPE_STRING,
    TYPE_ARRAY,
    TYPE_FUNCTION
} TypeID;

/* symbol table */
struct SymbolTable {
    Symbol *head;
    Symbol **pTail;
};

/* symbol structure */
struct Symbol {
    Symbol *next;
    StorageClass storageClass;
    Type *type;
    int offset;
    int16_t value;
    char name[1];
};

/* type table */
struct TypeTable {
    NamedType *head;
    NamedType **pTail;
};

/* named type structure */
struct NamedType {
    NamedType *next;
    Type *type;
    char name[1];
};

/* type definition */
struct Type {
    TypeID  id;
    union {
        struct {
            int length;
        } StringInfo;
        struct {
            Type *elementType;
            int16_t dim1;
            int16_t dim2;
        } ArrayInfo;
        struct {
            SymbolTable fields;
        } UserInfo;
        struct {
            Type *returnType;
            SymbolTable arguments;
            int argumentCount;
        } FunctionInfo;
    };
};

/* parse file */
typedef struct ParseFile ParseFile;
struct ParseFile {
    ParseFile *next;        /* next file in stack */
    FILE *fp;               /* input file pointer */
    char lineBuf[MAXLINE];  /* current input line */
    char *linePtr;          /* pointer to the current character */
    int lineNumber;         /* current line number */
    char name[1];           /* file name */
};

/* parse context */
typedef struct ParseContext ParseContext;
struct ParseContext {
    ParseFile *currentFile; /* scan - current file */
    Token savedToken;       /* scan - lookahead token */
    int tokenOffset;        /* scan - offset to the start of the current token */
    char token[MAXTOKEN];   /* scan - current token string */
    int value;              /* scan - current token integer value */
    float fvalue;           /* scan - current token float value */
    int inComment;          /* scan - inside of a slash/star comment */
    Type floatType;         /* parse - float type */
    Type integerType;       /* parse - integer type */
    Type byteType;          /* parse - byte type */
    Type stringType;        /* parse - string type */
    TypeTable types;        /* parse - type table */
    SymbolTable globals;    /* parse - global variables and constants */
    int globalCount;        /* parse - global variable count */
    String *strings;        /* parse - string constants */
    Symbol *function;       /* parse - symbol for current function definition */
    Type *functionType;     /* parse - type of the current function */
    SymbolTable locals;     /* parse - local variables of current function definition */
    int localCount;         /* parse - local variable count */
    Block blockBuf[10];     /* parse - stack of nested blocks */
    Block *bptr;            /* parse - current block */
    Block *btop;            /* parse - top of block stack */
    uint16_t codeOff;       /* bytecode - offset to next available code byte */
    uint8_t codeBuf[1024];  /* bytecode - code buffer */
    uint8_t *cbase;         /* bytecode - base of current function */
    uint8_t *cptr;          /* bytecode - next available code position */
    uint8_t *ctop;          /* bytecode - top of code buffer */
    FILE *ofp;              /* output - output file pointer */
};

typedef enum {
    PV_LOAD,
    PV_STORE
} PValOp;

/* partial value structure */
typedef struct PVAL PVAL;
struct PVAL {
    Type *type;
    void (*fcn)(ParseContext *c, PValOp op, PVAL *pv);
    int val;
};

/* db_bascom.c */
int AddString(ParseContext *c, char *value, int offset);
void WriteFunction(ParseContext *c, char *name);
void *Alloc(size_t size);
void Fatal(char *fmt, ...);

/* db_stat.c */
void ParseStatement(ParseContext *c);

/* db_expr.c */
void ParseExpr(ParseContext *c, PVAL *pv);
void ParsePrimary(ParseContext *c, PVAL *pv);
void ParseSubroutineCall(ParseContext *c);
void GetLValue(ParseContext *c, char *name, PVAL *pv);
Type *DefaultType(ParseContext *c, char *name);
int IsLiteral(PVAL *pv);
void rvalue(ParseContext *c, PVAL *pv);
void chklvalue(ParseContext *c, PVAL *pv);

/* db_scan.c */
void InitParseContext(ParseContext *c);
int PushFile(ParseContext *c, char *name);
void CloseParseContext(ParseContext *c);
int GetLine(ParseContext *c);
void FRequire(ParseContext *c, Token requiredToken);
void Require(ParseContext *c, Token token, Token requiredToken);
int GetToken(ParseContext *c);
void SaveToken(ParseContext *c, Token token);
char *TokenName(Token token);
void ParseError(ParseContext *c, char *fmt, ...);

/* db_sym.c */
void InitSymbolTable(SymbolTable *table);
void FreeSymbols(SymbolTable *table);
Symbol *AddSymbol(SymbolTable *table, char *name, StorageClass storageClass, Type *type, int offset, int value);
Symbol *FindSymbol(SymbolTable *table, char *name);
void DumpSymbols(SymbolTable *table, int showTags);

/* db_type.c */
void InitTypeTable(TypeTable *table);
Type *AddType(TypeTable *table, char *name, Type *type);
Type *FindType(TypeTable *table, char *name);
int CompareTypes(Type *type1, Type *type2);
int ValueSize(Type *type, int *pPrefix);
int ValueByteSize(Type *type);

/* db_bcode.c */
int codeaddr(ParseContext *c);
int putcbyte(ParseContext *c, int b);
int putcword(ParseContext *c, int w);
void fixup(ParseContext *c, int chn, int val);
void fixupbranch(ParseContext *c, int chn, int val);

/* db_xmsnd.c */
int SendFileToXGS(char *port, char *srcname, char *dstname);

#endif

