/* db_scan.c - token scanner
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "db_bascom_v010.h"

#ifdef WIN32
#define strcasecmp(a, b)    _stricmp(a, b)
#endif

/* keyword table */
static struct {
    char *keyword;
    Token token;
} ktab[] = {

/* these must be in the same order as the Token enum */
{   "REM",      T_REM       },
{   "DEF",      T_DEF       },
{   "SUB",      T_SUB       },
{   "TYPE",     T_TYPE      },
{   "DIM",      T_DIM       },
{   "AS",       T_AS        },
{   "LET",      T_LET       },
{   "IF",       T_IF        },
{   "THEN",     T_THEN      },
{   "ELSE",     T_ELSE      },
{   "END",      T_END       },
{   "FOR",      T_FOR       },
{   "TO",       T_TO        },
{   "STEP",     T_STEP      },
{   "NEXT",     T_NEXT      },
{   "DO",       T_DO        },
{   "WHILE",    T_WHILE     },
{   "UNTIL",    T_UNTIL     },
{   "LOOP",     T_LOOP      },
{   "GOTO",     T_GOTO      },
{   "CALL",     T_CALL      },
{   "MOD",      T_MOD       },
{   "AND",      T_AND       },
{   "OR",       T_OR        },
{	"NOT",		T_NOT		},
{   "STOP",     T_STOP      },
{   "PRINT",    T_PRINT     },
{   "INPUT",    T_INPUT     },
{   NULL,       0           }
};

/* local function prototypes */
static int NextToken(ParseContext *c);
static int IdentifierToken(ParseContext *c, int ch);
static int NumberToken(ParseContext *c, int ch);
static int HexNumberToken(ParseContext *c);
static int IdentifierCharP(int ch);
static int StringToken(ParseContext *c);
static int SkipSpaces(ParseContext *c);
static int SkipComment(ParseContext *c);
static int GetChar(ParseContext *c);
static int XGetC(ParseContext *c);
static void UngetC(ParseContext *c);

/* InitParseContext - initialize a parse context */
void InitParseContext(ParseContext *c)
{
    memset(c, 0, sizeof(ParseContext));
    c->floatType.id = TYPE_FLOAT;
    c->integerType.id = TYPE_INTEGER;
    c->byteType.id = TYPE_BYTE;
    c->stringType.id = TYPE_STRING;
    c->btop = (Block *)((char *)c->blockBuf + sizeof(c->blockBuf));
    c->bptr = c->blockBuf - 1;
    c->ctop = c->codeBuf + sizeof(c->codeBuf);
    c->cbase = c->codeBuf;
    c->cptr = c->cbase;
}

/* PushFile - push a file onto the input file stack */
int PushFile(ParseContext *c, char *name)
{
    ParseFile *f;
    
    /* allocate a parse file structure */
    if (!(f = (ParseFile *)malloc(sizeof(ParseFile) + strlen(name))))
        return FALSE;
    
    /* open the input file */
    if (!(f->fp = fopen(name, "r"))) {
        free(f);
        return FALSE;
    }

    /* initialize the parse context */
    strcpy(f->name, name);
    f->linePtr = f->lineBuf;
    *f->linePtr = '\0';
    f->lineNumber = 0;
    
    /* push the file onto the input file stack */
    f->next = c->currentFile;
    c->currentFile = f;
    
    /* return successfully */
    return TRUE;
}

/* CloseParseContext - close a parse context */
void CloseParseContext(ParseContext *c)
{
    ParseFile *f,*next;
    
    /* close all of the currently open files */
    for (f = c->currentFile; f != NULL; f = next) {
        next = f->next;
        fclose(f->fp);
        free(f);
    }

    /* clear the file stack */
    c->currentFile = NULL;
}

/* GetLine - get the next input line */
int GetLine(ParseContext *c)
{
    ParseFile *f;

    /* get the next input line */
    for (;;) {
        
        /* get the current input file */
        if (!(f = c->currentFile))
            return FALSE;
        
        /* get a line from the current file */
        if (fgets(f->lineBuf, sizeof(f->lineBuf), f->fp))
            break;

        /* pop the input file stack on end of file */
        else {
            c->currentFile = f->next;
            fclose(f->fp);
            free(f);
        }
    }
    
    /* initialize the input buffer */
    f->linePtr = f->lineBuf;
    ++f->lineNumber;
    
    /* return successfully */
    return TRUE;
}

/* FRequire - fetch a token and check it */
void FRequire(ParseContext *c, Token requiredToken)
{
    Require(c, GetToken(c), requiredToken);
}

/* Require - check for a required token */
void Require(ParseContext *c, Token token, Token requiredToken)
{
    char tknbuf[MAXTOKEN];
    if (token != requiredToken) {
        strcpy(tknbuf, TokenName(requiredToken));
        ParseError(c, "Expecting '%s', found '%s'", tknbuf, TokenName(token));
    }
}

/* GetToken - get the next token */
int GetToken(ParseContext *c)
{
    int tkn;

    /* check for a saved token */
    if ((tkn = c->savedToken) != T_NONE)
        c->savedToken = T_NONE;

    /* otherwise, get the next token */
    else
        tkn = NextToken(c);

    /* return the token */
    return tkn;
}

/* SaveToken - save a token */
void SaveToken(ParseContext *c, Token token)
{
    c->savedToken = token;
}

/* TokenName - get the name of a token */
char *TokenName(Token token)
{
    static char nameBuf[4];
    char *name;

    switch (token) {
    case T_NONE:
        name = "<NONE>";
        break;
    case T_REM:
    case T_DEF:
    case T_SUB:
    case T_TYPE:
    case T_DIM:
    case T_AS:
    case T_LET:
    case T_IF:
    case T_THEN:
    case T_ELSE:
    case T_END:
    case T_FOR:
    case T_TO:
    case T_STEP:
    case T_NEXT:
    case T_DO:
    case T_WHILE:
    case T_UNTIL:
    case T_LOOP:
    case T_GOTO:
    case T_CALL:
	case T_MOD:
    case T_AND:
    case T_OR:
	case T_NOT:
    case T_STOP:
    case T_PRINT:
    case T_INPUT:
        name = ktab[token - T_REM].keyword;
        break;
    case T_END_DEF:
        name = "END DEF";
        break;
    case T_END_SUB:
        name = "END SUB";
        break;
    case T_END_IF:
        name = "END IF";
        break;
    case T_DO_WHILE:
        name = "DO WHILE";
        break;
    case T_DO_UNTIL:
        name = "DO UNTIL";
        break;
    case T_LOOP_WHILE:
        name = "LOOP WHILE";
        break;
    case T_LOOP_UNTIL:
        name = "LOOP UNTIL";
        break;
    case T_LE:
        name = "<=";
        break;
    case T_NE:
        name = "<>";
        break;
    case T_GE:
        name = ">=";
        break;
	case T_SHL:
		name = "<<";
		break;
	case T_SHR:
		name = ">>";
		break;
    case T_IDENTIFIER:
        name = "<IDENTIFIER>";
        break;
    case T_NUMBER:
        name = "<NUMBER>";
        break;
    case T_STRING:
        name = "<STRING>";
        break;
    case T_EOL:
        name = "<EOL>";
        break;
    case T_EOF:
        name = "<EOF>";
        break;
    default:
        nameBuf[0] = '\'';
        nameBuf[1] = token;
        nameBuf[2] = '\'';
        nameBuf[3] = '\0';
        name = nameBuf;
        break;
    }

    /* return the token name */
    return name;
}

/* NextToken - read the next token */
static int NextToken(ParseContext *c)
{
    ParseFile *f;
    int ch, tkn;
    
    /* get the current file */
    if (!(f = c->currentFile))
        return T_EOF;

    /* skip leading blanks */
    ch = SkipSpaces(c);

    /* remember the start of the current token */
    c->tokenOffset = f->linePtr - f->lineBuf;

    /* check the next character */
    switch (ch) {
    case EOF:
        tkn = T_EOL;
        break;
    case '"':
        tkn = StringToken(c);
        break;
    case '<':
        if ((ch = GetChar(c)) == '=')
            tkn = T_LE;
        else if (ch == '>')
            tkn = T_NE;
		else if (ch == '<')
			tkn = T_SHL;
        else {
            UngetC(c);
            tkn = '<';
        }
        break;
    case '>':
        if ((ch = GetChar(c)) == '=')
            tkn = T_GE;
		else if (ch == '>')
			tkn = T_SHR;
        else {
            UngetC(c);
            tkn = '>';
        }
        break;
	case '0':
		if ((ch = GetChar(c)) == 'x')
			tkn = HexNumberToken(c);
		else {
			UngetC(c);
			tkn = NumberToken(c, '0');
		}
		break;
    default:
        if (isdigit(ch))
            tkn = NumberToken(c, ch);
        else if (IdentifierCharP(ch)) {
            char *savePtr;
            switch (tkn = IdentifierToken(c,ch)) {
            case T_ELSE:
                savePtr = f->linePtr;
                if ((ch = SkipSpaces(c)) != EOF && IdentifierCharP(ch)) {
                    switch (IdentifierToken(c, ch)) {
                    case T_IF:
                        tkn = T_ELSE_IF;
                        break;
                    default:
                        f->linePtr = savePtr;
                        break;
                    }
                }
                else
                    f->linePtr = savePtr;
                break;
            case T_END:
                savePtr = f->linePtr;
                if ((ch = SkipSpaces(c)) != EOF && IdentifierCharP(ch)) {
                    switch (IdentifierToken(c, ch)) {
                    case T_DEF:
                        tkn = T_END_DEF;
                        break;
                    case T_SUB:
                        tkn = T_END_SUB;
                        break;
                    case T_IF:
                        tkn = T_END_IF;
                        break;
                    default:
                        f->linePtr = savePtr;
                        break;
                    }
                }
                else
                    f->linePtr = savePtr;
                break;
            case T_DO:
                savePtr = f->linePtr;
                if ((ch = SkipSpaces(c)) != EOF && IdentifierCharP(ch)) {
                    switch (IdentifierToken(c, ch)) {
                    case T_WHILE:
                        tkn = T_DO_WHILE;
                        break;
                    case T_UNTIL:
                        tkn = T_DO_UNTIL;
                        break;
                    default:
                        f->linePtr = savePtr;
                        break;
                    }
                }
                else
                    f->linePtr = savePtr;
                break;
            case T_LOOP:
                savePtr = f->linePtr;
                if ((ch = SkipSpaces(c)) != EOF && IdentifierCharP(ch)) {
                    switch (IdentifierToken(c, ch)) {
                    case T_WHILE:
                        tkn = T_LOOP_WHILE;
                        break;
                    case T_UNTIL:
                        tkn = T_LOOP_UNTIL;
                        break;
                    default:
                        f->linePtr = savePtr;
                        break;
                    }
                }
                else
                    f->linePtr = savePtr;
                break;
            }
        }
        else
            tkn = ch;
    }

    /* return the token */
    return tkn;
}

/* IdentifierToken - get an identifier */
static int IdentifierToken(ParseContext *c, int ch)
{
    int len, i;
    char *p;

    /* get the identifier */
    p = c->token; *p++ = ch; len = 1;
    while ((ch = GetChar(c)) != EOF && IdentifierCharP(ch)) {
        if (++len > MAXTOKEN)
            ParseError(c, "Identifier too long");
        *p++ = ch;
    }
    UngetC(c);
    *p = '\0';

    /* check to see if it is a keyword */
    for (i = 0; ktab[i].keyword != NULL; ++i)
        if (strcasecmp(ktab[i].keyword, c->token) == 0)
            return ktab[i].token;

    /* otherwise, it is an identifier */
    return T_IDENTIFIER;
}

/* NumberToken - get a number */
static int NumberToken(ParseContext *c, int ch)
{
    char *p = c->token;

    /* get the number */
    *p++ = ch;
    while ((ch = GetChar(c)) != EOF) {
		if (isdigit(ch))
			*p++ = ch;
		else if (ch != '_')
			break;
	}
    UngetC(c);
    *p = '\0';
    
    /* convert the string to an integer */
    c->value = atol(c->token);
    
    /* return the token */
    return T_NUMBER;
}

/* HexNumberToken - get a hexadecimal number */
static int HexNumberToken(ParseContext *c)
{
    char *p = c->token;
	int ch;

    /* get the number */
    while ((ch = GetChar(c)) != EOF) {
		if (isxdigit(ch))
			*p++ = ch;
		else if (ch != '_')
			break;
	}
    UngetC(c);
    *p = '\0';
    
    /* convert the string to an integer */
    c->value = strtoul(c->token, NULL, 16);
    
    /* return the token */
    return T_NUMBER;
}

/* IdentifierCharP - is this an identifier character? */
static int IdentifierCharP(int ch)
{
    return isupper(ch)
        || islower(ch)
        || isdigit(ch)
        || strchr("$%_", ch) != NULL;
}

/* StringToken - get a string */
static int StringToken(ParseContext *c)
{
    int ch,len;
    char *p;

    /* collect the string */
    p = c->token; len = 0;
    while ((ch = GetChar(c)) != EOF && ch != '"') {
        if (++len > MAXTOKEN)
            ParseError(c, "String too long");
        *p++ = ch;
    }
    *p = '\0';

    /* check for premature end of file */
    if (ch != '"')
        ParseError(c, "unterminated string");

    /* return the token */
    return T_STRING;
}

/* SkipSpaces - skip leading spaces and the the next non-blank character */
static int SkipSpaces(ParseContext *c)
{
    int ch;
    while ((ch = GetChar(c)) != EOF)
        if (!isspace(ch))
            break;
    return ch;
}

/* SkipComment - skip characters up to the end of a comment */
static int SkipComment(ParseContext *c)
{
    int lastch, ch;
    lastch = '\0';
    while ((ch = XGetC(c)) != EOF) {
        if (lastch == '*' && ch == '/')
            return TRUE;
        lastch = ch;
    }
    return FALSE;
}

/* GetChar - get the next character */
static int GetChar(ParseContext *c)
{
    int ch;

    /* check for being in a comment */
    if (c->inComment) {
        if (!SkipComment(c))
            return EOF;
        c->inComment = FALSE;
    }

    /* loop until we find a non-comment character */
    for (;;) {
        
        /* get the next character */
        if ((ch = XGetC(c)) == EOF)
            return EOF;

        /* check for a comment (starts with '//' or '/*') */
        else if (ch == '/') {
            if ((ch = XGetC(c)) == '/') {
                while ((ch = XGetC(c)) != EOF)
                    ;
            }
            else if (ch == '*') {
                if (!SkipComment(c)) {
                    c->inComment = TRUE;
                    return EOF;
                }
            }
            else {
                UngetC(c);
                ch = '/';
                break;
            }
        }

        /* handle a normal character */
        else
            break;
    }

    /* return the character */
    return ch;
}

/* XGetC - get the next character without checking for comments */
static int XGetC(ParseContext *c)
{
    ParseFile *f;
    int ch;
    
    /* get the current input file */
    if (!(f = c->currentFile))
        return EOF;
    
    /* get the next character on the current line */
    if (!(ch = *f->linePtr++)) {
        --f->linePtr;
        return EOF;
    }
    
    /* return the character */
    return ch;
}

/* UngetC - unget the most recent character */
static void UngetC(ParseContext *c)
{
    ParseFile *f;
    
    /* get the current input file */
    if (!(f = c->currentFile)) {
        ParseError(c, "internal error");
        return;
    }

    /* backup the input pointer */
    --f->linePtr;
}

/* ParseError - report a parsing error */
void ParseError(ParseContext *c, char *fmt, ...)
{
    ParseFile *f;
    va_list ap;

    /* print the error message */
    va_start(ap, fmt);
    fputs("error: ", stderr);
    vfprintf(stderr, fmt, ap);
    putc('\n', stderr);
    va_end(ap);

    /* show the context */
    if ((f = c->currentFile) != NULL) {
        fprintf(stderr, "  file '%s', line %d\n", f->name, f->lineNumber);
        fprintf(stderr, "    %s\n", f->lineBuf);
        fprintf(stderr, "    %*s\n", c->tokenOffset, "^");
    }

    /* exit until we fix the compiler so it can recover from parse errors */
    exit(1);
}
