/* db_vmmain.c - a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

/* needs to be built with:

XGS_PIC_GFX_DRV_V010.c
XGS_PIC_NTSC_160_192_2_V010.s
XGS_PIC_UART_DRV_V010.c
XGS_PIC_GAMEPAD_DRV_V010.c
XGS_PIC_SYSTEM_V010.c
XGS_PIC_SOUND_DRV_V010.c
FSIO.c
SD-SPI.c
db_vmmain_v010.c
db_vmimage_v010.c
db_vmint_v010.c
db_vmdebug_v010.c
db_vmfcn_v010.c
db_text_v010.c
db_xmrcv_v010.c
db_crctab_v010.c

In order to make serial downloads reliable, it is necessary to change the following definition in XGS_PIC_UART_DRV_V010.h:

  #define UART1_RX_BUFFER_SIZE	134

The receive buffer needs to be large enough to hold an XMODEM/CRC packet which is 133 bytes long.

In MPLAB under Project -> Build Options -> Project
and then under the "MPLAB LINK30" tab, set the heap size to 2048. 

*/

#include "XGS_PIC_SYSTEM_V010.h"			
#include "XGS_PIC_NTSC_160_192_2_V010.h"
#include "XGS_PIC_GFX_DRV_V010.h"
#include "XGS_PIC_UART_DRV_V010.h"
#include "XGS_PIC_GAMEPAD_DRV_V010.h"
#include "XGS_PIC_SOUND_DRV_V010.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "db_vm_v010.h"
#include "db_text_v010.h"
#include "db_xmrcv_v010.h"

// Not defined but actually located in the gcc library!
int vsprintf( char *, const char *, va_list );
int vsnprintf(char *, size_t, const char *, va_list);

unsigned char	g_VRAMBuffer[VRAM_BUF_SIZE] __attribute__((far));	
unsigned short 	g_PaletteMap[PALETTE_MAP_SIZE];
unsigned char	g_Palettes[MIN_PALETTE_SIZE];

static VMFILE FileOpen(char *name, char *mode);
static void FileClose(VMFILE file);
static size_t FileRead(void *buf, size_t count, VMFILE file);
static size_t FileWrite(void *buf, size_t count, VMFILE file);
static int FileSeek(VMFILE file, long offset, int whence);
static long FileTell(VMFILE file);

int main(void)
{
	int imageLoaded;
	Interpreter i;

	UART_Init(9600);
	XmInit();

	Gamepad_Init();
	SND_Init();

	SYS_ConfigureClock(FCY_NTSC);

	FSInit();

	GFX_InitBitmap(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP, g_VRAMBuffer, g_Palettes, g_PaletteMap);
	GFX_SetClipRegion(0, SCREEN_HEIGHT - 1, 0, SCREEN_WIDTH - 1);
	
	memset(g_PaletteMap, 0, sizeof(g_PaletteMap));
	
	g_Palettes[0] = NTSC_BLACK;
	g_Palettes[1] = NTSC_RED;
	g_Palettes[2] = NTSC_GREEN;
	g_Palettes[3] = NTSC_BLUE;

	GFX_StartDrawing(SCREEN_TYPE_NTSC);

	i.fileOpen = FileOpen;
	i.fileClose = FileClose;
	i.fileRead = FileRead;
	i.fileWrite = FileWrite;
	i.fileSeek = FileSeek;
	i.fileTell = FileTell;
	i.poll = XmPoll;
	InitFcn(&i);

	if (!(imageLoaded = LoadImage(&i, "autorun.bai"))) {
	    DrawBitmapFontString("XGS Basic Runtime 1.0", 0, 20, g_VRAMBuffer);
	    DrawBitmapFontString("load failed: autorun.bai", 0, 40, g_VRAMBuffer);
    }   

	for (;;) {
		if (imageLoaded)
			Execute(&i);
        else
            XmPoll();
	}

	return 0;
}

void VM_printf(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
	VM_vprintf(fmt, ap);
	va_end(ap);
}

void VM_vprintf(char *fmt, va_list ap)
{
	char buf[128];
	vsnprintf(buf, sizeof(buf), fmt, ap);
	VM_puts(buf);
}

void VM_puts(char *str)
{
	while (*str) {
		if (*str == '\n')
			UART_putchar('\r');
		UART_putchar(*str++);
	}
}

void VM_putchar(int ch)
{
	if (ch == '\n')
		UART_putchar('\r');
	UART_putchar(ch);
}

static VMFILE FileOpen(char *name, char *mode)
{
	return (VMFILE)FSfopen(name, mode);
}

static void FileClose(VMFILE file)
{
	FSfclose((FSFILE *)file);
}

static size_t FileRead(void *buf, size_t count, VMFILE file)
{
	return FSfread(buf, 1, count, (FSFILE *)file);
}

static size_t FileWrite(void *buf, size_t count, VMFILE file)
{
	return FSfwrite(buf, 1, count, (FSFILE *)file);
}

static int FileSeek(VMFILE file, long offset, int whence)
{
	return FSfseek((FSFILE *)file, offset, whence);
}

static long FileTell(VMFILE file)
{
	return FSftell((FSFILE *)file);
}
