/* db_sendfile.c - file send protocol
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <string.h>
#include "../common/db_packet_v010.h"

/* local variables */
static int pktn;		/* current packet number */

/* prototypes */
static int snddata(FILE *fp, long fsize);
static int sndinit(void);
static int sndhdr(char *name);
static int sndblk(uint8_t *buf,int len);
static int sndeot(void);
static int sndpkt(int type, uint8_t *buf, int len);

int SendFile(char *srcname, char *dstname)
{
    long fsize;
    FILE *fp;
    int sts;

    /* open the file */
    if (!(fp = fopen(srcname, "rb")))
        return VMFALSE;

    /* get the file size */
    fseek(fp, 0, SEEK_END);
    fsize = ftell(fp);

    /* seek back to the start of the file */
    fseek(fp, 0, SEEK_SET);

    /* initialize the packet connection */
    if ((sts = sndinit()) != ER_OK) {
        fclose(fp);
        return VMFALSE;
    }

    /* send the filename header and the data */
    if ((sts = sndhdr(dstname)) == ER_OK)
        sts = snddata(fp, fsize);

    /* close the file */
    fclose(fp);

    /* return successfully */
    return sts;
}

int RunFile(char *name)
{
    int len = (int)strlen(name);

    /* make sure the name isn't too long */
    if (len >= PKTMAXLEN)
        return VMFALSE;

    /* send the packet */
    return sndpkt(PKT_RUN, (uint8_t *)name, len + 1);
}

/* snddata - send data using x/ymodem */
static int snddata(FILE *fp, long fsize)
{
    unsigned char buf[PKTMAXLEN], *bp;
    int bc, ch, sts;

    /* send each byte */
    for (bp = buf, bc = 0; (ch = getc(fp)) != EOF; ) {
        *bp++ = ch;
        if (++bc == sizeof(buf)) {
            printf("Remaining: %ld\r", fsize);
            if ((sts = sndblk(buf, sizeof(buf))) != ER_OK)
                return sts;
            fsize -= bc;
            bp = buf;
            bc = 0;
        }
    }

    /* flush partial buffer */
    if (bc > 0) {
        if ((sts = sndblk(buf, bc)) != ER_OK)
            return sts;
        printf("Download complete\n");
    }

    /* send EOT and wait for ACK */
    return sndeot();
}

/* sndinit - initialize to send data using x/ymodem */
static int sndinit(void)
{
    pktn = 0;
    return ER_OK;
}

/* sndhdr - send a ymodem batch header */
static int sndhdr(char *name)
{
    int len = (int)strlen(name);

    /* make sure the name isn't too long */
    if (len >= PKTMAXLEN)
        return VMFALSE;

    /* send the packet */
    return sndpkt(PKT_OPEN, (uint8_t *)name, len + 1);
}

/* sndblk - send a data block using x/ymodem */
static int sndblk(uint8_t *buf, int len)
{
    return sndpkt(PKT_DATA, buf, len);
}

/* sndeot - end a data transfer using x/ymodem */
static int sndeot(void)
{
    return sndpkt(PKT_CLOSE, NULL, 0);
}

/* sndpkt - send a packet */
static int sndpkt(int type, uint8_t *buf, int len)
{
    int retries;

    PktBuildFrame(type, buf, len);

    /* send data and wait for an ACK */
    for (retries = RETRY; retries > 0; --retries) {

        /* send the current frame */
        PktSendCurrentFrame();

        /* wait for ACK */
        for (;;) {

            /* get the next ack/nak packet */
            type = PktPoll(XTIME);
            
            /* check for timeout */
            if (type == PKT_TIMEOUT) {
                //md_log("timeout waiting for ACK/NAK");
                break;
            }

            /* return successfully on ACK */
            else if (type == PKT_ACK) {
                //md_log("received ACK");
                return ER_OK;
            }

            /* send packet again on NAK */
            else {
                //md_log("received NAK");
                break;
            }
        }
    }

    /* retry count ran out */
    return ER_DATA;
}

