/* db_vm.h - definitions for a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#ifndef __DB_VM_H__
#define __DB_VM_H__

#include <stdio.h>
#include "../common/db_types_v010.h"

/* forward type declarations */
typedef struct Interpreter Interpreter;
typedef struct FileHdr FileHdr;

/* intrinsic function handler type */
typedef void IntrinsicFcn(Interpreter *i);

/* interpreter state structure */
struct Interpreter {
    IntrinsicFcn *intrinsics[16];
    int16_t *variables;
    int16_t *stack;
    int16_t *stackTop;
    int16_t *fp;
    int16_t *sp;
    uint8_t *code;
    uint8_t *pc;
    uint16_t mainCodeOffset;
    uint16_t linePos;
    VMFILE *(*fileOpen)(char *name, char *mode);
    void (*fileClose)(VMFILE *fp);
    size_t (*fileRead)(void *buf, size_t count, VMFILE *fp);
    size_t (*fileWrite)(void *buf, size_t count, VMFILE *fp);
    int (*fileSeek)(VMFILE *fp, long offset, int whence);
    long (*fileTell)(VMFILE *fp);
    void (*poll)(void);
    int pollCount;
    int target;
};

/* number of bytecode instructions per poll */
#define BC_POLL_INTERVAL	10

/* stack manipulation macros */
#define Reserve(i, n)   do { if ((i)->sp + (n) > (i)->stackTop) StackOverflow(); else (i)->sp += (n); } while (0)
#define CPush(i, v)     do { if ((i)->sp < (i)->stackTop) Push(i, v); else StackOverflow(); } while (0)
#define Push(i, v)      (*++(i)->sp = (v))
#define Pop(i)          (*(i)->sp--)
#define Drop(i, n)      ((i)->sp -= (n))

/* prototypes */
void StackOverflow(void);

/* opcodes */
#define OP_HALT         0x00    /* halt */
#define OP_BRT          0x01    /* branch on true */
#define OP_BRTSC        0x02    /* branch on true (for short circuit booleans) */
#define OP_BRF          0x03    /* branch on false */
#define OP_BRFSC        0x04    /* branch on false (for short circuit booleans) */
#define OP_BR           0x05    /* branch unconditionally */
#define OP_NOT          0x06    /* logical negate top of stack */
#define OP_NEG          0x07    /* negate */
#define OP_ADD          0x08    /* add two numeric expressions */
#define OP_SUB          0x09    /* subtract two numeric expressions */
#define OP_MUL          0x0a    /* multiply two numeric expressions */
#define OP_DIV          0x0b    /* divide two numeric expressions */
#define OP_REM          0x0c    /* remainder of two numeric expressions */
#define OP_BNOT         0x0d    /* bitwise not of two numeric expressions */
#define OP_BAND         0x0e    /* bitwise and of two numeric expressions */
#define OP_BOR          0x0f    /* bitwise or of two numeric expressions */
#define OP_BXOR         0x10    /* bitwise exclusive or */
#define OP_SHL          0x11    /* shift left */
#define OP_SHR          0x12    /* shift right */
#define OP_LT           0x13    /* less than */
#define OP_LE           0x14    /* less than or equal to */
#define OP_EQ           0x15    /* equal to */
#define OP_NE           0x16    /* not equal to */
#define OP_GE           0x17    /* greater than or equal to */
#define OP_GT           0x18    /* greater than */
#define OP_LIT          0x19    /* load literal */
#define OP_GREF         0x1a    /* load a global variable */
#define OP_GSET         0x1b    /* set a global variable */
#define OP_GREFB        0x1c    /* load a global byte variable */
#define OP_GSETB        0x1d    /* set a global byte variable */
#define OP_GREFF        0x1e    /* load a global float variable */
#define OP_GSETF        0x1f    /* set a global float variable */
#define OP_LOFF         0x20    /* load the offset into variable space for a local variable */
#define OP_LREF         0x21    /* load a local variable relative to the frame pointer */
#define OP_LSET         0x22    /* set a local variable relative to the frame pointer */
#define OP_LREFB        0x23    /* load a local byte variable relative to the frame pointer */
#define OP_LSETB        0x24    /* set a local byte variable relative to the frame pointer */
#define OP_LREFF        0x25    /* load a local float variable relative to the frame pointer */
#define OP_LSETF        0x26    /* set a local float variable relative to the frame pointer */
#define OP_AREF1        0x27    /* load an element of an array */
#define OP_ASET1        0x28    /* set an element of an array */
#define OP_AREF2        0x29    /* load an element of an array */
#define OP_ASET2        0x2a    /* set an element of an array */
#define OP_AREF1B       0x2b    /* load a byte element of an array */
#define OP_ASET1B       0x2c    /* set a byte element of an array */
#define OP_AREF2B       0x2d    /* load a byte element of an array */
#define OP_ASET2B       0x2e    /* set a byte element of an array */
#define OP_AREF1F       0x2f    /* load a float element of an array */
#define OP_ASET1F       0x30    /* set a float element of an array */
#define OP_AREF2F       0x31    /* load a float element of an array */
#define OP_ASET2F       0x32    /* set a float element of an array */
#define OP_RESERVE      0x33    /* reserve space on the stack */
#define OP_CALL         0x34    /* call a function */
#define OP_FRET         0x35    /* return from a function leaving the result on the stack */
#define OP_SRET         0x36    /* return from a subroutine */
#define OP_PRINTF       0x37    /* print a float */
#define OP_PRINTI       0x38    /* print an integer */
#define OP_PRINTS       0x39    /* print a string */
#define OP_PRINTT       0x3a    /* print a tab */
#define OP_PRINTN       0x3b    /* print a newline */
#define OP_OPEN         0x3c    /* open a file */
#define OP_CLOSE        0x3d    /* close a file */
#define OP_WRITES       0x3e    /* write a string */

/* stack frame offsets */
#define F_FP    1
#define F_PC    2
#define F_SIZE  2

/* image file signature */
#define FileTag         "DMBASIC"
#define FileVersion     0x0001

/* target runtimes */
#define TARGET_UNKNOWN      0
#define TARGET_BITMAP		1
#define TARGET_TILE			2

/* image file header */
struct FileHdr {
    char        tag[8];         /* must be 'DMBASIC' */
    uint16_t    version;        /* must be 0x0001 */
    uint8_t     target;         /* target runtime */
    uint8_t     unused;         /* (unused) */
    uint16_t    codeOffset;     /* offset to the bytecodes */
    uint16_t    codeSize;       /* size of bytecodes in bytes */
    uint16_t    mainCodeOffset; /* offset to the main code */
    uint16_t    variableOffset; /* offset to the variable table */
    uint16_t    variableCount;  /* number of variables */
    uint16_t    stackSize;      /* stack size */
};

/* prototypes from db_vmimage.c  */
int LoadImage(Interpreter *i, char *name);

/* prototypes for image memory allocation functions */
int AllocateImage(Interpreter *i, int codeSize, int dataSize);
void FreeImage(Interpreter *i);

/* prototypes from db_vmint.c */
void InitInterpreter(Interpreter *i, int target);
void LoadProgram(char *name);
void Execute(Interpreter *i);

/* prototypes from db_vmdebug.c */
void DecodeFunction(uint16_t base, uint8_t *code, int len);
int DecodeInstruction(uint16_t base, uint8_t *code, uint8_t *lc);
void ShowStack(Interpreter *i);

/* intrinsic functions */
void InitFcn(Interpreter *i);

#endif
