/* db_vmint.c - bytecode interpreter for a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "db_vm_v010.h"

#define INDEX1                                                              \
            do {                                                            \
			    ind1 = Pop(i);                                              \
			    tmp = i->sp[0];                                             \
			    if (ind1 < 1 || ind1 > i->variables[tmp - 1])               \
				    Abort("array subscript out of bounds: %d", ind1);       \
            } while (0)

#define INDEX2                                                              \
            do {                                                            \
			    ind2 = Pop(i);                                              \
			    ind1 = Pop(i);                                              \
			    tmp = i->sp[0];                                             \
			    if (ind1 < 1 || ind1 > i->variables[tmp - 2])               \
				    Abort("first array subscript out of bounds: %d", ind1); \
                ind1 += (ind2 - 1) * i->variables[tmp - 2];                 \
			    if (ind1 < 1 || ind1 > i->variables[tmp - 1])               \
				    Abort("second array subscript out of bounds: %d", ind2);\
            } while (0)

/* current interpreter */
static Interpreter *interpreter = NULL;

/* prototypes for local functions */
static void Restart(Interpreter *i);
static void PrintFloat(Interpreter *i, float value);
static void PrintInteger(Interpreter *i, int16_t value);
static void PrintString(Interpreter *i, char *value);
static void PrintTab(Interpreter *i);
static void PrintNewline(Interpreter *i);
static void OpenFile(Interpreter *i);
static void CloseFile(Interpreter *i);
static void WriteString(Interpreter *i);
static void Abort(char *fmt, ...);

/* InitInterpreter - initialize the interpreter */
void InitInterpreter(Interpreter *i, int target)
{
    i->target = target;
    i->pollCount = BC_POLL_INTERVAL;
    i->linePos = 0;
    interpreter = i;
}

/* LoadProgram - load a new program */
void LoadProgram(char *name)
{
    LoadImage(interpreter, name);
    Restart(interpreter);
}

/* Execute - execute the main code */
void Execute(Interpreter *i)
{
    int16_t tmp, tmp2, ind1, ind2;
    
	Restart(i);

	for (;;) {
		if (--i->pollCount <= 0) {
			i->pollCount = BC_POLL_INTERVAL;
			(*i->poll)();
		}
#if 0
		ShowStack(i);
		DecodeInstruction(0, i->code, i->pc);
#endif
		switch (*i->pc++) {
		case OP_HALT:
			return;
		case OP_BRT:
			tmp = *i->pc++;
			tmp |= (int16_t)(*i->pc++ << 8);
			if (Pop(i))
				i->pc += tmp;
			break;
		case OP_BRTSC:
			tmp = *i->pc++;
			tmp |= (int16_t)(*i->pc++ << 8);
			if (*i->sp)
				i->pc += tmp;
			else
				Drop(i, 1);
			break;
		case OP_BRF:
			tmp = *i->pc++;
			tmp |= (int16_t)(*i->pc++ << 8);
			if (!Pop(i))
				i->pc += tmp;
			break;
		case OP_BRFSC:
			tmp = *i->pc++;
			tmp |= (int16_t)(*i->pc++ << 8);
			if (!*i->sp)
				i->pc += tmp;
			else
				Drop(i, 1);
			break;
		case OP_BR:
			tmp = *i->pc++;
			tmp |= (int16_t)(*i->pc++ << 8);
			i->pc += tmp;
			break;
		case OP_NOT:
			*i->sp = (*i->sp ? VMFALSE : VMTRUE);
			break;
		case OP_NEG:
			*i->sp = -*i->sp;
			break;
		case OP_ADD:
			tmp = Pop(i);
			*i->sp += tmp;
			break;
		case OP_SUB:
			tmp = Pop(i);
			*i->sp -= tmp;
			break;
		case OP_MUL:
			tmp = Pop(i);
			*i->sp *= tmp;
			break;
		case OP_DIV:
			tmp = Pop(i);
			*i->sp = (tmp == 0 ? 0 : *i->sp / tmp);
			break;
		case OP_REM:
			tmp = Pop(i);
			*i->sp = (tmp == 0 ? 0 : *i->sp % tmp);
			break;
		case OP_BNOT:
			*i->sp = ~*i->sp;
			break;
		case OP_BAND:
			tmp = Pop(i);
			*i->sp &= tmp;
			break;
		case OP_BOR:
			tmp = Pop(i);
			*i->sp |= tmp;
			break;
		case OP_BXOR:
			tmp = Pop(i);
			*i->sp ^= tmp;
			break;
		case OP_SHL:
			tmp = Pop(i);
			*i->sp <<= tmp;
			break;
		case OP_SHR:
			tmp = Pop(i);
			*i->sp >>= tmp;
			break;
		case OP_LT:
			tmp = Pop(i);
			*i->sp = (*i->sp < tmp ? VMTRUE : VMFALSE);
			break;
		case OP_LE:
			tmp = Pop(i);
			*i->sp = (*i->sp <= tmp ? VMTRUE : VMFALSE);
			break;
		case OP_EQ:
			tmp = Pop(i);
			*i->sp = (*i->sp == tmp ? VMTRUE : VMFALSE);
			break;
		case OP_NE:
			tmp = Pop(i);
			*i->sp = (*i->sp != tmp ? VMTRUE : VMFALSE);
			break;
		case OP_GE:
			tmp = Pop(i);
			*i->sp = (*i->sp >= tmp ? VMTRUE : VMFALSE);
			break;
		case OP_GT:
			tmp = Pop(i);
			*i->sp = (*i->sp > tmp ? VMTRUE : VMFALSE);
			break;
		case OP_LIT:
			tmp = *i->pc++;
			CPush(i, (int16_t)(*i->pc++ << 8) | tmp);
			break;
		case OP_GREF:
			tmp = *i->pc++;
			CPush(i, i->variables[(int16_t)(*i->pc++ << 8) | tmp]);
			break;
		case OP_GSET:
			tmp = *i->pc++;
			i->variables[(int16_t)(*i->pc++ << 8) | tmp] = Pop(i);
			break;
		case OP_GREFB:
			tmp = *i->pc++;
			CPush(i, *(uint8_t *)&i->variables[(int16_t)(*i->pc++ << 8) | tmp]);
			break;
		case OP_GSETB:
			tmp = *i->pc++;
			*(uint8_t *)&i->variables[(int16_t)(*i->pc++ << 8) | tmp] = (uint8_t)Pop(i);
			break;
		case OP_GREFF:
			tmp = *i->pc++;
			CPush(i, *(uint8_t *)&i->variables[(int16_t)(*i->pc++ << 8) | tmp]);
			break;
		case OP_GSETF:
			tmp = *i->pc++;
			*(uint8_t *)&i->variables[(int16_t)(*i->pc++ << 8) | tmp] = (uint8_t)Pop(i);
			break;
		case OP_LOFF:
			tmp = *i->pc++;
			CPush(i, (int16_t)(i->fp - i->variables) + tmp);
			break;
		case OP_LREF:
			CPush(i, i->fp[(int)*(int8_t *)(i->pc)++]);
			break;
		case OP_LSET:
			i->fp[(int)*(int8_t *)(i->pc)++] = Pop(i);
			break;
		case OP_LREFB:
			CPush(i, *(uint8_t *)&i->fp[(int)*(int8_t *)(i->pc)++]);
			break;
		case OP_LSETB:
			*(uint8_t *)&i->fp[(int)*(int8_t *)(i->pc)++] = (uint8_t)Pop(i);
			break;
		case OP_LREFF:
			CPush(i, *(uint8_t *)&i->fp[(int)*(int8_t *)(i->pc)++]);
			break;
		case OP_LSETF:
			*(uint8_t *)&i->fp[(int)*(int8_t *)(i->pc)++] = (uint8_t)Pop(i);
			break;
		case OP_AREF1:
            INDEX1;
			i->sp[0] = i->variables[tmp + ind1 - 1];
			break;
		case OP_ASET1:
			tmp2 = Pop(i);
            INDEX1;
			i->variables[tmp + ind1 - 1] = tmp2;
            Drop(i, 1);
			break;
		case OP_AREF2:
            INDEX2;
			i->sp[0] = i->variables[tmp + ind1 - 1];
			break;
		case OP_ASET2:
			tmp2 = Pop(i);
            INDEX2;
            i->variables[tmp + ind1 - 1] = tmp2;
            Drop(i, 1);
			break;
		case OP_AREF1B:
            INDEX1;
			i->sp[0] = ((uint8_t *)&i->variables[tmp])[ind1 - 1];
			break;
		case OP_ASET1B:
			tmp2 = Pop(i);
            INDEX1;
			((uint8_t *)&i->variables[tmp])[ind1 - 1] = (uint8_t)tmp2;
            Drop(i, 1);
			break;
		case OP_AREF2B:
            INDEX2;
			i->sp[0] = ((uint8_t *)&i->variables[tmp])[ind1 - 1];
			break;
		case OP_ASET2B:
			tmp2 = Pop(i);
            INDEX2;
            ((uint8_t *)&i->variables[tmp])[ind1 - 1] = (uint8_t)tmp2;
            Drop(i, 1);
			break;
		case OP_AREF1F:
            INDEX1;
			i->sp[0] = ((uint8_t *)&i->variables[tmp])[ind1 - 1];
			break;
		case OP_ASET1F:
			tmp2 = Pop(i);
            INDEX1;
			((uint8_t *)&i->variables[tmp])[ind1 - 1] = (uint8_t)tmp2;
            Drop(i, 1);
			break;
		case OP_AREF2F:
            INDEX2;
			i->sp[0] = ((uint8_t *)&i->variables[tmp])[ind1 - 1];
			break;
		case OP_ASET2F:
			tmp2 = Pop(i);
            INDEX2;
            ((uint8_t *)&i->variables[tmp])[ind1 - 1] = (uint8_t)tmp2;
            Drop(i, 1);
			break;
		case OP_RESERVE:
			tmp = *i->pc++;
			Reserve(i, tmp);
			break;
		case OP_CALL:
			tmp = i->sp[-(int16_t)*i->pc++];
			if (tmp & 1)
				(*i->intrinsics[tmp >> 1])(i);
			else {
				tmp2 = (int16_t)(i->fp - i->stack);
				i->fp = i->sp;
				Reserve(i, F_SIZE);
				i->fp[F_FP] = tmp2;
				i->fp[F_PC] = (int16_t)(i->pc - i->code);
				i->pc = i->code + tmp;
			}
			break;
		case OP_FRET:
			i->pc = i->code + i->fp[F_PC];
			i->sp = i->fp;
			Drop(i, i->pc[-1]);
			i->fp = i->stack + i->fp[F_FP];
			break;
		case OP_SRET:
			i->pc = i->code + i->fp[F_PC];
			i->sp = i->fp;
			Drop(i, i->pc[-1] + 1);
			i->fp = i->stack + i->fp[F_FP];
			break;
        case OP_PRINTF:
            PrintFloat(i, *(float *)&i->sp[-1]);
            break;
        case OP_PRINTI:
            PrintInteger(i, Pop(i));
            break;
        case OP_PRINTS:
	        PrintString(i, (char *)(i->code + Pop(i)));
            break;
        case OP_PRINTT:
            PrintTab(i);
            break;
        case OP_PRINTN:
            PrintNewline(i);
            break;
        case OP_OPEN:
            OpenFile(i);
            break;
        case OP_CLOSE:
            CloseFile(i);
            break;
        case OP_WRITES:
            WriteString(i);
            break;
		default:
			Abort("undefined opcode 0x%02x", i->pc[-1]);
			break;
		}
	}
}

static void Restart(Interpreter *i)
{
    i->pc = i->code + i->mainCodeOffset;
    i->sp = i->fp = i->stack - 1;
}

static void PrintFloat(Interpreter *i, float value)
{
    char buf[20];
    sprintf(buf, "%g", (double)value);
    PrintString(i, buf);
}

static void PrintInteger(Interpreter *i, int16_t value)
{
    char buf[10];
    sprintf(buf, "%d", value);
    PrintString(i, buf);
}

static void PrintString(Interpreter *i, char *value)
{
    int ch;
    while ((ch = *value++) != '\0') {
        VM_putchar(ch);
        ++i->linePos;
    }
}

static void PrintTab(Interpreter *i)
{
    do {
        VM_putchar(' ');
        ++i->linePos;
    } while ((i->linePos % 8) != 0);
}

static void PrintNewline(Interpreter *i)
{
    VM_putchar('\n');
    i->linePos = 0;
}

static void OpenFile(Interpreter *i)
{
}

static void CloseFile(Interpreter *i)
{
}

static void WriteString(Interpreter *i)
{
}

void StackOverflow(void)
{
    Abort("stack overflow");
}

static void Abort(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_puts("error: ");
    VM_vprintf(fmt, ap);
    VM_putchar('\n');
    va_end(ap);
    exit(1);
}
