/* db_vmint.c - bytecode interpreter for a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "db_vm.h"

#define INDEX1                                                              \
            do {                                                            \
                ind1 = Pop(i);                                              \
                tmp = i->sp[0];                                             \
                if (ind1 < 1 || ind1 > i->variables[tmp - 1])               \
                    Abort("array subscript out of bounds: %d", ind1);       \
            } while (0)

#define INDEX2                                                              \
            do {                                                            \
                ind2 = Pop(i);                                              \
                ind1 = Pop(i);                                              \
                tmp = i->sp[0];                                             \
                if (ind1 < 1 || ind1 > i->variables[tmp - 2])               \
                    Abort("first array subscript out of bounds: %d", ind1); \
                ind1 += (ind2 - 1) * i->variables[tmp - 2];                 \
                if (ind1 < 1 || ind1 > i->variables[tmp - 1])               \
                    Abort("second array subscript out of bounds: %d", ind2);\
            } while (0)

/* prototypes for local functions */
static void PrintFloat(Interpreter *i, float value);
static void PrintInteger(Interpreter *i, int16_t value);
static void PrintString(Interpreter *i, char *value);
static void PrintTab(Interpreter *i);
static void PrintNewline(Interpreter *i);
static void OpenFile(Interpreter *i);
static void CloseFile(Interpreter *i);
static void WriteString(Interpreter *i);
static void Abort(char *fmt, ...);

/* InitInterpreter - initialize the interpreter */
void InitInterpreter(Interpreter *i, int target)
{
    i->target = target;
    i->pollCount = BC_POLL_INTERVAL;
    i->linePos = 0;
}

/* VerifyAndSetupImageHeader - verify and setup an image header */
static int VerifyAndSetupImageHeader(Interpreter *i, FileHdr *hdr)
{
    int dataSize;

    /* verify the image header */
    if (strncmp(hdr->tag, FILE_TAG, strlen(FILE_TAG)) != 0 || hdr->version != FILE_VERSION) {
        VM_printf("header verify failed\n");
        return VMFALSE;
    }

    /* verify the target */
    if (hdr->target != i->target) {
        VM_printf("not built for this target: expected %d, got %d\n", i->target, hdr->target);
        return VMFALSE;
    }

    /* allocate the code and data space */
    dataSize = (hdr->variableCount + hdr->stackSize) * sizeof(int16_t);
    if (!VM_AllocateImage(i, hdr->codeSize, hdr->initSize, dataSize)) {
        VM_printf("image space allocation failed: code %d, data %d\n", hdr->codeSize, dataSize);
        return VMFALSE;
    }

    /* clear the variable space */
    memset(i->variables, 0, hdr->variableCount * sizeof(int16_t));
    i->init_state = INIT_HDR1;

    /* setup the stack */
    i->stack = (int16_t *)(i->variables + hdr->variableCount);
    i->stackTop = i->stack + hdr->stackSize;

    /* return successfully */
    return VMTRUE;
}

/* InitMemoryImage - initialize an image already in memory */
int InitMemoryImage(Interpreter *i)
{    
    uint16_t off, cnt;
    uint8_t buf[128];
    
    /* get the memory image header */
    if (!VM_LoadMemoryImageHeader(i))
        return VMFALSE;

    /* verify and setup the header */
    if (!VerifyAndSetupImageHeader(i, &i->hdr))
        return VMFALSE;

    /* process the initialization data */
    for (off = 0; off < i->hdr.initSize; off += cnt) {

        /* determine how much initializer data to read */
        if ((cnt = i->hdr.initSize - off) > sizeof(buf))
			cnt = sizeof(buf);

        /* read the next block of initializer data */
        VM_ReadInit(off, buf, cnt);

        /* process the initializer data */
        ProcessInitData(i, buf, cnt);
    }

    /* finish loading */
    FinishImageLoad(i);

    /* return successfully */
    return VMTRUE;
}

/* StartImageLoad - start loading an image */
int StartImageLoad(Interpreter *i, FileHdr *hdr)
{
    /* verify and setup the image header */
    if (!VerifyAndSetupImageHeader(i, hdr))
        return VMFALSE;

    /* store the image header */
    i->hdr = *hdr;

    /* no image loaded yet */
    i->imageLoaded = VMFALSE;

    /* return successfully */
    return VMTRUE;
}

/* ProcessInitData - process a block of variable initializers */
void ProcessInitData(Interpreter *i, uint8_t *buf, int len)
{
    uint8_t *end = buf + len;
    while (buf < end) {
        switch (i->init_state) {
        case INIT_HDR1:
            i->init_off = *(uint16_t *)buf;
            i->init_state = INIT_HDR2;
            buf += sizeof(uint16_t);
            break;
        case INIT_HDR2:
            i->init_cnt = *(uint16_t *)buf;
            i->init_state = INIT_DATA;
            buf += sizeof(uint16_t);
            break;
        case INIT_DATA:
            i->variables[i->init_off++] = *(int16_t *)buf;
            buf += sizeof(int16_t);
            if (--i->init_cnt == 0)
                i->init_state = INIT_HDR1;
            break;
        }
    }
}

/* FinishImageLoad - finish loading an image */
void FinishImageLoad(Interpreter *i)
{
    i->imageLoaded = VMTRUE;
}

/* Execute - execute the main code */
void Execute(Interpreter *i)
{
    int16_t tmp, tmp2, ind1, ind2;
    int8_t tmpb;
    
    Start(i);

    for (;;) {
        if (--i->pollCount <= 0) {
            i->pollCount = BC_POLL_INTERVAL;
            VM_poll();
            if (!i->imageLoaded)
                return;
        }
#if 0
        ShowStack(i);
        DecodeInstruction(0, i->code, i->pc);
#endif
        switch (VMCODEBYTE(i->pc++)) {
        case OP_HALT:
            i->imageLoaded = VMFALSE;
            return;
        case OP_BRT:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)(VMCODEBYTE(i->pc++) << 8);
            if (Pop(i))
                i->pc += tmp;
            break;
        case OP_BRTSC:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)(VMCODEBYTE(i->pc++) << 8);
            if (*i->sp)
                i->pc += tmp;
            else
                Drop(i, 1);
            break;
        case OP_BRF:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)(VMCODEBYTE(i->pc++) << 8);
            if (!Pop(i))
                i->pc += tmp;
            break;
        case OP_BRFSC:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)(VMCODEBYTE(i->pc++) << 8);
            if (!*i->sp)
                i->pc += tmp;
            else
                Drop(i, 1);
            break;
        case OP_BR:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)(VMCODEBYTE(i->pc++) << 8);
            i->pc += tmp;
            break;
        case OP_NOT:
            *i->sp = (*i->sp ? VMFALSE : VMTRUE);
            break;
        case OP_NEG:
            *i->sp = -*i->sp;
            break;
        case OP_ADD:
            tmp = Pop(i);
            *i->sp += tmp;
            break;
        case OP_SUB:
            tmp = Pop(i);
            *i->sp -= tmp;
            break;
        case OP_MUL:
            tmp = Pop(i);
            *i->sp *= tmp;
            break;
        case OP_DIV:
            tmp = Pop(i);
            *i->sp = (tmp == 0 ? 0 : *i->sp / tmp);
            break;
        case OP_REM:
            tmp = Pop(i);
            *i->sp = (tmp == 0 ? 0 : *i->sp % tmp);
            break;
        case OP_BNOT:
            *i->sp = ~*i->sp;
            break;
        case OP_BAND:
            tmp = Pop(i);
            *i->sp &= tmp;
            break;
        case OP_BOR:
            tmp = Pop(i);
            *i->sp |= tmp;
            break;
        case OP_BXOR:
            tmp = Pop(i);
            *i->sp ^= tmp;
            break;
        case OP_SHL:
            tmp = Pop(i);
            *i->sp <<= tmp;
            break;
        case OP_SHR:
            tmp = Pop(i);
            *i->sp >>= tmp;
            break;
        case OP_LT:
            tmp = Pop(i);
            *i->sp = (*i->sp < tmp ? VMTRUE : VMFALSE);
            break;
        case OP_LE:
            tmp = Pop(i);
            *i->sp = (*i->sp <= tmp ? VMTRUE : VMFALSE);
            break;
        case OP_EQ:
            tmp = Pop(i);
            *i->sp = (*i->sp == tmp ? VMTRUE : VMFALSE);
            break;
        case OP_NE:
            tmp = Pop(i);
            *i->sp = (*i->sp != tmp ? VMTRUE : VMFALSE);
            break;
        case OP_GE:
            tmp = Pop(i);
            *i->sp = (*i->sp >= tmp ? VMTRUE : VMFALSE);
            break;
        case OP_GT:
            tmp = Pop(i);
            *i->sp = (*i->sp > tmp ? VMTRUE : VMFALSE);
            break;
        case OP_LIT:
            tmp = VMCODEBYTE(i->pc++);
            CPush(i, (int16_t)(VMCODEBYTE(i->pc++) << 8) | tmp);
            break;
        case OP_GREF:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            CPush(i, i->variables[tmp]);
            break;
        case OP_GSET:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            i->variables[tmp] = Pop(i);
            break;
        case OP_GREFB:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            CPush(i, *(uint8_t *)&i->variables[tmp]);
            break;
        case OP_GSETB:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            *(uint8_t *)&i->variables[tmp] = (uint8_t)Pop(i);
            break;
        case OP_GREFF:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            CPush(i, *(uint8_t *)&i->variables[tmp]);
            break;
        case OP_GSETF:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            *(uint8_t *)&i->variables[tmp] = (uint8_t)Pop(i);
            break;
        case OP_LOFF:
            tmp = VMCODEBYTE(i->pc++);
            CPush(i, (int16_t)(i->fp - i->variables) + tmp);
            break;
        case OP_LREF:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            CPush(i, i->fp[(int)tmpb]);
            break;
        case OP_LSET:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            i->fp[(int)tmpb] = Pop(i);
            break;
        case OP_LREFB:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            CPush(i, *(uint8_t *)&i->fp[(int)tmpb]);
            break;
        case OP_LSETB:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            *(uint8_t *)&i->fp[(int)tmpb] = (uint8_t)Pop(i);
            break;
        case OP_LREFF:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            CPush(i, *(uint8_t *)&i->fp[(int)tmpb]);
            break;
        case OP_LSETF:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            *(uint8_t *)&i->fp[(int)tmpb] = (uint8_t)Pop(i);
            break;
        case OP_AREF1:
            INDEX1;
            i->sp[0] = i->variables[tmp + ind1 - 1];
            break;
        case OP_ASET1:
            tmp2 = Pop(i);
            INDEX1;
            i->variables[tmp + ind1 - 1] = tmp2;
            Drop(i, 1);
            break;
        case OP_AREF2:
            INDEX2;
            i->sp[0] = i->variables[tmp + ind1 - 1];
            break;
        case OP_ASET2:
            tmp2 = Pop(i);
            INDEX2;
            i->variables[tmp + ind1 - 1] = tmp2;
            Drop(i, 1);
            break;
        case OP_AREF1B:
            INDEX1;
            i->sp[0] = ((uint8_t *)&i->variables[tmp])[ind1 - 1];
            break;
        case OP_ASET1B:
            tmp2 = Pop(i);
            INDEX1;
            ((uint8_t *)&i->variables[tmp])[ind1 - 1] = (uint8_t)tmp2;
            Drop(i, 1);
            break;
        case OP_AREF2B:
            INDEX2;
            i->sp[0] = ((uint8_t *)&i->variables[tmp])[ind1 - 1];
            break;
        case OP_ASET2B:
            tmp2 = Pop(i);
            INDEX2;
            ((uint8_t *)&i->variables[tmp])[ind1 - 1] = (uint8_t)tmp2;
            Drop(i, 1);
            break;
        case OP_AREF1F:
            INDEX1;
            i->sp[0] = ((uint8_t *)&i->variables[tmp])[ind1 - 1];
            break;
        case OP_ASET1F:
            tmp2 = Pop(i);
            INDEX1;
            ((uint8_t *)&i->variables[tmp])[ind1 - 1] = (uint8_t)tmp2;
            Drop(i, 1);
            break;
        case OP_AREF2F:
            INDEX2;
            i->sp[0] = ((uint8_t *)&i->variables[tmp])[ind1 - 1];
            break;
        case OP_ASET2F:
            tmp2 = Pop(i);
            INDEX2;
            ((uint8_t *)&i->variables[tmp])[ind1 - 1] = (uint8_t)tmp2;
            Drop(i, 1);
            break;
        case OP_RESERVE:
            tmp = VMCODEBYTE(i->pc++);
            Reserve(i, tmp);
            break;
        case OP_CALL:
            tmp = i->sp[-(int16_t)VMCODEBYTE(i->pc++)];
            if (tmp & 1)
                (*i->intrinsics[tmp >> 1])(i);
            else {
                tmp2 = (int16_t)(i->fp - i->stack);
                i->fp = i->sp;
                Reserve(i, F_SIZE);
                i->fp[F_FP] = tmp2;
                i->fp[F_PC] = (int16_t)(i->pc - i->code);
                i->pc = i->code + tmp;
            }
            break;
        case OP_FRET:
            i->pc = i->code + i->fp[F_PC];
            i->sp = i->fp;
            Drop(i, VMCODEBYTE(i->pc - 1));
            i->fp = i->stack + i->fp[F_FP];
            break;
        case OP_SRET:
            i->pc = i->code + i->fp[F_PC];
            i->sp = i->fp;
            Drop(i, VMCODEBYTE(i->pc - 1) + 1);
            i->fp = i->stack + i->fp[F_FP];
            break;
        case OP_PRINTF:
            PrintFloat(i, *(float *)&i->sp[-1]);
            break;
        case OP_PRINTI:
            PrintInteger(i, Pop(i));
            break;
        case OP_PRINTS:
            PrintString(i, (char *)(i->code + Pop(i)));
            break;
        case OP_PRINTT:
            PrintTab(i);
            break;
        case OP_PRINTN:
            PrintNewline(i);
            break;
        case OP_OPEN:
            OpenFile(i);
            break;
        case OP_CLOSE:
            CloseFile(i);
            break;
        case OP_WRITES:
            WriteString(i);
            break;
        default:
            Abort("undefined opcode 0x%02x", VMCODEBYTE(i->pc - 1));
            break;
        }
    }
}

void Start(Interpreter *i)
{
    i->pc = i->code + i->hdr.mainCodeOffset;
    i->sp = i->fp = i->stack - 1;
}

static void PrintFloat(Interpreter *i, float value)
{
    char buf[20], *p = buf;
    sprintf(buf, "%g", (double)value);
    while (*p != '\0') {
        VM_putchar(*p++);
        ++i->linePos;
    }
}

static void PrintInteger(Interpreter *i, int16_t value)
{
    char buf[20], *p = buf;
    sprintf(buf, "%d", value);
    while (*p != '\0') {
        VM_putchar(*p++);
        ++i->linePos;
    }
}

static void PrintString(Interpreter *i, char *value)
{
    int ch;
    while ((ch = VMCODEBYTE(value++)) != '\0') {
        VM_putchar(ch);
        ++i->linePos;
    }
}

static void PrintTab(Interpreter *i)
{
    do {
        VM_putchar(' ');
        ++i->linePos;
    } while ((i->linePos % 8) != 0);
}

static void PrintNewline(Interpreter *i)
{
    VM_putchar('\n');
    i->linePos = 0;
}

static void OpenFile(Interpreter *i)
{
}

static void CloseFile(Interpreter *i)
{
}

static void WriteString(Interpreter *i)
{
}

void StackOverflow(void)
{
    Abort("stack overflow");
}

static void Abort(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_printf("error: ");
    VM_vprintf(fmt, ap);
    VM_putchar('\n');
    va_end(ap);
    VM_flush();
    exit(1);
}
