/* db_vmavr.c - a simple virtual machine for the XGS AVR system
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

/* needs to be built with:

db_crctab_v010.c
db_packet_v010.c
db_picuart_v010.c
db_vmdebug_v010.c
db_vmfcn_v010.c
db_vmimage_v010.c
db_vmint_v010.c
db_vmavr_v010.c
XGS_AVR_GAMEPAD_DRV_V010.c
XGS_AVR_GFX_DRV_V010.c
XGS_AVR_SOUND_DRV_V010.c
XGS_AVR_SYSTEM_V010.c

For the bitmap target define the symbol BITMAP_GRAPHICS and include:

db_text_v010.c
XGS_AVR_NTSC_160_192_2_V010.s

For the tile target define the symbol TILE_GRAPHICS and include:

XGS_AVR_NTSC_TILE6_V010.s

Add the following linker option to place the bootloader code at the correct address:

-Wl,--section-start=.bootloader=0xe000

*/

#include "XGS_AVR_SYSTEM_V010.h"			
#include "XGS_AVR_GFX_LIB_V010.h"
#include "XGS_AVR_GAMEPAD_DRV_V010.h"
#include "XGS_AVR_SOUND_DRV_V010.h"

#ifdef BITMAP_GRAPHICS
#include "XGS_AVR_NTSC_160_192_2_V010.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/boot.h>
#include "../runtime/db_vm.h"
#include "../runtime/db_text.h"
#include "../common/db_packet.h"

#define SPM_PAGEOFFSET  (SPM_PAGESIZE - 1)
#define SPM_PAGEMASK    ~SPM_PAGEOFFSET

#define VMCODE_SIZE		4096
#define VMDATA_SIZE    	2500

static uint8_t PROGMEM vmcodebuf[VMCODE_SIZE + 2 * SPM_PAGESIZE - 1] = { 0 };

static FileHdr *vmimage;
static uint8_t *vmcode;
static uint8_t vmdata[VMDATA_SIZE];	

#ifdef BITMAP_GRAPHICS
//unsigned char	g_VRAMBuffer[VRAM_BUF_SIZE] __attribute__((far));	
//unsigned short 	g_PaletteMap[PALETTE_MAP_SIZE];
//unsigned char	g_Palettes[MIN_PALETTE_SIZE];
#define TARGET_TYPE TARGET_BITMAP
#endif

#ifdef TILE_GRAPHICS
#define TARGET_TYPE TARGET_TILE
#endif

Interpreter interp;

static void WriteFlash(uint8_t *dst, uint8_t *src, int cnt) BOOTLOADER_SECTION;

int main(void)
{
    Interpreter *i = &interp;

	// make sure the code starts on a flash page boundry
	vmimage = (FileHdr *)(((uint16_t)vmcodebuf + SPM_PAGESIZE - 1) & SPM_PAGEMASK);
    vmcode = (uint8_t *)vmimage + SPM_PAGESIZE;

    PktOpen(NULL);
	sei();

    VM_printf("XGS Basic Runtime 1.0\n");

    Gamepad_Init();
    Sound_Init();

#ifdef BITMAP_GRAPHICS
//    GFX_InitBitmap(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP, g_VRAMBuffer, g_Palettes, g_PaletteMap);
//    GFX_SetClipRegion(0, SCREEN_HEIGHT - 1, 0, SCREEN_WIDTH - 1);
    
//    memset(g_PaletteMap, 0, sizeof(g_PaletteMap));
    
//    g_Palettes[0] = NTSC_BLACK;
//    g_Palettes[1] = NTSC_RED;
//    g_Palettes[2] = NTSC_GREEN;
//    g_Palettes[3] = NTSC_BLUE;
//
//    GFX_StartDrawing(SCREEN_TYPE_NTSC);
#endif

    InitInterpreter(i, TARGET_TYPE);
    InitFcn(i);

//    if (!LoadImage(i, "autorun.bai")) {
//#ifdef BITMAP_GRAPHICS
//        DrawBitmapFontString("XGS Basic Runtime 1.0", 0, 20, g_VRAMBuffer);
//        DrawBitmapFontString("load failed: autorun.bai", 0, 40, g_VRAMBuffer);
//#else
//        VM_printf("load failed: autorun.bai\n");
//#endif
//    }

    InitMemoryImage(i);

    for (;;) {
        if (i->imageLoaded)
            Execute(i);
        else
            PktPoll(0);
    }

    return 0;
}

/* VM_LoadMemoryImageHeader - load the header of an image already in memory */
int VM_LoadMemoryImageHeader(Interpreter *i)
{
    memcpy_P(&i->hdr, vmimage, sizeof(FileHdr));
    return VMTRUE;
}

/* VM_StoreMemoryImageHeader - store the header of an image in memory */
void VM_StoreMemoryImageHeader(Interpreter *i)
{
    WriteFlash((uint8_t *)vmimage, (uint8_t *)&i->hdr, sizeof(FileHdr));
}

/* VM_AllocateImage - allocate space for an image */
int VM_AllocateImage(Interpreter *i, int codeSize, int initSize, int dataSize)
{
    /* round up the code size to a flash page boundry */
    codeSize = (codeSize + CODE_BLOCK_SIZE - 1) & ~(CODE_BLOCK_SIZE - 1);
    
    /* make sure there is enough memory */
    if (codeSize + initSize > VMCODE_SIZE || dataSize > VMDATA_SIZE)
        return VMFALSE;

    /* setup the code and data space */
    i->code = vmcode;
    i->variables = (int16_t *)vmdata;

    /* return successfully */
    return VMTRUE;
}

/* VM_FreeImage - free an image */
void VM_FreeImage(Interpreter *i)
{
}

/* VM_LoadCode - load code into flash */
void VM_LoadCode(uint16_t dstoff, uint8_t *src, int cnt)
{
    WriteFlash(vmcode + dstoff, src, cnt);
}

/* VM_LoadInit - load initialization data into flash */
void VM_LoadInit(uint16_t dstoff, uint8_t *src, int cnt)
{
    uint8_t *init = vmcode + ((interp.hdr.codeSize + CODE_BLOCK_SIZE - 1) & ~(CODE_BLOCK_SIZE - 1));
    WriteFlash(init + dstoff, src, cnt);
}

/* VM_ReadInit - read initialization data from flash */
void VM_ReadInit(uint16_t off, uint8_t *buf, int cnt)
{
    uint8_t *p = vmcode + ((interp.hdr.codeSize + CODE_BLOCK_SIZE) & ~(CODE_BLOCK_SIZE - 1)) + off;
    while (--cnt >= 0)
        *buf++ = VMCODEBYTE(p++);
}

static void WriteFlash(uint8_t *dst, uint8_t *src, int cnt)
{
    uint16_t page, pageend, end, wdst;
    uint16_t *wsrc = (uint16_t *)src;

    /* make sure src and dst are word addresses */
    if (((uint16_t)src & 1) || ((uint16_t)dst & 1))
        return; // should report an error here

    /* make cnt an even number of words */
    cnt = (cnt + 1) & ~1;
    
    /* determine the destination end address */
    end = (uint16_t)dst + cnt;

    /* disable interrupts */
    cli();

    /* wait for any eeprom operations to complete */
    eeprom_busy_wait();

    /* transfer one flash page at a time */
    for (page = (uint16_t)dst & SPM_PAGEMASK; page < end; page += SPM_PAGESIZE) {

        /* determine the address of the end of the flash page */
        pageend = page + SPM_PAGESIZE;

        /* copy any data on the page that comes before what we are writing */
        for (wdst = page; wdst < (uint16_t)dst; wdst += 2) {
            uint16_t tmp = VMCODEBYTE(wdst);
            tmp |= VMCODEBYTE(wdst + 1) << 8;
            boot_page_fill(wdst, tmp);
        }

        /* copy the data we are writing */
        for (; wdst < pageend && wdst < end; wdst += 2)
            boot_page_fill(wdst, *wsrc++);

        /* copy the data on the page that comes after what we are writing */
        for (; wdst < pageend; wdst += 2) {
            uint16_t tmp = VMCODEBYTE(wdst);
            tmp |= VMCODEBYTE(wdst + 1) << 8;
            boot_page_fill(wdst, tmp);
        }

        /* erase the page */
        boot_page_erase(page);
        boot_spm_busy_wait();

        /* write the updated page */
        boot_page_write(page);
        boot_spm_busy_wait();

        /* enable the read-while-write memory */
        boot_rww_enable();
    }

    /* enable the interrupts */
    sei();
}

void VM_printf(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_vprintf(fmt, ap);
    va_end(ap);
}

void VM_vprintf(char *fmt, va_list ap)
{
    char buf[80], *p;
    vsnprintf(buf, sizeof(buf), fmt, ap);
    for (p = buf; *p != '\0'; ++p)
        VM_putchar(*p);
}

static char debugBuf[16];
static int debugLen = 0;

static void DebugBufferCh(int ch)
{
    if (debugLen < sizeof(debugBuf)) {
        debugBuf[debugLen++] = ch;
        if (debugLen == sizeof(debugBuf))
            VM_flush();
    }
}

void VM_putchar(int ch)
{
    if (ch == '\n')
        DebugBufferCh('\r');
    DebugBufferCh(ch);
    if (ch == '\n')
        VM_flush();
}

void VM_flush(void)
{
    if (debugLen > 0) {
        PktSendFrame(PKT_DEBUG, (uint8_t *)debugBuf, debugLen);
        debugLen = 0;
    }
}

VMFILE *VM_fopen(const char *name, const char *mode)
{
    return NULL;
}

int VM_fclose(VMFILE *fp)
{
	return 0;
}

size_t VM_fread(void *buf, size_t count, VMFILE *fp)
{
    return 0;
}

size_t VM_fwrite(const void *buf, size_t count, VMFILE *fp)
{
    return 0;
}

int VM_fseek(VMFILE *fp, long offset, int whence)
{
    return 0;
}

long VM_ftell(VMFILE *fp)
{
    return 0;
}

void VM_poll(void)
{
	PktPoll(0);
}
