/* db_vmpic.c - a simple virtual machine for the XGS PIC system
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

/* needs to be built with:

db_vmpic.c
db_vmfcnpic.c
db_picuart.c
../common/db_crctab.c
../common/db_packet.c
../runtime/db_vmdebug.c
../runtime/db_vmfcn.c
../runtime/db_vmimage.c
../runtime/db_vmint.c
FSIO.c
SD-SPI.c
XGS_PIC_GAMEPAD_DRV_V010.c
XGS_PIC_GFX_DRV_V010.c
XGS_PIC_SOUND_DRV_V010.c
XGS_PIC_SYSTEM_V010.c
XGS_PIC_UART_DRV_V010.c

For the bitmap target define the symbol BITMAP_GRAPHICS and include:

db_text_v010.c
XGS_PIC_NTSC_160_192_2_V010.s

For the tile target define the symbol TILE_GRAPHICS and include:

XGS_PIC_NTSC_TILE2_V010.s

In order to make serial downloads reliable, it is necessary to change the following definition in XGS_PIC_UART_DRV_V010.h:

  #define UART1_RX_BUFFER_SIZE	135

The receive buffer needs to be large enough to hold a 5+128+2=135 byte packet.

In MPLAB under Project -> Build Options -> Project
and then under the "MPLAB LINK30" tab, set the heap size to 512. 

*/

#include "XGS_PIC_SYSTEM_V010.h"			
#include "XGS_PIC_GFX_DRV_V010.h"
#include "XGS_PIC_UART_DRV_V010.h"
#include "XGS_PIC_GAMEPAD_DRV_V010.h"
#include "XGS_PIC_SOUND_DRV_V010.h"

#ifdef BITMAP_GRAPHICS
#include "XGS_PIC_NTSC_160_192_2_V010.h"
#define VMIMAGE_SIZE    3000
#endif

#ifdef TILE_GRAPHICS
#define VMIMAGE_SIZE    4000
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "../runtime/db_vm.h"
#include "../runtime/db_text.h"
#include "../common/db_packet.h"

static uint8_t vmimage[VMIMAGE_SIZE] __attribute__((far));	

// Not defined but actually located in the gcc library!
int vsprintf( char *, const char *, va_list );
int vsnprintf(char *, size_t, const char *, va_list);

#ifdef BITMAP_GRAPHICS
unsigned char	g_VRAMBuffer[VRAM_BUF_SIZE] __attribute__((far));	
unsigned short 	g_PaletteMap[PALETTE_MAP_SIZE];
unsigned char	g_Palettes[MIN_PALETTE_SIZE];
#define TARGET_TYPE TARGET_BITMAP
#endif

#ifdef TILE_GRAPHICS
#define TARGET_TYPE TARGET_TILE
#endif

Interpreter interp;

int main(void)
{
    Interpreter *i = &interp;

    SYS_ConfigureClock(FCY_NTSC);

    FSInit();

    PktOpen(NULL);
    VM_printf("XGS Basic Runtime 1.0\n");

    Gamepad_Init();
    SND_Init();

#ifdef BITMAP_GRAPHICS
    GFX_InitBitmap(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP, g_VRAMBuffer, g_Palettes, g_PaletteMap);
    GFX_SetClipRegion(0, SCREEN_HEIGHT - 1, 0, SCREEN_WIDTH - 1);
    
    memset(g_PaletteMap, 0, sizeof(g_PaletteMap));
    
    g_Palettes[0] = NTSC_BLACK;
    g_Palettes[1] = NTSC_RED;
    g_Palettes[2] = NTSC_GREEN;
    g_Palettes[3] = NTSC_BLUE;

    GFX_StartDrawing(SCREEN_TYPE_NTSC);
#endif

    InitInterpreter(i, TARGET_TYPE);
    InitFcn(i);

    if (!LoadImage(i, "autorun.bai")) {
#ifdef BITMAP_GRAPHICS
        DrawBitmapFontString("XGS Basic Runtime 1.0", 0, 20, g_VRAMBuffer);
        DrawBitmapFontString("load failed: autorun.bai", 0, 40, g_VRAMBuffer);
#else
        VM_printf("load failed: autorun.bai\n");
#endif
    }   

    for (;;) {
        if (i->imageLoaded)
            Execute(i);
        else
            PktPoll(0);
    }

    return 0;
}

/* VM_LoadMemoryImageHeader - load the header of an image already in memory */
int VM_LoadMemoryImageHeader(Interpreter *i)
{
    return VMFALSE;
}

/* VM_StoreMemoryImageHeader - store the header of an image in memory */
void VM_StoreMemoryImageHeader(Interpreter *i)
{
}

/* VM_AllocateImage - allocate space for an image */
int VM_AllocateImage(Interpreter *i, int codeSize, int initSize, int dataSize)
{
    /* make sure there is enough memory */
    if (codeSize + dataSize > VMIMAGE_SIZE)
        return VMFALSE;

    /* setup the code and data space (code last in case it is an odd number of bytes) */
    i->variables = (int16_t *)vmimage;
    i->code = vmimage + dataSize;

    /* return successfully */
    return VMTRUE;
}

/* VM_FreeImage - free an image */
void VM_FreeImage(Interpreter *i)
{
}

/* VM_LoadCode - load code into flash */
void VM_LoadCode(uint16_t dstoff, uint8_t *src, int cnt)
{
    memcpy(interp.code + dstoff, src, cnt);
}

/* VM_LoadInit - load initialization data into flash */
void VM_LoadInit(uint16_t dstoff, uint8_t *src, int cnt)
{
    /* don't need to save initialization data because we can't restart */
}

/* VM_ReadInit - read initialization data from flash */
void VM_ReadInit(uint16_t off, uint8_t *buf, int cnt)
{
    /* no saved initialization data to load */
}

void VM_printf(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_vprintf(fmt, ap);
    va_end(ap);
}

void VM_vprintf(char *fmt, va_list ap)
{
    char buf[128], *p;
    vsnprintf(buf, sizeof(buf), fmt, ap);
    for (p = buf; *p != '\0'; ++p)
        VM_putchar(*p);
}

static char debugBuf[16];
static int debugLen = 0;

static void DebugBufferCh(int ch)
{
    if (debugLen < sizeof(debugBuf)) {
        debugBuf[debugLen++] = ch;
        if (debugLen == sizeof(debugBuf))
            VM_flush();
    }
}

void VM_putchar(int ch)
{
    if (ch == '\n')
        DebugBufferCh('\r');
    DebugBufferCh(ch);
    if (ch == '\n')
        VM_flush();
}

void VM_flush(void)
{
    if (debugLen > 0) {
        PktSendFrame(PKT_DEBUG, (uint8_t *)debugBuf, debugLen);
        debugLen = 0;
    }
}

VMFILE *VM_fopen(const char *name, const char *mode)
{
    return (VMFILE *)FSfopen(name, mode);
}

int VM_fclose(VMFILE *fp)
{
    return FSfclose((FSFILE *)fp);
}

size_t VM_fread(void *buf, size_t count, VMFILE *fp)
{
    return FSfread(buf, 1, count, (FSFILE *)fp);
}

size_t VM_fwrite(const void *buf, size_t count, VMFILE *fp)
{
    return FSfwrite(buf, 1, count, (FSFILE *)fp);
}

int VM_fseek(VMFILE *fp, long offset, int whence)
{
    return FSfseek((FSFILE *)fp, offset, whence);
}

long VM_ftell(VMFILE *fp)
{
    return FSftell((FSFILE *)fp);
}

void VM_poll(void)
{
    PktPoll(0);
}

