/* db_compiler.c - a simple basic compiler
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <setjmp.h>
#include <string.h>
#include <ctype.h>
#include "db_compiler.h"
#include "../common/db_config.h"

/* error target */
jmp_buf errorTarget;

/* heap block header */
typedef struct HeapHdr HeapHdr;
struct HeapHdr {
    char *tag;
    HeapHdr *prev;
    HeapHdr *next;
    size_t size;
};

/* list of currently allocated blocks */
static HeapHdr *heapBlocks = NULL;

/* number of allocated bytes */
static size_t heapSize = 0;

/* local function prototypes */
static void PlaceStrings(ParseContext *c, int offset);
static void WriteStrings(ParseContext *c);
static void FreeStrings(ParseContext *c);
static void CreateOutputFile(ParseContext *c, char *name);
static void WriteOutputFile(ParseContext *c);

/* CompileFile - compile a source file */
int CompileFile(const char *infile, char *outfile, BoardType boardType, int debugMode)
{
    ParseContext c;

    /* initialize the parse context */
    InitParseContext(&c);

    /* setup an error target */
    if (setjmp(errorTarget) != 0) {
        CloseParseContext(&c);
        FreeAll();
        return VMFALSE;
    }

    /* create the output file */
    CreateOutputFile(&c, outfile);

    /* initialize the global symbol table */
    InitSymbolTable(&c.globals);

    /* target is unknown */
    switch (boardType) {
	case BOARD_UNKNOWN:
		c.hdr.target = TARGET_UNKNOWN;
		break;
	case BOARD_XGS_AVR:
	case BOARD_XGS_PIC:
		SelectRuntime(&c, "xgs");
		break;
	case BOARD_CHAMELEON_AVR:
	case BOARD_CHAMELEON_PIC:
		SelectRuntime(&c, "chameleon");
		break;
	default:
		return VMFALSE;
	}

    /* remember whether we're in debug mode */
    c.debugMode = debugMode;

    /* initialize the type table */
    InitTypeTable(&c.types);
    AddType(&c.types, "INTEGER",    &c.integerType);
    AddType(&c.types, "BYTE",       &c.byteType);
    AddType(&c.types, "STRING",     &c.stringType);
    
    /* add the intrinsic functions */
    AddIntrinsic(&c, "ABS",             "i",        "i",    FN_ABS);
    AddIntrinsic(&c, "RND",             "i",        "i",    FN_RND);
    AddIntrinsic(&c, "LEFT$",           "si",       "s",    FN_LEFT);
    AddIntrinsic(&c, "RIGHT$",          "si",       "s",    FN_RIGHT);
    AddIntrinsic(&c, "MID$",            "sii",      "s",    FN_MID);
    AddIntrinsic(&c, "CHR$",            "i",        "s",    FN_CHR);
    AddIntrinsic(&c, "STR$",            "i",        "s",    FN_STR);
    AddIntrinsic(&c, "VAL",             "s",        "i",    FN_VAL);
    AddIntrinsic(&c, "ASC",             "s",        "i",    FN_ASC);
    AddIntrinsic(&c, "LEN",             "s",        "i",    FN_LEN);
    
    /* open the input file */
    if (!PushFile(&c, infile))
        Fatal("opening '%s'", infile);

    /* get the next line */
    while (GetLine(&c)) {
        Token tkn;
        if ((tkn = GetToken(&c)) != T_EOL)
            ParseStatement(&c, tkn);
    }

    /* default the string heap size if it wasn't already set */
    if (!c.heapSizeSet) {
        c.hdr.heapSize = DEFAULT_HEAP_SIZE;
        c.hdr.heapHandles = DEFAULT_HEAP_HANDLES;
    }

    /* write the output file */
    WriteOutputFile(&c);

    /* make sure a target was defined */
    if (c.hdr.target == TARGET_UNKNOWN)
        Fatal("target unknown, please add an OPTION target = [ tile | bitmap ] to your source file");

    /* free the named types */
    FreeTypes(&c, &c.types);
    
    /* free the global symbol table */
    FreeSymbols(&c, &c.globals);

    /* free the string literals */
    FreeStrings(&c);

    /* close the parse context */
    CloseParseContext(&c);

    /* report any memory leaks */
    if (heapSize > 0) {
        VM_printf("warning: memory leak of %d bytes\n", heapSize);
        DumpAll();
    }

    /* return successfully */
    return VMTRUE;
}

/* AddString - add a string to the string table */
String *AddString(ParseContext *c, char *value)
{
    String *str;
    
    /* check to see if the string is already in the table */
    for (str = c->strings; str != NULL; str = str->next)
        if (strcmp(value, str->value) == 0)
            return str;

    /* allocate the string structure */
    str = (String *)Alloc("STR", sizeof(String) + strlen(value));
    memset(str, 0, sizeof(String));
    strcpy(str->value, value);
    str->next = c->strings;
    c->strings = str;

    /* return the string table entry */
    return str;
}

/* AddStringRef - add a reference to a string in the string table */
int AddStringRef(String *str, int offset)
{
    int link;

    /* handle strings that have already been placed */
    if (str->placed)
        return str->offset;

    /* add a new entry to the fixup list */
    link = str->fixups;
    str->fixups = offset;
    return link;
}

/* PlaceStrings - place any string literals defined in the current function */
static void PlaceStrings(ParseContext *c, int offset)
{
    String *str;
    for (str = c->strings; str != NULL; str = str->next) {
        if (str->fixups) {
            int size = (int)strlen(str->value) + 1;
            str->placed = VMTRUE;
            str->offset = 0x8000 | offset;
            fixup(c, str->fixups, str->offset);
            offset += size;
        }
    }
}

/* WriteStrings - write any string literals defined in the current function */
static void WriteStrings(ParseContext *c)
{
    String *str, *next;
    for (str = c->strings; str != NULL; str = next) {
        next = str->next;
        if (str->placed) {
            int size = (int)strlen(str->value) + 1;
            fwrite(str->value, size, 1, c->ofp);
            c->codeOff += size;
            str->fixups = 0;
        }
    }
}

/* FreeStrings - free the string literals */
static void FreeStrings(ParseContext *c)
{
    String *str, *next;
    for (str = c->strings; str != NULL; str = next) {
        next = str->next;
        Free(str);
    }
}

/* AddConstant - add a constant to the global symbol table */
void AddConstant(ParseContext *c, char *name, int16_t value)
{
    AddSymbol(&c->globals, name, SC_CONSTANT, &c->integerType, 0, value);
}

/* AddIntrinsic - add an intrinsic function to the global symbol table */
void AddIntrinsic(ParseContext *c, char *name, char *argTypes, char *retType, int index)
{
    Type *type, *argType;

    /* create a type for the function */
    type = NewType(TYPE_FUNCTION);
    InitSymbolTable(&type->FunctionInfo.arguments);

    /* set the return type */
    if (retType) {
        switch (*retType) {
        case 'i':
            type->FunctionInfo.returnType = &c->integerType;
            break;
        case 'b':
            type->FunctionInfo.returnType = &c->byteType;
            break;
        case 's':
            type->FunctionInfo.returnType = &c->stringType;
            break;
        default:
            Fatal("unknown return type: %s", retType);
            break;
        }
    }
    else
        type->FunctionInfo.returnType = NULL;


    /* add a dummy symbol for the return value */
    if (retType) {
        if (!AddSymbol(&type->FunctionInfo.arguments, name, SC_VARIABLE, NULL, 0, 0))
            Fatal("insufficient memory");
    }

    /* add the argument types */
    if (argTypes) {
        while (*argTypes) {
            switch (*argTypes++) {
            case 'i':
                argType = &c->integerType;
                break;
            case 'I':
                argType = &c->integerArrayType;
                break;
            case 'b':
                argType = &c->byteType;
                break;
            case 'B':
                argType = &c->byteArrayType;
                break;
            case 's':
                argType = &c->stringType;
                break;
            default:
                Fatal("unknown argument type: %c", argTypes[-1]);
                break;
            }
            if (!AddSymbol(&type->FunctionInfo.arguments, "", SC_VARIABLE, argType, 0, 0))
                Fatal("insufficient memory");
        }
    }

    /* add a global symbol for the intrinsic function */
    if (!AddSymbol(&c->globals, name, SC_CONSTANT, type, 0, (index << 1) | 1))
        Fatal("insufficient memory");
}

/* CreateOutputFile - create the output file */
static void CreateOutputFile(ParseContext *c, char *name)
{
    /* create the output file */
    if (!(c->ofp = fopen(name, "wb")))
        Fatal("creating output file: %s", name);

    /* skip past where the header will be */
    fseek(c->ofp, sizeof(FileHdr), SEEK_SET);
}

/* WriteFunction - write byte for a function to the output file */
void WriteFunction(ParseContext *c, char *name)
{
    size_t codeSize = c->cptr - c->codeBuf;

    /* make sure all referenced labels were defined */
    CheckLabelsAndFree(c);

    /* place string literals defined in this function */
    PlaceStrings(c, (int)c->codeOff + codeSize);
    
    if (c->debugMode) {
        printf("%s:\n", name);
        DecodeFunction(c->codeOff, c->codeBuf, (int)codeSize);
    }

    /* write the procedure to the output file */
    if (fwrite(c->codeBuf, 1, codeSize, c->ofp) != codeSize)
        Fatal("error writing function to output file");

    /* update the offset into the code area */
    c->codeOff += codeSize;

    /* write the string literals defined in this function */
    WriteStrings(c);
    
    /* restore the pointers to the main code */
    c->cptr = c->codeBuf;
}

/* WriteOutputFile - write the output file */
static void WriteOutputFile(ParseContext *c)
{
    uint16_t mainCodeOffset, fileOffset;
    int i;

    /* end the main code with a halt */
    putcbyte(c, OP_HALT);
    
    /* write the main code */
    mainCodeOffset = c->codeOff;
    WriteFunction(c, "main");

    /* setup the file header structure */
    strcpy(c->hdr.tag, FILE_TAG);
    c->hdr.version = FILE_VERSION;
    c->hdr.codeOffset = sizeof(FileHdr);
    c->hdr.codeSize = c->codeOff;
    c->hdr.mainCodeOffset = mainCodeOffset;
    c->hdr.initOffset = c->hdr.codeOffset + ((c->hdr.codeSize + 1) & ~1);
    c->hdr.initSize = 0;
    c->hdr.variableCount = c->dataOff;
    c->hdr.stackSize = 100;

    /* setup the file offset */
    fileOffset = c->hdr.codeOffset + ((c->hdr.codeSize + 1) & ~1);

    /* setup the init section */
    c->hdr.initOffset = fileOffset;

    /* write the initializers */
    fseek(c->ofp, (long)c->hdr.initOffset, SEEK_SET);
    for (i = 0; i < c->dataOff; ) {
        uint16_t offset, count;

        /* find the next non-zero value */
        for (; i < c->dataOff; ++i)
            if (c->dataBuf[i] != 0)
                break;

        /* remember the start of the non-zero block */
        offset = i;

        /* find the next zero value */
        for (; i < c->dataOff; ++i)
            if (c->dataBuf[i] == 0)
                break;

        /* determine the size of the block */
        count = i - offset;
        
        /* write the block if it is not empty */
        if (count > 0) {

            /* write the block header */
            fwrite(&offset, sizeof(uint16_t), 1, c->ofp);
            fwrite(&count, sizeof(uint16_t), 1, c->ofp);
            fileOffset += 2 * sizeof(uint16_t);

            /* write the block data */
            while (count > 0) {
                fwrite(&c->dataBuf[offset], sizeof(uint16_t), 1, c->ofp);
                fileOffset += sizeof(int16_t);
                ++offset;
                --count;
            }
        }
    }

    /* setup the init section size */
    c->hdr.initSize = fileOffset - c->hdr.initOffset;

    /* write the output file header */
    fseek(c->ofp, 0, SEEK_SET);
    if (fwrite(&c->hdr, 1, sizeof(FileHdr), c->ofp) != sizeof(FileHdr))
        Fatal("error writing output file header");

    /* close the output file */
    fclose(c->ofp);

    if (c->debugMode)
        DumpSymbols(&c->globals, VMTRUE);

    {
        int variableSize = c->hdr.variableCount  * sizeof(int16_t);
        int stackSize = c->hdr.stackSize * sizeof(int16_t);
        int heapSize = c->hdr.heapSize ? sizeof(StringHeap) + c->hdr.heapSize + sizeof(uint16_t) * c->hdr.heapHandles : 0;
        int dataSize = variableSize + stackSize + heapSize;
        VM_printf("code size:     %d\n", c->hdr.codeSize);
        VM_printf("variable size: %d\n", variableSize);
        VM_printf("stack size:    %d\n", stackSize);
        VM_printf("heap size:     %d\n", heapSize); 
        VM_printf("total sram:    %d\n", dataSize);
    }
}

/* Alloc - allocate memory from the heap */
void *Alloc(char *tag, size_t size)
{
    HeapHdr *hdr;
    if (!(hdr = (HeapHdr *)malloc(sizeof(HeapHdr) + size)))
        Fatal("insufficient memory");
    heapSize += size;
    hdr->tag = tag;
    hdr->size = size;
    hdr->prev = NULL;
    hdr->next = heapBlocks;
    if (heapBlocks)
        heapBlocks->prev = hdr;
    heapBlocks = hdr;
    return (void *)(hdr + 1);
}

/* Free - free memory allocated from the heap */
void Free(void *p)
{
    HeapHdr *hdr = (HeapHdr *)((char *)p - sizeof(HeapHdr));
    heapSize -= hdr->size;
    if (hdr->prev)
        hdr->prev->next = hdr->next;
    else
        heapBlocks = hdr->next;
    if (hdr->next)
        hdr->next->prev = hdr->prev;
    free(hdr);
}

/* FreeAll - free all remaining allocated memory */
void FreeAll(void)
{
    HeapHdr *hdr, *next;
    for (hdr = heapBlocks; hdr != NULL; hdr = next) {
        next = hdr->next;
        heapSize -= hdr->size;
        free(hdr);
    }
    heapBlocks = NULL;
}

/* DumpAll - dump all allocated memory */
void DumpAll(void)
{
    HeapHdr *hdr;
    for (hdr = heapBlocks; hdr != NULL; hdr = hdr->next)
        printf("  %s %d\n", hdr->tag, hdr->size);
}

/* Fatal - report a fatal error and exit */
void Fatal(char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_printf("error: ");
    VM_vprintf(fmt, ap);
    VM_putchar('\n');
    va_end(ap);
    longjmp(errorTarget, 1);
}
