/* db_vm.h - definitions for a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#ifndef __DB_VM_H__
#define __DB_VM_H__

#include <stdio.h>
#include <stdarg.h>
#include <setjmp.h>
#include "db_vmimage.h"

/* forward type declarations */
typedef struct Interpreter Interpreter;

/* the only global used by the interpreter */
extern Interpreter interp;

/* intrinsic function handler type */
typedef void IntrinsicFcn(Interpreter *i);

/* number of input devices */
#define MAX_IN_DEV      1   /* just terminal for now */

/* number of output devices */
#define MAX_OUT_DEV     2   /* just terminal and display for now */

/* input device structure */
typedef struct {
    char inputBuf[81];
    char *inputPtr;
} InputDev;

/* output device structure */
typedef struct {
    uint16_t linePos;
} OutputDev;

/* interpreter state structure */
struct Interpreter {
    const uint8_t *image;
    const uint8_t *code;
    const uint8_t *main;
    const uint8_t *pc;
    uint8_t *dataBuf;
    uint16_t dataBufSize;
    uint8_t *freeSpace;
    uint16_t freeSpaceSize;
    jmp_buf errorTarget;
    int16_t *variables;
    int16_t *stack;
    int16_t *stackTop;
    int16_t *fp;
    int16_t *sp;
    StringHeap *heap;
    InputDev inputDevs[MAX_IN_DEV];
    OutputDev outputDevs[MAX_OUT_DEV];
    int target;
};

/* stack manipulation macros */
#define Reserve(i, n)   do {                                    \
                            if ((i)->sp + (n) > (i)->stackTop)  \
                                StackOverflow(i);               \
                            else  {                             \
                                int cnt = (n);                  \
                                while (--cnt >= 0)              \
                                    Push(i, 0);                 \
                            }                                   \
                        } while (0)
#define CPush(i, v)     do {                                    \
                            if ((i)->sp + 1 >= (i)->stackTop)   \
                                StackOverflow(i);               \
                            else                                \
                                Push(i, v);                     \
                        } while (0)
#define Push(i, v)      (*++(i)->sp = (v))
#define Pop(i)          (*(i)->sp--)
#define Drop(i, n)      ((i)->sp -= (n))

/* prototypes */
void StackOverflow(Interpreter *i);

extern FLASH_SPACE char str_file_tag[];
extern FLASH_SPACE char str_string_index_range_err[];
extern FLASH_SPACE char str_no_heap_err[];
extern FLASH_SPACE char str_no_heap_handles_err[];
extern FLASH_SPACE char str_no_heap_space_err[];
extern FLASH_SPACE char str_heap_debug_output[];
extern FLASH_SPACE char str_heap_debug_output_tail[];
extern FLASH_SPACE char str_array_subscript_err[];
extern FLASH_SPACE char str_array_subscript1_err[];
extern FLASH_SPACE char str_array_subscript2_err[];
extern FLASH_SPACE char str_image_header_err[];
extern FLASH_SPACE char str_image_target_err[]; // (unused)
extern FLASH_SPACE char str_data_space_err[];
extern FLASH_SPACE char str_opcode_err[];
extern FLASH_SPACE char str_stack_overflow_err[];
extern FLASH_SPACE char str_warn_prefix[];
extern FLASH_SPACE char str_abort_prefix[];
extern FLASH_SPACE char str_program_id[];

/* prototypes and variables from db_vmfcn.c */
extern IntrinsicFcn * FLASH_SPACE Intrinsics[];

/* prototypes from db_vmimage.c  */
int LoadImageFile(Interpreter *i, char *name);

/* prototypes from db_vmint.c */
int InitInterpreter(Interpreter *i, const uint8_t *image, uint8_t *dataBuf, uint16_t dataBufSize);
int Execute(Interpreter *i);
void ShowStack(Interpreter *i);
void Warn(const char *fmt, ...);                    /* fmt in FLASH_SPACE */
void Abort(Interpreter *i, const char *fmt, ...);   /* fmt in FLASH_SPACE */

/* prototypes from db_vmheap.c */
void InitHeap(Interpreter *i, uint16_t size, uint16_t nHandles);
int16_t StringAlloc(Interpreter *i, uint16_t length);
void StringAddRef(Interpreter *i, int16_t handle);
void StringRelease(Interpreter *i, int16_t handle);
uint8_t *GetStringPtr(Interpreter *i, int16_t value, int *pInHeap);
int GetStringLen(Interpreter *i, int16_t value);
void CompactHeap(Interpreter *i);
void DumpHeap(Interpreter *i);

/* functions provided by target platform */
void VM_sysinit(int argc, char *argv[], const uint8_t **image, uint8_t **data, uint16_t *dataSize);
void VM_fcninit(void);
void VM_fcnstop(void);
void VM_displaychar(int ch);

/* platform specific built in functions */
#ifdef XGS_COMMON
void fcn_waitforvsync(Interpreter *i);
void fcn_gamepad(Interpreter *i);
void fcn_sound(Interpreter *i);
void fcn_plot(Interpreter *i);
void fcn_line(Interpreter *i);
void fcn_hline(Interpreter *i);
void fcn_vline(Interpreter *i);
void fcn_clear(Interpreter *i);
void fcn_fillrect(Interpreter *i);
void fcn_circle(Interpreter *i);
void fcn_position(Interpreter *i);
void fcn_text(Interpreter *i);
void fcn_settilemap(Interpreter *i);
void fcn_settiles(Interpreter *i);
void fcn_setcolortable(Interpreter *i);
void fcn_setsprite(Interpreter *i);
#endif
#ifdef CHAMELEON_COMMON
void fcn_prop(Interpreter *i);
void fcn_prop_8(Interpreter *i);
void fcn_prop_8_8(Interpreter *i);
void fcn_prop_16(Interpreter *i);
void fcn_prop_result(Interpreter *i);
void fcn_define_tile_map(Interpreter *i);
void fcn_display_tile_map(Interpreter *i);
void fcn_set_term_tile_map(Interpreter *i);
void fcn_load_tile(Interpreter *i);
void fcn_load_palette(Interpreter *i);
void fcn_write_tile(Interpreter *i);
void fcn_read_tile(Interpreter *i);
#endif

#endif
