/* db_vmimage.h - definitions for a simple virtual machine
 *
 * Copyright (c) 2009-2010 by David Michael Betz.  All rights reserved.
 *
 */

#ifndef __DB_VMIMAGE_H__
#define __DB_VMIMAGE_H__

#include "vmlib.h"

/* booleans */
#define VMTRUE		1
#define VMFALSE   	0

/* image file signature */
#define FILE_TAG        "DMBASIC"
#define FILE_VERSION    0x0200

/* target runtimes */
#define TARGET_UNKNOWN      0
#define TARGET_XGS          1
#define TARGET_CHAMELEON    2

/* image file header */
typedef struct {
    char        tag[8];         /* must be 'DMBASIC' (tag[7] must be '\0') */
    uint16_t    version;        /* must be equal to FILE_VERSION above */
    uint8_t     target;         /* target runtime */
    uint8_t     unused;         /* (unused) */
    uint16_t    codeOffset;     /* offset to the bytecodes */
    uint16_t    codeSize;       /* size of bytecodes in bytes */
    uint16_t    mainCodeOffset; /* offset to the main code */
    uint16_t    initOffset;     /* offset to the initialization table */
    uint16_t    initSize;       /* size of the initialization table in bytes */
    uint16_t    variableCount;  /* number of variables */
    uint16_t    stackSize;      /* stack size */
    uint16_t    heapSize;       /* string heap size */
    uint16_t    heapHandles;    /* number of heap handles */
} FileHdr;

/* string handle structure */
typedef struct {
    uint8_t *data;
} StringHandle;

/* string header structure */
typedef struct {
    StringHandle *handle;
    uint16_t refCnt;
    uint16_t length;
} StringHdr;

/* macro to get a pointer to an object in the string heap */
#define GetHeapStringPtr(i, h)  (((i)->heap->endHandles - (h))->data)
#define GetHeapStringLen(i, h)  (((StringHdr *)GetHeapStringPtr(i, h) - 1)->length)

/* string heap structure */
typedef struct {
    int nHandles;               /* number of handles */
    StringHandle *handles;      /* array of handles */
    StringHandle *endHandles;   /* end of the array of handles */
    StringHandle *freeHandles;  /* list of free handles */
    uint8_t *data;              /* heap data */
    uint8_t *free;              /* next free heap location */
} StringHeap;

/* opcodes */
#define OP_HALT         0x00    /* halt */
#define OP_BRT          0x01    /* branch on true */
#define OP_BRTSC        0x02    /* branch on true (for short circuit booleans) */
#define OP_BRF          0x03    /* branch on false */
#define OP_BRFSC        0x04    /* branch on false (for short circuit booleans) */
#define OP_BR           0x05    /* branch unconditionally */
#define OP_NOT          0x06    /* logical negate top of stack */
#define OP_NEG          0x07    /* negate */
#define OP_ADD          0x08    /* add two numeric expressions */
#define OP_SUB          0x09    /* subtract two numeric expressions */
#define OP_MUL          0x0a    /* multiply two numeric expressions */
#define OP_DIV          0x0b    /* divide two numeric expressions */
#define OP_REM          0x0c    /* remainder of two numeric expressions */
#define OP_BNOT         0x0d    /* bitwise not of two numeric expressions */
#define OP_BAND         0x0e    /* bitwise and of two numeric expressions */
#define OP_BOR          0x0f    /* bitwise or of two numeric expressions */
#define OP_BXOR         0x10    /* bitwise exclusive or */
#define OP_SHL          0x11    /* shift left */
#define OP_SHR          0x12    /* shift right */
#define OP_LT           0x13    /* less than */
#define OP_LE           0x14    /* less than or equal to */
#define OP_EQ           0x15    /* equal to */
#define OP_NE           0x16    /* not equal to */
#define OP_GE           0x17    /* greater than or equal to */
#define OP_GT           0x18    /* greater than */
#define OP_CAT          0x19    /* concatenate two strings */
#define OP_LTS          0x1a    /* string less than */
#define OP_LES          0x1b    /* string less than or equal to */
#define OP_EQS          0x1c    /* string equal to */
#define OP_NES          0x1d    /* string not equal to */
#define OP_GES          0x1e    /* string greater than or equal to */
#define OP_GTS          0x1f    /* string greater than */
#define OP_LIT          0x20    /* load literal */
#define OP_GREF         0x21    /* load a global variable */
#define OP_GSET         0x22    /* set a global variable */
#define OP_GREFB        0x23    /* load a global byte variable */
#define OP_GSETB        0x24    /* set a global byte variable */
#define OP_GREFS        0x25    /* load a global string variable */
#define OP_GSETS        0x26    /* set a global string variable */
#define OP_LOFF         0x27    /* load the offset into variable space for a local variable */
#define OP_LREF         0x28    /* load a local variable relative to the frame pointer */
#define OP_LSET         0x29    /* set a local variable relative to the frame pointer */
#define OP_LREFB        0x2a    /* load a local byte variable relative to the frame pointer */
#define OP_LSETB        0x2b    /* set a local byte variable relative to the frame pointer */
#define OP_LREFS        0x2c    /* load a local string variable relative to the frame pointer */
#define OP_LSETS        0x2d    /* set a local string variable relative to the frame pointer */
#define OP_LDECREF      0x2e    /* decrement the reference count of a local variable */
#define OP_AREF1        0x2f    /* load an element of an array */
#define OP_ASET1        0x30    /* set an element of an array */
#define OP_AREF2        0x31    /* load an element of a two-dimensional array */
#define OP_ASET2        0x32    /* set an element of a two-dimensional array */
#define OP_AREF1B       0x33    /* load a byte element of an array */
#define OP_ASET1B       0x34    /* set a byte element of an array */
#define OP_AREF2B       0x35    /* load a byte element of a two-dimensional array */
#define OP_ASET2B       0x36    /* set a byte element of a two-dimensional array */
#define OP_AREF1S       0x37    /* load a string element of an array */
#define OP_ASET1S       0x38    /* set a string element of an array */
#define OP_AREF2S       0x39    /* load a string element of a two-dimensional array */
#define OP_ASET2S       0x3a    /* set a string element of a two-dimensional array */
#define OP_RESERVE      0x3b    /* reserve space on the stack */
#define OP_CALL         0x3c    /* call a function */
#define OP_FRET         0x3d    /* return from a function leaving an integer result on the stack */
#define OP_SRET         0x3e    /* return from a subroutine */
#define OP_DUP          0x3f    /* duplicate the top element of the stack */
#define OP_DROP         0x40    /* drop the top element of the stack */
#define OP_PRINTI       0x41    /* print an integer */
#define OP_PRINTS       0x42    /* print a string */
#define OP_PRINTT       0x43    /* print a tab */
#define OP_PRINTN       0x44    /* print a newline */
#define OP_GETLINE      0x45    /* get a new input line */
#define OP_INPUTI       0x46    /* input an integer */
#define OP_INPUTS       0x47    /* input a string */

/* stack frame offsets */
#define F_FP    1
#define F_PC    2
#define F_SIZE  2

/* intrinsic function numbers */
enum {
    FN_ABS              = 0x00,
    FN_RND,
    FN_LEFT,
    FN_RIGHT,
    FN_MID,
    FN_CHR,
    FN_STR,
    FN_VAL,
    FN_ASC,
    FN_LEN,

    /* target: xgs */
    FN_WAITFORVSYNC = FN_LEN + 1,
    FN_GAMEPAD,
    FN_SOUND,
    FN_PLOT,
    FN_LINE,
    FN_HLINE,
    FN_VLINE,
    FN_CLEAR,
    FN_FILLRECT,
    FN_CIRCLE,
    FN_POSITION,
    FN_TEXT,
    FN_SETTILEMAP,
    FN_SETTILES,
    FN_SETCOLORTABLE,
    FN_SETSPRITE,

    /* target: chameleon */
    FN_PROP = FN_LEN + 1,
    FN_PROP_8,
    FN_PROP_8_8,
    FN_PROP_16,
    FN_PROP_RESULT,
    FN_DEFINE_TILE_MAP,
    FN_DISPLAY_TILE_MAP,
    FN_SET_TERM_TILE_MAP,
    FN_LOAD_TILE,
    FN_LOAD_PALETTE,
    FN_WRITE_TILE,
    FN_READ_TILE
};

/* prototypes from db_vmdebug.c (used by both compiler and runtime) */
void DecodeFunction(uint16_t base, const uint8_t *code, int len);
int DecodeInstruction(uint16_t base, const uint8_t *code, const uint8_t *lc);

#endif
