/* db_vmint.c - bytecode interpreter for a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "../common/db_config.h"
#include "db_vm.h"

/* initializer processing states */
#define INIT_HDR1   1
#define INIT_HDR2   2
#define INIT_DATA   3

/* macro to index a one dimensional array */
#define INDEX1                                                  \
            do {                                                \
                ind1 = Pop(i);                                  \
                tmp = i->sp[0];                                 \
                if (ind1 < 1 || ind1 > i->variables[tmp - 1])   \
                    Abort(i, str_array_subscript_err, ind1);    \
            } while (0)

/* macro to index a two dimensional array */
#define INDEX2                                                  \
            do {                                                \
                ind2 = Pop(i);                                  \
                ind1 = Pop(i);                                  \
                tmp = i->sp[0];                                 \
                if (ind1 < 1 || ind1 > i->variables[tmp - 2])   \
                    Abort(i, str_array_subscript1_err, ind1);   \
                ind1 += (ind2 - 1) * i->variables[tmp - 2];     \
                if (ind1 < 1 || ind1 > i->variables[tmp - 1])   \
                    Abort(i, str_array_subscript2_err, ind2);   \
            } while (0)

/* prototypes for local functions */
static int VerifyAndSetupImageHeader(Interpreter *i, FileHdr *hdr);
static int AllocateDataSpace(Interpreter *i, uint16_t dataSize, uint16_t heapSize);
static void StringCat(Interpreter *i);
static int StringCmp(Interpreter *i);
static void PrintInteger(Interpreter *i, int dev, int16_t value);
static void PrintString(Interpreter *i, int dev, int16_t value);
static void PrintTab(Interpreter *i, int dev);
static void PrintNewline(Interpreter *i, int dev);
static void InputGetLine(Interpreter *i, int dev);
static void InputInteger(Interpreter *i, int dev);
static void InputString(Interpreter *i, int dev);

/* InitInterpreter - initialize the interpreter */
int InitInterpreter(Interpreter *i, const uint8_t *image, uint8_t *dataBuf, uint16_t dataBufSize)
{
    int init_state;
    uint16_t init_off, init_cnt, off, cnt;
    uint8_t buf[128];
    FileHdr hdr;
    
    /* setup the new image */
    memset(i, 0, sizeof(Interpreter));
    i->target = RUNTIME_TARGET;
    i->image = image;
    i->dataBuf = dataBuf;
    i->dataBufSize = dataBufSize;

    /* get the image header */
    VMCODECOPY((uint8_t *)&hdr, i->image, sizeof(hdr));

    /* verify and setup the header */
    if (!VerifyAndSetupImageHeader(i, &hdr))
        return VMFALSE;

    /* process the initialization data */
    for (init_state = INIT_HDR1, init_off = init_cnt = off = 0; off < hdr.initSize; off += cnt) {
        uint8_t *p, *end;

        /* determine how much initializer data to read */
        if ((cnt = hdr.initSize - off) > sizeof(buf))
			cnt = sizeof(buf);

        /* read the next block of initializer data */
        VMCODECOPY(buf, image + hdr.initOffset + off, cnt);

        /* process the initializer data */
        for (p = buf, end = buf + cnt; p < end; ) {
            switch (init_state) {
            case INIT_HDR1:
                init_off = *(uint16_t *)p;
                init_state = INIT_HDR2;
                p += sizeof(uint16_t);
                break;
            case INIT_HDR2:
                init_cnt = *(uint16_t *)p;
                init_state = INIT_DATA;
                p += sizeof(uint16_t);
                break;
            case INIT_DATA:
                i->variables[init_off++] = *(int16_t *)p;
                p += sizeof(int16_t);
                if (--init_cnt == 0)
                    init_state = INIT_HDR1;
                break;
            }
        }
    }

    /* return successfully */
    return VMTRUE;
}

/* VerifyAndSetupImageHeader - verify and setup an image header */
static int VerifyAndSetupImageHeader(Interpreter *i, FileHdr *hdr)
{
    uint16_t dataSize, heapSize;

    /* verify the image header */
    if (VMSTRNCMP(hdr->tag, str_file_tag, sizeof(FILE_TAG) - 1) != 0 || hdr->version != FILE_VERSION) {
        Warn(str_image_header_err);
        return VMFALSE;
    }

    /* verify the target */
    if (hdr->target != i->target) {
        Warn(str_image_target_err, i->target, hdr->target);
        return VMFALSE;
    }

    /* setup the code space */
    i->code = i->image + hdr->codeOffset;
    i->main = i->code + hdr->mainCodeOffset;

    /* allocate the variable and heap space */
    dataSize = (hdr->variableCount + hdr->stackSize) * sizeof(int16_t);
    heapSize = hdr->heapSize ? sizeof(StringHeap) + hdr->heapSize + hdr->heapHandles * sizeof(void *) : 0;
    if (!AllocateDataSpace(i, dataSize, heapSize)) {
        Warn(str_data_space_err, dataSize, heapSize);
        return VMFALSE;
    }

    /* clear the variable space */
    memset(i->variables, 0, hdr->variableCount * sizeof(int16_t));

    /* setup the stack */
    i->stack = (int16_t *)(i->variables + hdr->variableCount);
    i->stackTop = i->stack + hdr->stackSize;

    /* setup the heap if one was configured */
    if (i->heap)
        InitHeap(i, hdr->heapSize, hdr->heapHandles);

    /* return successfully */
    return VMTRUE;
}

/* AllocateDataSpace - allocate space for an image */
static int AllocateDataSpace(Interpreter *i, uint16_t dataSize, uint16_t heapSize)
{
    /* make sure there is enough memory */
    if (dataSize + heapSize > i->dataBufSize)
        return VMFALSE;

    /* setup the variable space */
    i->variables = (int16_t *)i->dataBuf;

    /* setup the heap space */
    if (heapSize)
        i->heap = (StringHeap *)(i->dataBuf + dataSize);

    /* setup free data space */
    i->freeSpace = i->dataBuf + dataSize + heapSize;
    i->freeSpaceSize = i->dataBufSize - dataSize - heapSize;

    /* return successfully */
    return VMTRUE;
}

/* Execute - execute the main code */
int Execute(Interpreter *i)
{
    int16_t tmp, tmp2, ind1, ind2;
    int8_t tmpb;

    /* initialize */
    i->pc = i->main;
    i->sp = i->fp = i->stack - 1;
    VM_fcninit();

    if (setjmp(i->errorTarget)) {
        VM_fcnstop();
        return VMFALSE;
    }

    for (;;) {
#if 0
        ShowStack(i);
        DecodeInstruction(0, i->code, i->pc);
#endif
        switch (VMCODEBYTE(i->pc++)) {
        case OP_HALT:
            VM_fcnstop();
            return VMTRUE;
        case OP_BRT:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)(VMCODEBYTE(i->pc++) << 8);
            if (Pop(i))
                i->pc += tmp;
            break;
        case OP_BRTSC:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)(VMCODEBYTE(i->pc++) << 8);
            if (*i->sp)
                i->pc += tmp;
            else
                Drop(i, 1);
            break;
        case OP_BRF:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)(VMCODEBYTE(i->pc++) << 8);
            if (!Pop(i))
                i->pc += tmp;
            break;
        case OP_BRFSC:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)(VMCODEBYTE(i->pc++) << 8);
            if (!*i->sp)
                i->pc += tmp;
            else
                Drop(i, 1);
            break;
        case OP_BR:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)(VMCODEBYTE(i->pc++) << 8);
            i->pc += tmp;
            break;
        case OP_NOT:
            *i->sp = (*i->sp ? VMFALSE : VMTRUE);
            break;
        case OP_NEG:
            *i->sp = -*i->sp;
            break;
        case OP_ADD:
            tmp = Pop(i);
            *i->sp += tmp;
            break;
        case OP_SUB:
            tmp = Pop(i);
            *i->sp -= tmp;
            break;
        case OP_MUL:
            tmp = Pop(i);
            *i->sp *= tmp;
            break;
        case OP_DIV:
            tmp = Pop(i);
            *i->sp = (tmp == 0 ? 0 : *i->sp / tmp);
            break;
        case OP_REM:
            tmp = Pop(i);
            *i->sp = (tmp == 0 ? 0 : *i->sp % tmp);
            break;
        case OP_BNOT:
            *i->sp = ~*i->sp;
            break;
        case OP_BAND:
            tmp = Pop(i);
            *i->sp &= tmp;
            break;
        case OP_BOR:
            tmp = Pop(i);
            *i->sp |= tmp;
            break;
        case OP_BXOR:
            tmp = Pop(i);
            *i->sp ^= tmp;
            break;
        case OP_SHL:
            tmp = Pop(i);
            *i->sp <<= tmp;
            break;
        case OP_SHR:
            tmp = Pop(i);
            *i->sp >>= tmp;
            break;
        case OP_LT:
            tmp = Pop(i);
            *i->sp = (*i->sp < tmp ? VMTRUE : VMFALSE);
            break;
        case OP_LE:
            tmp = Pop(i);
            *i->sp = (*i->sp <= tmp ? VMTRUE : VMFALSE);
            break;
        case OP_EQ:
            tmp = Pop(i);
            *i->sp = (*i->sp == tmp ? VMTRUE : VMFALSE);
            break;
        case OP_NE:
            tmp = Pop(i);
            *i->sp = (*i->sp != tmp ? VMTRUE : VMFALSE);
            break;
        case OP_GE:
            tmp = Pop(i);
            *i->sp = (*i->sp >= tmp ? VMTRUE : VMFALSE);
            break;
        case OP_GT:
            tmp = Pop(i);
            *i->sp = (*i->sp > tmp ? VMTRUE : VMFALSE);
            break;
        case OP_CAT:
            StringCat(i);
            break;
        case OP_LTS:
            tmp = (StringCmp(i) < 0 ? VMTRUE : VMFALSE);
            *i->sp = tmp;
            break;
        case OP_LES:
            tmp = (StringCmp(i) <= 0 ? VMTRUE : VMFALSE);
            *i->sp = tmp;
            break;
        case OP_EQS:
            tmp = (StringCmp(i) == 0 ? VMTRUE : VMFALSE);
            *i->sp = tmp;
            break;
        case OP_NES:
            tmp = (StringCmp(i) != 0 ? VMTRUE : VMFALSE);
            *i->sp = tmp;
            break;
        case OP_GES:
            tmp = (StringCmp(i) >= 0 ? VMTRUE : VMFALSE);
            *i->sp = tmp;
            break;
        case OP_GTS:
            tmp = (StringCmp(i) > 0 ? VMTRUE : VMFALSE);
            *i->sp = tmp;
            break;
        case OP_LIT:
            tmp = VMCODEBYTE(i->pc++);
            CPush(i, (int16_t)(VMCODEBYTE(i->pc++) << 8) | tmp);
            break;
        case OP_GREF:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            CPush(i, i->variables[tmp]);
            break;
        case OP_GSET:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            i->variables[tmp] = Pop(i);
            break;
        case OP_GREFB:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            CPush(i, *(uint8_t *)&i->variables[tmp]);
            break;
        case OP_GSETB:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            *(uint8_t *)&i->variables[tmp] = (uint8_t)Pop(i);
            break;
        case OP_GREFS:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            StringAddRef(i, i->variables[tmp]);
            CPush(i, i->variables[tmp]);
            break;
        case OP_GSETS:
            tmp = VMCODEBYTE(i->pc++);
            tmp |= (int16_t)VMCODEBYTE(i->pc++) << 8;
            StringRelease(i, i->variables[tmp]);
            i->variables[tmp] = Pop(i);
            break;
        case OP_LOFF:
            tmp = VMCODEBYTE(i->pc++);
            CPush(i, (int16_t)(i->fp - i->variables) + tmp);
            break;
        case OP_LREF:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            CPush(i, i->fp[(int)tmpb]);
            break;
        case OP_LSET:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            i->fp[(int)tmpb] = Pop(i);
            break;
        case OP_LREFB:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            CPush(i, *(uint8_t *)&i->fp[(int)tmpb]);
            break;
        case OP_LSETB:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            *(uint8_t *)&i->fp[(int)tmpb] = (uint8_t)Pop(i);
            break;
        case OP_LREFS:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            StringAddRef(i, i->fp[(int)tmpb]);
            CPush(i, i->fp[(int)tmpb]);
            break;
        case OP_LSETS:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            StringRelease(i, i->fp[(int)tmpb]);
            i->fp[(int)tmpb] = Pop(i);
            break;
        case OP_LDECREF:
            tmpb = (int8_t)VMCODEBYTE(i->pc++);
            StringRelease(i, i->fp[(int)tmpb]);
            break;
        case OP_AREF1:
            INDEX1;
            i->sp[0] = i->variables[tmp + ind1 - 1];
            break;
        case OP_ASET1:
            tmp2 = Pop(i);
            INDEX1;
            i->variables[tmp + ind1 - 1] = tmp2;
            Drop(i, 1);
            break;
        case OP_AREF2:
            INDEX2;
            i->sp[0] = i->variables[tmp + ind1 - 1];
            break;
        case OP_ASET2:
            tmp2 = Pop(i);
            INDEX2;
            i->variables[tmp + ind1 - 1] = tmp2;
            Drop(i, 1);
            break;
        case OP_AREF1B:
            INDEX1;
            i->sp[0] = ((uint8_t *)&i->variables[tmp])[ind1 - 1];
            break;
        case OP_ASET1B:
            tmp2 = Pop(i);
            INDEX1;
            ((uint8_t *)&i->variables[tmp])[ind1 - 1] = (uint8_t)tmp2;
            Drop(i, 1);
            break;
        case OP_AREF2B:
            INDEX2;
            i->sp[0] = ((uint8_t *)&i->variables[tmp])[ind1 - 1];
            break;
        case OP_ASET2B:
            tmp2 = Pop(i);
            INDEX2;
            ((uint8_t *)&i->variables[tmp])[ind1 - 1] = (uint8_t)tmp2;
            Drop(i, 1);
            break;
        case OP_AREF1S:
            INDEX1;
            tmp = i->variables[tmp + ind1 - 1];
            i->sp[0] = tmp;
            StringAddRef(i, tmp);
            break;
        case OP_ASET1S:
            tmp2 = Pop(i);
            INDEX1;
            tmp += ind1 - 1;
            StringRelease(i, tmp);
            i->variables[tmp] = tmp2;
            Drop(i, 1);
            break;
        case OP_AREF2S:
            INDEX2;
            tmp = i->variables[tmp + ind1 - 1];
            i->sp[0] = tmp;
            StringAddRef(i, tmp);
            break;
        case OP_ASET2S:
            tmp2 = Pop(i);
            INDEX2;
            tmp += ind1 - 1;
            StringRelease(i, tmp);
            i->variables[tmp] = tmp2;
            Drop(i, 1);
            break;
        case OP_RESERVE:
            tmp = VMCODEBYTE(i->pc++);
            Reserve(i, tmp);
            break;
        case OP_CALL:
            tmp = i->sp[-(int16_t)VMCODEBYTE(i->pc++)];
            if (tmp & 1)
                (*VMINTRINSIC(tmp >> 1))(i);
            else {
                tmp2 = (int16_t)(i->fp - i->stack);
                i->fp = i->sp;
                Reserve(i, F_SIZE);
                i->fp[F_FP] = tmp2;
                i->fp[F_PC] = (int16_t)(i->pc - i->code);
                i->pc = i->code + tmp;
            }
            break;
        case OP_FRET:
            i->pc = i->code + i->fp[F_PC];
            i->sp = i->fp;
            Drop(i, VMCODEBYTE(i->pc - 1));
            i->fp = i->stack + i->fp[F_FP];
            break;
        case OP_SRET:
            i->pc = i->code + i->fp[F_PC];
            i->sp = i->fp;
            Drop(i, VMCODEBYTE(i->pc - 1) + 1);
            i->fp = i->stack + i->fp[F_FP];
            break;
        case OP_DUP:
            tmp = *(i)->sp;
            CPush(i, tmp);
            break;
        case OP_DROP:
            Drop(i, 1);
            break;
        case OP_PRINTI:
            tmp = VMCODEBYTE(i->pc++);
            PrintInteger(i, tmp, Pop(i));
            break;
        case OP_PRINTS:
            tmp = VMCODEBYTE(i->pc++);
            PrintString(i, tmp, i->sp[0]);
            StringRelease(i, Pop(i));
            break;
        case OP_PRINTT:
            tmp = VMCODEBYTE(i->pc++);
            PrintTab(i, tmp);
            break;
        case OP_PRINTN:
            tmp = VMCODEBYTE(i->pc++);
            PrintNewline(i, tmp);
            break;
        case OP_GETLINE:
            tmp = VMCODEBYTE(i->pc++);
            InputGetLine(i, tmp);
            break;
        case OP_INPUTI:
            tmp = VMCODEBYTE(i->pc++);
            InputInteger(i, tmp);
            break;
        case OP_INPUTS:
            tmp = VMCODEBYTE(i->pc++);
            InputString(i, tmp);
            break;
        default:
            Abort(i, str_opcode_err, VMCODEBYTE(i->pc - 1));
            break;
        }
    }
}

static void StringCat(Interpreter *i)
{
    int16_t hStr2 = Pop(i);
    int16_t hStr1 = i->sp[0];
    int len1, inHeap1, len2, inHeap2;
    uint8_t *str1, *str2, *str;

    /* get the first string pointer and length */
    str1 = GetStringPtr(i, hStr1, &inHeap1);
    len1 = GetStringLen(i, hStr1);

    /* get the second string pointer and length */
    str2 = GetStringPtr(i, hStr2, &inHeap2);
    len2 = GetStringLen(i, hStr2);

    /* allocate the result string */
    i->sp[0] = StringAlloc(i, (uint16_t)(len1 + len2));
    str = GetHeapStringPtr(i, i->sp[0]);

    /* copy the source strings into the result string */
    VMCOPY(str, str1, len1, inHeap1);
    VMCOPY(str + len1, str2, len2, inHeap2);

    /* adjust the reference counts */
    StringRelease(i, hStr1);
    StringRelease(i, hStr2);
}

static int StringCmp(Interpreter *i)
{
    int16_t hStr2 = Pop(i);
    int16_t hStr1 = i->sp[0];
    int len1, inHeap1, len2, inHeap2, cnt, cmp;
    uint8_t *str1, *str2;

    /* get the first string pointer and length */
    str1 = GetStringPtr(i, hStr1, &inHeap1);
    len1 = GetStringLen(i, hStr1);

    /* get the second string pointer and length */
    str2 = GetStringPtr(i, hStr2, &inHeap2);
    len2 = GetStringLen(i, hStr2);

    /* handle the case where the first string is shorter than the second */
    if (len1 < len2) {
        for (cnt = len1; cnt > 0; --cnt) {
            if ((cmp = VMGETBYTE(str1++, inHeap1) - VMGETBYTE(str2++, inHeap2)) != 0)
                break;
        }
        if (cnt <= 0)
            cmp = -1;
    }
    
    /* handle the case where the first string is longer or the same length as the first */
    else {
        for (cnt = len2; cnt > 0; --cnt) {
            if ((cmp = VMGETBYTE(str1++, inHeap1) - VMGETBYTE(str2++, inHeap2)) != 0)
                break;
        }
        if (cnt <= 0)
            cmp = (len1 == len2 ? 0 : 1);
    }

    /* adjust the reference counts */
    StringRelease(i, hStr1);
    StringRelease(i, hStr2);

    /* return the comparision result */
    return cmp;
}

static void PrintC(Interpreter *i, int dev, int ch)
{
    OutputDev *outDev = &i->outputDevs[dev];
    switch (dev) {
    case 0: /* terminal */
        VM_putchar(ch);
        break;
    case 1: /* display */
        VM_displaychar(ch);
        break;
    }
    if (ch == '\n')
        outDev->linePos = 0;
    else
        ++outDev->linePos;
}

static void PrintInteger(Interpreter *i, int dev, int16_t value)
{
    char buf[10], *p;
    itoa(value, buf, 10);
    for (p = buf; *p != '\0'; )
        PrintC(i, dev, *p++);
}

static void PrintString(Interpreter *i, int dev, int16_t value)
{
    int inHeap;
    uint8_t *str = GetStringPtr(i, value, &inHeap);
    int len = GetStringLen(i, value);
    while (--len >= 0)
        PrintC(i, dev, VMGETBYTE(str++, inHeap));
}

static void PrintTab(Interpreter *i, int dev)
{
    OutputDev *outDev = &i->outputDevs[dev];
    do {
        PrintC(i, dev, ' ');
    } while ((outDev->linePos % 8) != 0);
}

static void PrintNewline(Interpreter *i, int dev)
{
    PrintC(i, dev, '\n');
}

static char *GetInputToken(Interpreter *i)
{
    InputDev *inDev = &i->inputDevs[0];
    char *token, *p;
    int ch;
    
    /* make sure there is a current line */
    if (!inDev->inputPtr)
        return VMFALSE;
    
    /* skip leading spaces */
    for (token = inDev->inputPtr; *token != '\0' && isspace(*token); ++token)
        ;

    /* check for the end of line */
    if (*token == '\0' || *token == '\n') {
        inDev->inputPtr = NULL;
        return VMFALSE;
    }

    /* handle quoted strings */
    if (*token == '"') {
        p = ++token;
        while (*p != '\0' && *p != '\n' && *p != '"')
            ++p;
    }

    /* handle tokens without quotes */
    else {
        p = token;
        while (*p != '\0' && *p != '\n' && *p != ',')
            ++p;
    }

    /* skip past the terminator */
    if (*p == '\0' || *p == '\n')
        inDev->inputPtr = NULL;
    else {
        inDev->inputPtr = p;
        if (*inDev->inputPtr == '"') {
            while ((ch = *inDev->inputPtr) != '\0' && ch != '\n' && ch != ',')
                ++inDev->inputPtr;
        }
        if (*inDev->inputPtr == ',')
            ++inDev->inputPtr;
        else
            inDev->inputPtr = NULL;
    }

    /* strip trailing blanks */
    while (p > token) {
        if (isspace(p[-1]))
            --p;
        else
            break;
    }
    
    /* terminate the token string */
    *p = '\0';

    /* return the token */
    return token;
}

static void InputGetLine(Interpreter *i, int dev)
{
    InputDev *inDev = &i->inputDevs[0];
    VM_getline(inDev->inputBuf, sizeof(inDev->inputBuf));
    inDev->inputPtr = inDev->inputBuf;
}

static void InputInteger(Interpreter *i, int dev)
{
    char *token;
    while (!(token = GetInputToken(i)))
        InputGetLine(i, dev);
    CPush(i, atoi(token));
}

static void InputString(Interpreter *i, int dev)
{
    char *token, *str;
    uint16_t len;
    int16_t hStr;
    while (!(token = GetInputToken(i)))
        InputGetLine(i, dev);
    len = (uint16_t)strlen(token);
    hStr = StringAlloc(i, len);
    str = (char *)GetHeapStringPtr(i, hStr);
    memcpy(str, token, len);
    CPush(i, hStr);
}

void ShowStack(Interpreter *i)
{
    int16_t *p;
    for (p = i->stack; p <= i->sp; ++p) {
        if (p == i->fp)
            VM_printf(" <fp>");
        VM_printf(" %d", *p);
        if (*p < 0)
            VM_printf("(%04x)", *p);
    }
    VM_putchar('\n');
}

void StackOverflow(Interpreter *i)
{
    Abort(i, str_stack_overflow_err);
}

void Warn(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_printf(str_warn_prefix);
    VM_vprintf(fmt, ap);
    VM_putchar('\n');
    va_end(ap);
}

void Abort(Interpreter *i, const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    VM_printf(str_abort_prefix);
    VM_vprintf(fmt, ap);
    VM_putchar('\n');
    va_end(ap);
    if (i)
        longjmp(i->errorTarget, 1);
    else
        exit(1);
}
