///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2008
// 
// Filename: XGS_AVR_GFX_LIB_V010.c
// 
// Last Modified: 9.1.08
//
// Description: 
//
// Overview:
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// define chip type here
#define __AVR_ATmega644__

// define CPU frequency in Mhz here if not defined in Makefile, this is used for conditional compilation, so needs
// to be first, before other XGS includes, also note there is a copy of this constant in XGS_AVR_SYSTEM_V010, so if you 
// change it here, also change it in the header to make certain everthing is sync'ed
#define F_CPU 28636360UL

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/pgmspace.h>

// include local XGS API header files
#include "XGS_AVR_SYSTEM_V010.h"
#include "XGS_AVR_GFX_LIB_V010.h"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES //////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS/////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

volatile UCHAR *tile_map_ptr       = NULL;
volatile UCHAR *tile_bitmaps_ptr   = NULL;
volatile UCHAR *vram_ptr           = NULL;

// screen screen variable for bitmap and tile modes
int g_screen_width    = 0;
int g_screen_height   = 0;

int g_tile_map_width  = 0;
int g_tile_map_height = 0;

// clipping variables
int g_clip_min_x      = 0;
int g_clip_max_x      = 0;
int g_clip_min_y      = 0;
int g_clip_max_y      = 0;

// these globals keep track of the tile map console printing cursor position
int g_tmap_cursor_x = 0;
int g_tmap_cursor_y = 0;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// EXTERNALS //////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// all of these are exported from the ASM drivers always
volatile extern unsigned int  curr_line;
volatile extern unsigned char curr_raster_line;
volatile extern unsigned char tile_map_logical_width;
volatile extern char          tile_map_vscroll;
volatile extern unsigned char palette_table[];

// single missile for VGA tile mode 1 (exported from ASM driver)
volatile extern unsigned char missile_x;
volatile extern unsigned char missile_y; 
volatile extern unsigned char missile_color;
volatile extern unsigned char missile_state;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PROTOTYPES /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FUNCTIONS  /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GRAPHICS FUNCTIONS /////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void GFX_Init(int video_interrupt_rate)
{
// this function initializes the I/O ports for NTSC/PAL/VGA graphics as well as initilizes variables and turns on the 
// driver interrupt which will immediately start feeding the data sources to the currently installed ASM driver, thus
// make sure any pointers to data structures are initialized before calling this function
// when done with graphics and you want to stop calling the graphics interrupt, call GFX_Shutdown(..)
//
// Inputs:
//     video_interrupt_rate - the clock divider, so that the number of clocks per interrupt is equal to the HSYNC rate
//                            since the interrupt handler runs every HSYNC in both NTSC/PAL and VGA modes

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// VIDEO INTERRUPT EXPERIMENT /////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// main graphics port that luma/chroma and VGA signals come from, port A
#define GFX_MAIN_PORT       PORTA
#define GFX_MAIN_DDR        DDRA

// the "saturation" control is connected to port B
#define GFX_SATURATION_PORT PORTB
#define GFX_SATURATION_DDR  DDRB

// set video ports to output
GFX_MAIN_DDR  = 0xFF; // NTSC/PAL LUMA7:4 | CHROMA3:0, VGA7:0 [vsync, hsync, B1,B0, G1,G0, G1,G0] 
              //                                        b7      b6    b5:4   b3:2   b1:0
GFX_MAIN_PORT = 0x00;

// set sat control bits PB2, PB3 as outputs
GFX_SATURATION_DDR  =  SETPORTBITS(0,0,0,0,1,1,0,0);
PORTB               =  SETPORTBITS(0,0,0,0,1,1,0,0);

// set up timer                                     
TCCR1A = (0 << COM1A1) | (0 << COM1A0) | (0 << COM1B1) | (0 << COM1B0) | (0 << WGM11) | (0 << WGM10);

// set clock pre-scaler and counter mode
TCCR1B = (0 << WGM13) | (1 << WGM12) | (0 << CS12) | (0 << CS11)  | (1 << CS10);    

// 168. 644 use TIMSK1,TIFR1
// 162 use TIMSK,TIFR

// enable timer interrupts for timer 
TIMSK1 = (1 << OCIE1A);                  //(1<<TOIE1);                 

// clear and pendings timer x CTC interrupts (1 clears it)
TIFR1  = (1 << OCF1A); // (1<<TOV1);                  

// set output compare register value
OCR1A  = video_interrupt_rate; // 1817+-1 for NTSC at 28.636 Mhz = 15.748Hkz hsync rate, 910 for VGA @ 28.636Mhz = 31.468Khz, hsync rate

// enable interrupt
sei();

} // end GFX_Init

void GFX_Stop(void)
{
TIMSK1 = (0 << OCIE1A);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void GFX_Fill_Screen(int c, UCHAR *vbuffer)
{
// this functions fills the entire screen with the sent 2-bit color

memset( vbuffer, (UCHAR)( c | (c << 2) | (c << 4) | (c << 6) ), (g_screen_width >> 2) * g_screen_height );

} // end GFX_Fill_Screen

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void GFX_Fill_Region(int c, int y1, int y2, UCHAR *vbuffer)
{
// this functions fills a vertical region from y1 to y2 with the sent 2-bit color

int ytemp;

// clip
if ( y1 < 0 )
  y1 = 0;
else 
if ( y1 >= g_screen_height )
  y1 = g_screen_height-1;

if ( y2 < 0 )
  y2 = 0;
else
if ( y2 >= g_screen_height )
  y2 = g_screen_height-1;

// swap?
if (y2 < y1)
    {
    ytemp = y1;
    y1 = y2;
    y2 = ytemp;
    } // end if

// and fill memory
memset( vbuffer + (y1*(g_screen_width >> 2)), (UCHAR)( c | (c << 2) | (c << 4) | (c << 6) ), (g_screen_width >> 2) * (y2 - y1 + 1) );

} // end GFX_Fill_Region


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

inline void GFX_Plot(int x, int y, int c, UCHAR *vbuffer)
{
// plots a pixel on the screen at (x,y) in color c
// there are 2 bits per pixel, thus to locate the pixel, we need to divide the x by 4 and add that
// to the y * SCREEN_WIDTH/4, then that byte contains the pixel, next we need to shift the color over 0..3 pixels over
// where each pixel is 2-bit, and AND/OR mask it into the video byte
// Note: later convert to pure ASM, also try optimizing common terms (compiler should do this?)
// shifts are slower than multiples, so on AVR the mutliplications should be faster?

vbuffer [ (x >> 2) + (y * (g_screen_width >> 2) ) ] = 
                 (vbuffer [ (x >> 2) + (y * (g_screen_width >> 2)) ] & ~(0x03 << 2*(x & 0x03)) ) | (c << 2*(x & 0x03));

} // end GFX_Plot

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

inline void GFX_Plot_Clipped(int x, int y, int c, UCHAR *vbuffer)
{
// plots a pixel on the screen at (x,y) in color c
// there are 2 bits per pixel, thus to locate the pixel, we need to divide the x by 4 and add that
// to the y * SCREEN_WIDTH/4, then that byte contains the pixel, next we need to shift the color over 0..3 pixels over
// where each pixel is 2-bit, and AND/OR mask it into the video byte
// Note: later convert to pure ASM, also try optimizing common terms (compiler should do this?)
// shifts are slower than multiples, so on AVR the mutliplications should be faster?

// this version clips to the clipping rectangle which is global
if ( (x > g_clip_max_x) || ( x < g_clip_min_x ) || ( y > g_clip_max_y ) || ( y < g_clip_min_y ) )
    return;

// write the pixel to video RAM
vbuffer [ (x >> 2) + (y * (g_screen_width >> 2)) ] = 
                (vbuffer [ (x >> 2) + (y * (g_screen_width >> 2)) ] & ~(0x03 << 2*(x & 0x03)) ) | (c << 2*(x & 0x03));

} // end GFX_Plot_Clipped

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

unsigned char GFX_Get_Pixel(int x, int y, UCHAR *vbuffer)
{
// returns the 2-bit pixel code at screen location (x,y)
// there are 2 bits per pixel, thus to locate the pixel, we need to divide the x by 4 and add that
// to the y * SCREEN_WIDTH/4, then that byte contains the pixel, once retrieved, we need to shift the color over 0..3 pixels over
// where each pixel is 2-bit, and AND mask it to retrieve the pixel and then return it to caller
// Note: later convert to pure ASM, also try optimizing common terms (compiler should do this?)
// shifts are slower than multiples, so on AVR the mutliplications should be faster?

return ( (vbuffer [ (x >> 2) + (y * (g_screen_width >> 2)) ] >> (2*(x & 0x03))) & (0x03) );

} // end GFX_Get_Pixel

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void GFX_Set_Clipping_Rect(int clip_min_x, int clip_min_y, int clip_max_x, int clip_max_y)
{
// this function sets the clipping rectangle for ALL graphics functions that clip to a 2D viewplane; NTSC, VGA, etc.

g_clip_min_x = clip_min_x;
g_clip_max_x = clip_max_x;
g_clip_min_y = clip_min_y;
g_clip_max_y = clip_max_y;

} // end GFX_Set_Clipping_Rect

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void GFX_Draw_Clip_Line(int x0,int y0, int x1, int y1, int color, UCHAR *dest_buffer)
{
// this function draws a wireframe triangle

int cxs, cys,
	cxe, cye;

// clip and draw each line
cxs = x0;
cys = y0;
cxe = x1;
cye = y1;

// clip the line
if (GFX_Clip_Line(&cxs,&cys,&cxe,&cye))
	GFX_Draw_Line(cxs, cys, cxe,cye,color,dest_buffer);

} // end GFX_Draw_Clip_Line

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int GFX_Clip_Line(int *_x1,int *_y1,int *_x2, int *_y2)
{
// this function clips the sent line using the globally defined clipping
// region

// internal clipping codes
#define CLIP_CODE_C  0x0000
#define CLIP_CODE_N  0x0008
#define CLIP_CODE_S  0x0004
#define CLIP_CODE_E  0x0002
#define CLIP_CODE_W  0x0001

#define CLIP_CODE_NE 0x000a
#define CLIP_CODE_SE 0x0006
#define CLIP_CODE_NW 0x0009
#define CLIP_CODE_SW 0x0005

int x1=*_x1,
    y1=*_y1,
	x2=*_x2,
	y2=*_y2;

int xc1=x1,
    yc1=y1,
	xc2=x2,
	yc2=y2;

int p1_code=0,
    p2_code=0;

// determine codes for p1 and p2
if (y1 < g_clip_min_y)
	p1_code|=CLIP_CODE_N;
else
if (y1 > g_clip_max_y)
	p1_code|=CLIP_CODE_S;

if (x1 < g_clip_min_x)
	p1_code|=CLIP_CODE_W;
else
if (x1 > g_clip_max_x)
	p1_code|=CLIP_CODE_E;

if (y2 < g_clip_min_y)
	p2_code|=CLIP_CODE_N;
else
if (y2 > g_clip_max_y)
	p2_code|=CLIP_CODE_S;

if (x2 < g_clip_min_x)
	p2_code|=CLIP_CODE_W;
else
if (x2 > g_clip_max_x)
	p2_code|=CLIP_CODE_E;

// try and trivially reject
if ((p1_code & p2_code))
	return(0);

// test for totally visible, if so leave points untouched
if (p1_code==0 && p2_code==0)
	return(1);

// determine end clip point for p1
switch(p1_code)
	  {
	  case CLIP_CODE_C: break;

	  case CLIP_CODE_N:
		   {
		   yc1 = g_clip_min_y;
		   xc1 = x1 + 0.5+(g_clip_min_y-y1)*(x2-x1)/(y2-y1);
		   } break;
	  case CLIP_CODE_S:
		   {
		   yc1 = g_clip_max_y;
		   xc1 = x1 + 0.5+(g_clip_max_y-y1)*(x2-x1)/(y2-y1);
		   } break;

	  case CLIP_CODE_W:
		   {
		   xc1 = g_clip_min_x;
		   yc1 = y1 + 0.5+(g_clip_min_x-x1)*(y2-y1)/(x2-x1);
		   } break;

	  case CLIP_CODE_E:
		   {
		   xc1 = g_clip_max_x;
		   yc1 = y1 + 0.5+(g_clip_max_x-x1)*(y2-y1)/(x2-x1);
		   } break;

	// these cases are more complex, must compute 2 intersections
	  case CLIP_CODE_NE:
		   {
		   // north hline intersection
		   yc1 = g_clip_min_y;
		   xc1 = x1 + 0.5+(g_clip_min_y-y1)*(x2-x1)/(y2-y1);

		   // test if intersection is valid, of so then done, else compute next
			if (xc1 < g_clip_min_x || xc1 > g_clip_max_x)
				{
				// east vline intersection
				xc1 = g_clip_max_x;
				yc1 = y1 + 0.5+(g_clip_max_x-x1)*(y2-y1)/(x2-x1);
				} // end if

		   } break;

	  case CLIP_CODE_SE:
      	   {
		   // south hline intersection
		   yc1 = g_clip_max_y;
		   xc1 = x1 + 0.5+(g_clip_max_y-y1)*(x2-x1)/(y2-y1);

		   // test if intersection is valid, of so then done, else compute next
		   if (xc1 < g_clip_min_x || xc1 > g_clip_max_x)
		      {
			  // east vline intersection
			  xc1 = g_clip_max_x;
			  yc1 = y1 + 0.5+(g_clip_max_x-x1)*(y2-y1)/(x2-x1);
			  } // end if

		   } break;

	  case CLIP_CODE_NW:
      	   {
		   // north hline intersection
		   yc1 = g_clip_min_y;
		   xc1 = x1 + 0.5+(g_clip_min_y-y1)*(x2-x1)/(y2-y1);

		   // test if intersection is valid, of so then done, else compute next
		   if (xc1 < g_clip_min_x || xc1 > g_clip_max_x)
		      {
			  xc1 = g_clip_min_x;
		      yc1 = y1 + 0.5+(g_clip_min_x-x1)*(y2-y1)/(x2-x1);
			  } // end if

		   } break;

	  case CLIP_CODE_SW:
		   {
		   // south hline intersection
		   yc1 = g_clip_max_y;
		   xc1 = x1 + 0.5+(g_clip_max_y-y1)*(x2-x1)/(y2-y1);

		   // test if intersection is valid, of so then done, else compute next
		   if (xc1 < g_clip_min_x || xc1 > g_clip_max_x)
		      {
			  xc1 = g_clip_min_x;
		      yc1 = y1 + 0.5+(g_clip_min_x-x1)*(y2-y1)/(x2-x1);
			  } // end if

		   } break;

	  default:break;

	  } // end switch

// determine clip point for p2
switch(p2_code)
	  {
	  case CLIP_CODE_C: break;

	  case CLIP_CODE_N:
		   {
		   yc2 = g_clip_min_y;
		   xc2 = x2 + (g_clip_min_y-y2)*(x1-x2)/(y1-y2);
		   } break;

	  case CLIP_CODE_S:
		   {
		   yc2 = g_clip_max_y;
		   xc2 = x2 + (g_clip_max_y-y2)*(x1-x2)/(y1-y2);
		   } break;

	  case CLIP_CODE_W:
		   {
		   xc2 = g_clip_min_x;
		   yc2 = y2 + (g_clip_min_x-x2)*(y1-y2)/(x1-x2);
		   } break;

	  case CLIP_CODE_E:
		   {
		   xc2 = g_clip_max_x;
		   yc2 = y2 + (g_clip_max_x-x2)*(y1-y2)/(x1-x2);
		   } break;

		// these cases are more complex, must compute 2 intersections
	  case CLIP_CODE_NE:
		   {
		   // north hline intersection
		   yc2 = g_clip_min_y;
		   xc2 = x2 + 0.5+(g_clip_min_y-y2)*(x1-x2)/(y1-y2);

		   // test if intersection is valid, of so then done, else compute next
			if (xc2 < g_clip_min_x || xc2 > g_clip_max_x)
				{
				// east vline intersection
				xc2 = g_clip_max_x;
				yc2 = y2 + 0.5+(g_clip_max_x-x2)*(y1-y2)/(x1-x2);
				} // end if

		   } break;

	  case CLIP_CODE_SE:
      	   {
		   // south hline intersection
		   yc2 = g_clip_max_y;
		   xc2 = x2 + 0.5+(g_clip_max_y-y2)*(x1-x2)/(y1-y2);

		   // test if intersection is valid, of so then done, else compute next
		   if (xc2 < g_clip_min_x || xc2 > g_clip_max_x)
		      {
			  // east vline intersection
			  xc2 = g_clip_max_x;
			  yc2 = y2 + 0.5+(g_clip_max_x-x2)*(y1-y2)/(x1-x2);
			  } // end if

		   } break;

	  case CLIP_CODE_NW:
      	   {
		   // north hline intersection
		   yc2 = g_clip_min_y;
		   xc2 = x2 + 0.5+(g_clip_min_y-y2)*(x1-x2)/(y1-y2);

		   // test if intersection is valid, of so then done, else compute next
		   if (xc2 < g_clip_min_x || xc2 > g_clip_max_x)
		      {
			  xc2 = g_clip_min_x;
		      yc2 = y2 + 0.5+(g_clip_min_x-x2)*(y1-y2)/(x1-x2);
			  } // end if

		   } break;

	  case CLIP_CODE_SW:
		   {
		   // south hline intersection
		   yc2 = g_clip_max_y;
		   xc2 = x2 + 0.5+(g_clip_max_y-y2)*(x1-x2)/(y1-y2);

		   // test if intersection is valid, of so then done, else compute next
		   if (xc2 < g_clip_min_x || xc2 > g_clip_max_x)
		      {
			  xc2 = g_clip_min_x;
		      yc2 = y2 + 0.5+(g_clip_min_x-x2)*(y1-y2)/(x1-x2);
			  } // end if

		   } break;

	  default:break;

	  } // end switch

// do bounds check
if ((xc1 < g_clip_min_x) || (xc1 > g_clip_max_x) ||
	(yc1 < g_clip_min_y) || (yc1 > g_clip_max_y) ||
	(xc2 < g_clip_min_x) || (xc2 > g_clip_max_x) ||
	(yc2 < g_clip_min_y) || (yc2 > g_clip_max_y) )
	{
	return(0);
	} // end if

// store vars back
*_x1 = xc1;
*_y1 = yc1;
*_x2 = xc2;
*_y2 = yc2;

return(1);

} // end GFX_Clip_Line

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void GFX_Draw_Line(int x0, int y0,  // starting position
                   int x1, int y1,  // ending position
                   int color,       // color index
                   UCHAR *vb_start) // video buffer
{
// this function draws a line from x0,y0 to x1,y1 using differential error analysis
// terms (based on Bresenahams work)

int dx,             // difference in x's
    dy,             // difference in y's
    dx2,            // dx,dy * 2
    dy2,
    x_inc,          // amount in pixel space to move during drawing
    y_inc,          // amount in pixel space to move during drawing
    error,          // the discriminant i.e. error i.e. decision variable
    index;          // used for looping

// pre-compute first pixel address in video buffer
// vb_start = vb_start + x0 + y0*lpitch;


// compute horizontal and vertical deltas
dx = x1-x0;
dy = y1-y0;

// test which direction the line is going in i.e. slope angle
if (dx>=0)
   {
   x_inc = 1;

   } // end if line is moving right
else
   {
   x_inc = -1;
   dx    = -dx;  // need absolute value

   } // end else moving left

// test y component of slope

if (dy>=0)
   {
   y_inc = 1;
   } // end if line is moving down
else
   {
   y_inc = -1;
   dy    = -dy;  // need absolute value

   } // end else moving up

// compute (dx,dy) * 2
dx2 = dx << 1;
dy2 = dy << 1;

// now based on which delta is greater we can draw the line
if (dx > dy)
   {
   // initialize error term
   error = dy2 - dx;

   // draw the line
   for (index=0; index <= dx; index++)
       {
       // set the pixel
       // *vb_start = color;

       // plot the pixel, this can definitely be optimized with pre-computation etc. however, due to the pixel packing, it slows
       // down the direct addressing optimizations available with 1 byte per pixel
       vb_start [ (x0 >> 2) + (y0 * (g_screen_width >> 2) ) ] = 
                 (vb_start [ (x0 >> 2) + (y0 * (g_screen_width >> 2)) ] & ~(0x03 << 2*(x0 & 0x03)) ) | (color << 2*(x0 & 0x03));

       // test if error has overflowed
       if (error >= 0)
          {
          error-=dx2;

          // move to next line
          y0 += y_inc;

	   } // end if error overflowed

       // adjust the error term
       error+=dy2;

       // move to the next pixel
       x0 += x_inc;

       } // end for

   } // end if |slope| <= 1
else
   {
   // initialize error term
   error = dx2 - dy;

   // draw the line
   for (index=0; index <= dy; index++)
       {
       // set the pixel
       //*vb_start = color;

       // plot the pixel, this can definitely be optimized with pre-computation etc. however, due to the pixel packing, it slows
       // down the direct addressing optimizations available with 1 byte per pixel
       vb_start [ (x0 >> 2) + (y0 * (g_screen_width >> 2) ) ] = 
                 (vb_start [ (x0 >> 2) + (y0 * (g_screen_width >> 2)) ] & ~(0x03 << 2*(x0 & 0x03)) ) | (color << 2*(x0 & 0x03));

       // test if error overflowed
       if (error >= 0)
          {
          error-=dy2;

          // move to next line
          x0 += x_inc;

          } // end if error overflowed

       // adjust the error term
       error+=dx2;

       // move to the next pixel
       y0 += y_inc;

       } // end for

   } // end else |slope| > 1

} // end GFX_Draw_Line

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void GFX_Draw_HLine(int x1,int x2,int y, UCHAR c, UCHAR *vbuffer)
{
// draw a horizontal line using the memset function, end points are computed as special case, then the longest run
// of 4-pixels at a time is rendered ...|XXXXXXXXXXXXXXXXXXXXXXXXX|...
// clips to clipping rectangle, code is long due to special cases, but is fast

int temp;                    // used for temporary storage during endpoint swap
int xs, xe, dx;              // holds start/end byte to render into
UCHAR xp1, xp2;              // holds sub-pixel accurate pixels in each 4-pixel packed block
UCHAR mask, pixels;          // holds bit mask and pixels being drawn
unsigned int mask16, pixels16; // holds 16 bit version, bit mask and pixels being drawn

// these tables are for the special case of < 4 pixels, since it can straddle a byte boundary
// these lookups will cut down on computation
UCHAR colors[3*4] = { 0b00000000, 0b00000001, 0b00000010,  0b00000011,
                      0b00000000, 0b00000101, 0b00001010,  0b00001111,
                      0b00000000, 0b00010101, 0b00101010,  0b00111111 };

UCHAR bitmasks[1*3] = { 0b00000011, 0b00001111, 0b00111111 }; 

// perform trivial rejections
if (y > g_clip_max_y || y < g_clip_min_y)
   return;

// sort x1 and x2, so that x2 > x1
if (x1 > x2)
   {
   temp = x1;
   x1   = x2;
   x2   = temp;
   } // end swap

// perform trivial rejections
if (x1 > g_clip_max_x || x2 < g_clip_min_x)
   return;

// now clip
x1 = ((x1 < g_clip_min_x) ? g_clip_min_x : x1);
x2 = ((x2 > g_clip_max_x) ? g_clip_max_x : x2);

// advance vbuffer to the proper line
vbuffer += (y * (g_screen_width >> 2));

// draw the row of pixels, special case needed for endpoints due to pixel packing per byte [p3 | p2 | p1 | p0] 
// which are drawn [p0 | p1 | p2 | p3]

// first test if there are less than 4 pixels to draw? This is a special icky case since it can straddle a 
if ( (dx = 1 + x2 - x1) < 4 )
    {
    // compute initial byte position and pixel within byte
    xs  = x1 >> 2;            // xs  = x1 / 4
    xp1 = (x1 & 0x03) << 1;   // xp1 = (x1 mod 4)*2

    // set up to draw 1-3 pixels of color 0,1,2,3
    pixels16 = colors[ ( (dx-1) << 2) + c ] << xp1;
    mask16   = bitmasks[ (dx-1) ] << xp1;

    // advance screen pointer to start x postion
    vbuffer += xs;

    // write the pixels at one time with 16-bit write
    *(unsigned int *)vbuffer = ( (*(unsigned int *)vbuffer & ~mask16) | pixels16);

    // return
    return;
    } // end if

xs  = x1 >> 2;   // xs  = x1 / 4
xp1 = x1 & 0x03; // xp1 = x1 mod 4;

// which case are we dealing with?
switch (xp1)
    {
    case 1:
        {
        // draw pixel 1,2,3, _XXX|
        mask    = ~0b11111100;
        pixels  = ( c << 2 ) | ( c << 4) | (c << 6);
        vbuffer[ xs ] = (vbuffer[ xs ] & mask) | pixels;

        // advance xs 
        xs++;
        } break;
    
    case 2:
        {
        // draw pixel 2,3, __XX|
        mask    = ~0b11110000;
        pixels  = ( c << 4) | (c << 6);
        vbuffer[ xs ] = (vbuffer[ xs ] & mask) | pixels;

        // advance xs 
        xs++;

        } break;

    case 3:
        {
        // draw pixel 3, ___X|
        mask    = ~0b11000000;
        pixels  = (c << 6);
        vbuffer[ xs ] = (vbuffer[ xs ] & mask) | pixels;

        // advance xs 
        xs++;

        } break;

    } // end switch

// compute end point metrics
xe  = (x2+1) >> 2;  // xe  = (x2+1) / 4
xp2 = x2 & 0x03;    // xp2 = x2 mod 4;

// draw solid bytes in between endpoints, compute how many solid blocks of 4 pixels to draw?
dx = xe - xs;

pixels = ( (c) | ( c << 2 ) | ( c << 4) | (c << 6) );
memset ( vbuffer + xs, pixels, dx );

// and finally, the trailing pixels
xs += dx;

// which case are we dealing with?
switch (xp2)
    {
    case 0:
        {
        // draw pixel 0, X___|
        mask    = ~0b00000011;
        pixels  = (c);
        vbuffer[ xs ] = (vbuffer[ xs ] & mask) | pixels;

        } break;
    
    case 1:
        {
        // draw pixel 0,1, XX__|
        mask    = ~0b00001111;
        pixels  = ( c ) | (c << 2);
        vbuffer[ xs ] = (vbuffer[ xs ] & mask) | pixels;

        } break;

    case 2:
        {
        // draw pixel 0,1,2, XXX_|
        mask    = ~0b00111111;
        pixels  = ( c ) | ( c << 2 ) | ( c << 4);
        vbuffer[ xs ] = (vbuffer[ xs ] & mask) | pixels;

        } break;

    } // end switch

} // end GFX_Draw_HLine

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void GFX_Draw_VLine(int y1,int y2,int x,int c,UCHAR *vbuffer)
{
// draw a vertical line, note that a memset function can no longer be
// used since the pixel addresses are no longer contiguous in memory
// note that the end points of the line must be on the screen

UCHAR *start_offset; // starting memory offset of line

int index, // loop index
    temp;  // used for temporary storage during swap

UCHAR xp1;          // holds sub-pixel accurate pixels in each 4-pixel packed block
UCHAR mask, pixels; // holds bit mask and pixels being drawn

// perform trivial rejections
if (x > g_clip_max_x || x < g_clip_min_x)
   return;

// make sure y2 > y1
if (y1>y2)
   {
   temp = y1;
   y1   = y2;
   y2   = temp;
   } // end swap

// perform trivial rejections
if (y1 > g_clip_max_y || y2 < g_clip_min_y)
   return;

// now clip
y1 = ((y1 < g_clip_min_y) ? g_clip_min_y : y1);
y2 = ((y2 > g_clip_max_y) ? g_clip_max_y : y2);

// compute starting position byte position at x, y1
start_offset = vbuffer + y1*(g_screen_width >> 2) + (x >> 2);

// build mask and pixels
// compute initial byte position and pixel within byte
xp1 = ((x & 0x03) << 1); // xp1 = (x1 mod 4)*2

// get pixel and mask ready to iterate pixels into screen
pixels = (c << xp1);
mask   = (0x03 << xp1);

// draw line one pixel at a time
for (index=0; index<=y2-y1; index++)
    {
    // set the pixel, mask first then OR pixels
    *start_offset = ( (*start_offset & ~mask) | pixels);

    // move downward to next line
    start_offset+=(g_screen_width >> 2);

    } // end for index

} // end GFX_Draw_VLine

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

GFX_Draw_Circle(int x0, int y0, int r, int c, UCHAR *vbuffer)
{
// draws a circle at (x0, y0) with radius r and color c, based on DDA, midpoint algorithm, aka bresenham's circle
// basically uses differential analysis, that is the change from one iteration to another to compute the terms
// of the curve
int f = 1 - r;
int ddF_x = 1;
int ddF_y = -2 * r;
int x = 0;
int y = r;

// draw the 4 endpoints
GFX_Plot(x0, y0 + r, c, vbuffer);
GFX_Plot(x0, y0 - r, c, vbuffer);
GFX_Plot(x0 + r, y0, c, vbuffer);
GFX_Plot(x0 - r, y0, c, vbuffer);

// enter loop and rasterize curve
while (x < y) 
    {
    
    if(f >= 0) 
      {
      y--;
      ddF_y += 2;
      f += ddF_y;
      } // end if

  x++;
  ddF_x += 2;
  f += ddF_x;    

  // draw all octants
  GFX_Plot(x0 + x, y0 + y, c, vbuffer);
  GFX_Plot(x0 - x, y0 + y, c, vbuffer);
  GFX_Plot(x0 + x, y0 - y, c, vbuffer);
  GFX_Plot(x0 - x, y0 - y, c, vbuffer);
  GFX_Plot(x0 + y, y0 + x, c, vbuffer);
  GFX_Plot(x0 - y, y0 + x, c, vbuffer);
  GFX_Plot(x0 + y, y0 - x, c, vbuffer);
  GFX_Plot(x0 - y, y0 - x, c, vbuffer);

  } // end while

} // GFX_Draw_Circle

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

GFX_Draw_Circle_Clipped(int x0, int y0, int r, int c, UCHAR *vbuffer)
{
// draws a circle at (x0, y0) with radius r and color c, based on DDA, midpoint algorithm, aka bresenham's circle
// basically uses differential analysis, that is the change from one iteration to another to compute the terms
// of the curve
int f = 1 - r;
int ddF_x = 1;
int ddF_y = -2 * r;
int x = 0;
int y = r;

// first test for trivial rejection of circle
if ( (x0+r < 0) || (x0-r >= g_screen_width) || (y0+r < 0) || (y0-r >= g_screen_height) )
    return;

// draw the 4 endpoints
GFX_Plot_Clipped(x0, y0 + r, c, vbuffer);
GFX_Plot_Clipped(x0, y0 - r, c, vbuffer);
GFX_Plot_Clipped(x0 + r, y0, c, vbuffer);
GFX_Plot_Clipped(x0 - r, y0, c, vbuffer);

// enter loop and rasterize curve
while (x < y) 
    {
    
    if(f >= 0) 
      {
      y--;
      ddF_y += 2;
      f += ddF_y;
      } // end if

  x++;
  ddF_x += 2;
  f += ddF_x;    

  // draw all octants
  GFX_Plot_Clipped(x0 + x, y0 + y, c, vbuffer);
  GFX_Plot_Clipped(x0 - x, y0 + y, c, vbuffer);
  GFX_Plot_Clipped(x0 + x, y0 - y, c, vbuffer);
  GFX_Plot_Clipped(x0 - x, y0 - y, c, vbuffer);
  GFX_Plot_Clipped(x0 + y, y0 + x, c, vbuffer);
  GFX_Plot_Clipped(x0 - y, y0 + x, c, vbuffer);
  GFX_Plot_Clipped(x0 + y, y0 - x, c, vbuffer);
  GFX_Plot_Clipped(x0 - y, y0 - x, c, vbuffer);
  } // end while

} // GFX_Draw_Circle_Clipped
