/* xgsbasic.c - xgs basic compiler command line program
 *
 * Copyright (c) 2009-2010 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <setjmp.h>
#include <string.h>
#include <ctype.h>
#include "../compiler/db_compiler.h"

/* error target */
static jmp_buf errorTarget;

/* local function prototypes */
static void Usage(void);
static void Download(const char *port, int boardType, const char *file, int terminalMode);

int main(int argc, char *argv[])
{
    int boardType, compileOnly, terminalMode, debugMode, i;
    char *file, *port, *p;

    /* initialize */
    file = NULL;
    boardType = BOARD_UNKNOWN;
    port = "COM1";
    compileOnly = VMFALSE;
    terminalMode = VMFALSE;
    debugMode = VMFALSE;

    /* get the arguments */
    for(i = 1; i < argc; ++i) {

        /* handle switches */
        if(argv[i][0] == '-') {
            switch(argv[i][1]) {
            case 'b':
                if(argv[i][2])
                    p = &argv[i][2];
                else if(++i < argc)
                    p = argv[i];
                else
                    Usage();
                if (strcasecmp(p, "ca") == 0)
                    boardType = BOARD_CHAMELEON_AVR;
                else if (strcasecmp(p, "cp") == 0)
                    boardType = BOARD_CHAMELEON_PIC;
                else if (strcasecmp(p, "xa") == 0)
                    boardType = BOARD_XGS_AVR;
                else if (strcasecmp(p, "xp") == 0)
                    boardType = BOARD_XGS_PIC;
                else
                    Usage();
                break;
            case 'c':
                compileOnly = VMTRUE;
                break;
            case 'd':
                debugMode = VMTRUE;
                break;
            case 'p':
                if(argv[i][2])
                    port = &argv[i][2];
                else if(++i < argc)
                    port = argv[i];
                else
                    Usage();
                break;
            case 't':
                terminalMode = VMTRUE;
                break;
            default:
                Usage();
                break;
            }
        }

        /* handle the input filename */
        else {
            if (file)
                Usage();
            file = argv[i];
        }
    }

    /* make sure a board type was specified */
    if (boardType == BOARD_UNKNOWN) {
        fprintf(stderr, "error: please specify either board type:\n");
        fprintf(stderr, "         -b ca     for Chameleon AVR\n");
        fprintf(stderr, "         -b cp     for Chameleon PIC\n");
        fprintf(stderr, "         -b xa     for XGS AVR\n");
        fprintf(stderr, "         -b xp     for XGS PIC\n");
        exit(1);
    }
    
    /* check for a numeric port */
    if (isdigit(port[0])) {
        static char buf[10];
        sprintf(buf, "\\\\.\\COM%d", atoi(port));
        port = buf;
    }

    /* make sure there was at least an input file specified */
    if (file) {

        /* ".bas" is a source file */
        if (IsSourceFile(file)) {
            char outfile[_MAX_PATH];
            ConstructOutputName(file, outfile);
            if (CompileFile(file, outfile, boardType, debugMode) && !compileOnly)
                Download(port, boardType, outfile, terminalMode);
        }

        /* ".bai" is an already compiled image file */
        else if (IsImageFile(file))
            Download(port, boardType, file, terminalMode);

        /* unknown file extension */
        else
            Usage();
    }

	/* no file specified */
	else if (terminalMode)
		TerminalMode(port, boardType);

    /* exit successfully */
    return 0;
}

/* Usage - display a usage message and exit */
static void Usage(void)
{
    fprintf(stderr, "\
usage: xgsbasic\n\
         [ -b xa | xp | ca | cp ]   select an XGS AVR/PIC or a Chameleon AVR/PIC board\n\
         [ -c ]                     compile only and don't download\n\
         [ -p <port> ]              serial port (default is COM1)\n\
         [ -t ]                     enter terminal mode after optional compile/download\n\
         [ <name>.<ext> ]           <source>.bas or <compiled>.bai\n\
\n\
   for example:\n\
\n\
     chambasic -p COM6 pong.bas\n\
\n\
   will download the file 'pong.bas' to on-chip flash using serial port COM6\n\
\n\
   and:\n\
\n\
     chambasic life.bai\n\
\n\
   will download the already compiled file to on-chip flash\n\
\n\
   After the download is complete, terminal mode is entered to display\n\
   any output produced by the downloaded program.\n\
\n\
   If -t is given, no compile or download is performed and terminal mode\n\
   is entered directly.\n\
");
    exit(1);
}

/* Download - download a binary file to a board */
static void Download(const char *port, int boardType, const char *file, int terminalMode)
{
	if (!WriteImage(port, boardType, file, terminalMode))
        VM_printf("Download failed: %s\n", file);
}

void VM_printf(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end(ap);
}

void VM_vprintf(const char *fmt, va_list ap)
{
	vprintf(fmt, ap);
}

void VM_putchar(int ch)
{
	putchar(ch);
}


